/* 
 * (c) 2011 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */


#ifndef NMM_ADAPTBYLOOKUP_HPP
#define NMM_ADAPTBYLOOKUP_HPP NMM_ADAPTBYLOOKUP_HPP


#include "Adaptability.hpp"

#include <set>
#include <map>


namespace Adaptation {
  
  typedef std::map<double, OptimizedParameterSet*> plr_map;

  typedef std::map<int, plr_map*> rtt_map;

  typedef std::map<int, rtt_map*> num_recv_map;

  /** Base class for tables of different delay constraints.
   */
  class DelayCell
  {
    public :

      DelayCell(int delay);

      ~DelayCell();

      int getDelay() { return m_delay; }

      OptimizedParameterSet* getOptSet(int num_recv, int rtt, double plr);

      /** TODO use virtual pure methods for storing parameter sets into the table.
       * subclasses (table of different delay constraints) can add their parameter
       * sets by itself.
       */
      virtual void addOneReceiverMap();
      virtual void addTwoReceiversMap();
      virtual void addThreeReceiversMap();

      virtual void add12MbpsOneReceiverMap();
      virtual void add20MbpsOneReceiverMap();
      virtual void add12MbpsTwoReceiversMap();
      virtual void add12MbpsThreeReceiversMap();

      void removeMapOfNumReceivers(int num_recv);

      void removeMap();

    private :
      int m_delay;
      num_recv_map m_num_recv_map;

  };

  class TableFactory
  {
    public :
      void createCell();
  };

  /** Class for the AdaptByLookup of the PRRT.
   *
   * This class provides functions to adapting the PRRT on the fly, based on the feedback statistics
   * provided by the PRRT socket.
   */
  class AdaptByLookup : public Adaptability
  {
    public :
      /** @name Constructor & Destructor
       */
      //@{
      /** Create a new AdaptByLookup.
       *
       * This constructor creates an %AdaptByLookup function for the PRRT socket.
       */
      AdaptByLookup();

      /** Virtual Destructor.
       *
       * Releases the AdaptByLookup.
       */
      virtual ~AdaptByLookup();
      //@}
      
      /** Look up the adapted parameters set from the pre-computed optimized parameters sets.
       *
       * Look up the adapted parameters set from the pre-computed optimized parameters sets,
       * depending on the current socket statistics (PLR).
       *
       * @return a constant pointer to the adapted parameters set.
       */
      const OptimizedParameterSet* searchOptParamSet();
      

    
    protected :
      /** @name ThreadedObject methods
       */
      //@{
      /** Main thread method.
       * This method handles the computation of the parameters and applying them in a loop.
       * It runs in a separate thread owned by the Adapability.
       */
      virtual void run();
      //@}
      
      
    private :
      
      double smoothPLR(double plr, double fall_acceleration, double max_deviation, double min_deviation);
      uint16_t smoothRTT(uint16_t rtt, double fall_acceleration, double max_deviation, double min_deviation);

      uint16_t m_gplr;     ///< The global packet loss rate of the current socket statistics (need scale it with 65535).
      uint16_t m_grtt;     ///< The global round trip time of the current socket statistics.
      double m_plr;     ///< The global packet loss rate of the current socket statistics (need scale it with 65535).
      double m_rtt;     ///< The global round trip time of the current socket statistics.
      DelayCell m_delay200_cell;


  };

}

#endif
