/* 
 * (c) 2011 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "Adaptability.hpp"

#include <cerrno>       // for errnos

// For debug
#include <iostream>
#include <csignal>
#include <cstdlib>
#include <string.h>
#include <stdexcept>

using namespace media_transport_protocol;

namespace Adaptation {
  
  // Static members
  const int           Adaptability::STATS_SCALE     = 65535;
  
  Adaptability::Adaptability()
  : m_sock              ( 0 )
  , m_is_running        ( false )
  , m_update_interval   ( 200 )
  , m_opt_param_set     ( 0 )
  , m_packet_length         ( 1316 )
  , m_datarate              ( 4 )
  , m_packet_interval       ( m_packet_length * 8 / m_datarate / 1000 )
  {
    // Initialize mutex attributes
    pthread_mutexattr_t attr;
    pthread_mutexattr_init(&attr);
      // Recursive mutex (if locked again by the same thread => increment lock counter)
//       pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
      // Normal mutex (if locked again by the same thread => deadlock)
      pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_NORMAL);
      // Error-checking mutex (if locked again by the same thread => error)
      //pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_ERRORCHECK);

    // Initialize mutex
    pthread_mutex_init(&m_mutex, &attr);
    pthread_mutexattr_destroy(&attr);
  
    pthread_cond_init(&m_condition, 0);
    
    m_stats.deliveredData = 0;
    m_stats.deliveredRedundancy = 0;
    m_stats.lostPackets = 0;
    m_stats.droppedPackets = 0;
    m_stats.missedPackets = 0;
    m_stats.receivedData = 0;
    m_stats.receivedRedundancy = 0;
    m_stats.lastGRTT = 0;
    m_stats.lastPLR = 0;
    m_stats.lastGPLR = 0;
  }

  Adaptability::~Adaptability()
  {
    this->stop();
  }
  
  void Adaptability::setSock(prrt_socket *sock)
  {
    pthread_mutex_lock(&m_mutex);
    m_sock = sock;
    pthread_mutex_unlock(&m_mutex);
  }


  void Adaptability::start()
  {
    pthread_mutex_lock(&m_mutex);
    // If the adaptability is running, do nothing.
    if ( m_is_running ) {
      pthread_mutex_unlock(&m_mutex);
      return;
    }
    
    m_is_running = true;

    // Start adaptability thread
    try {
      if (pthread_create(&m_thread, 0, static_run, static_cast<void *>(this)) != 0) {
        std::cerr << "> Adaptability::start() pthread_create failed" << std::endl;
        throw std::runtime_error("not enough system resources to create a process for the new thread");
      }
    }
    catch(...) {
      std::cerr << "> Adaptability ERROR (start) : Unable to activate thread for the adaptability of the PRRT." << std::endl;
      pthread_mutex_unlock(&m_mutex);
      throw "Unable to activate thread for the adaptability of the PRRT.";
    }

    pthread_mutex_unlock(&m_mutex);
    pthread_cond_broadcast(&m_condition);
  }

  
  void Adaptability::stop()
  {
    pthread_mutex_lock(&m_mutex);
    if (m_is_running) {
      m_is_running = false;
    }
    else {
      pthread_mutex_unlock(&m_mutex);
      return;
    }
    pthread_mutex_unlock(&m_mutex);
    pthread_cond_broadcast(&m_condition);

    // Join thread
    if (pthread_join(m_thread, 0) == 0) {
      
    }
    else {
      std::cerr << "> Adaptability::stop() pthread_join failed" << std::endl;
    }
  }


  void Adaptability::applyOptParamSet(const OptimizedParameterSet *opt)
  {
    // Update the optimized parameter set.

    try {
      std::vector<unsigned> n_p ( opt->getRetransmissionRoundsVector() );
      n_p.insert( n_p.begin(), opt->getNumberOfPresentRedundancyPacketsPerBlock() );
      // we set the delay of sending feedbacks to 0 as default.
      int d_c = opt->getNumberOfDataPacketsPerBlock() * m_packet_interval;
      m_sock->set_param( param_AHEC( 200, d_c + m_stats.lastGRTT/2, d_c, m_stats.lastGRTT,
                         0, opt->getNumberOfDataPacketsPerBlock(), n_p ) ); 
    }
    catch (...) {
      std::cerr << "Adaptability applyOptParamSet() failed!" << std::endl;
    }
    
    m_opt_param_set = opt;
  }
  
  bool Adaptability::isOptParamSetChanged(const OptimizedParameterSet *opt) const
  {
    if (!m_opt_param_set || !opt) {
      if (!opt) {
        return false;
      }
      else {
        return true;
      }
    }
    return !(*m_opt_param_set == *opt);
  }
  

  bool Adaptability::isStatsChanged(uint32_t *sent_data, uint32_t *sent_red, uint32_t *lost, uint32_t *drop, uint32_t *miss,
                          uint32_t *recv_data, uint32_t *recv_red, uint16_t *grtt, uint16_t *plr, uint16_t *gplr) const
  {
    bool changed = false;
    
    // Is delivered data changed?
    if ( sent_data ) {
      unsigned long new_sent_data = m_stats.deliveredData;
      if (*sent_data != new_sent_data) {
        *sent_data = new_sent_data;
        changed = true;
      }
    }
    
    // Is delivered redundancy changed?
    if ( sent_red ) {
      unsigned long new_sent_red = m_stats.deliveredRedundancy;
      if (*sent_red != new_sent_red) {
        *sent_red = new_sent_red;
        changed = true;
      }
    }
    
    // Is lost packets changed?
    if ( lost ) {
      unsigned long new_lost = m_stats.lostPackets;
      if (*lost != new_lost) {
        *lost = new_lost;
        changed = true;
      }
    }
    
    // Is dropped packets changed?
    if ( drop ) {
      unsigned long new_drop = m_stats.droppedPackets;
      if (*drop != new_drop) {
        *drop = new_drop;
        changed = true;
      }
    }
    
    // Is missed packets changed?
    if ( miss ) {
      unsigned long new_miss = m_stats.missedPackets;
      if (*miss != new_miss) {
        *miss = new_miss;
        changed = true;
      }
    }
    
    // Is received data changed?
    if ( recv_data ) {
      unsigned long new_recv_data = m_stats.receivedData;
      if (*recv_data != new_recv_data) {
        *recv_data = new_recv_data;
        changed = true;
      }
    }
    
    // Is received redundancy changed?
    if ( recv_red ) {
      unsigned long new_recv_red = m_stats.receivedRedundancy;
      if (*recv_red != new_recv_red) {
        *recv_red = new_recv_red;
        changed = true;
      }
    }
    
    // Is last global RTT changed?
    if ( grtt ) {
      unsigned int new_grtt = m_stats.lastGRTT;
      if (*grtt != new_grtt) {
        *grtt = new_grtt;
        changed = true;
      }
    }
    
    // Is last PLR changed?
    if ( plr ) {
      unsigned int new_plr = m_stats.lastPLR;
      if (*plr != new_plr) {
        *plr = new_plr;
        changed = true;
      }
    }
    
    // Is last global PLR changed?
    if ( gplr ) {
      unsigned int new_gplr = m_stats.lastGPLR;
      if (*gplr != new_gplr) {
        *gplr = new_gplr;
        changed = true;
      }
    }
    
    return changed;
  }
  
  void Adaptability::updateStats()
  {
    try {
      m_sock->getStats( &m_stats );
    }
    catch ( prrt_error &e ) {
      std::cerr << "> Adaptability::updateStats() ERROR (update statistics)" << std::endl;
      throw "PRRT socket failed on Statistics.";
    }
  }

  void Adaptability::setUpdateInterval(int interval)
  {
    if (interval > 0) {
      m_update_interval = interval;
    }
    else {
      throw "Adaptability has invalid update interval.";
    }
  }

  int Adaptability::getUpdateInterval()
  {
    return m_update_interval;
  }

}

