/* 
 * (c) 2011 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */


#ifndef ADAPTABILITY_HPP
#define ADAPTABILITY_HPP ADAPTABILITY_HPP


#include <inttypes.h>   // for int types e.g. uint32_t

#include "OptimizedParameterSet.hpp"

/******** Include the PRRT headers ********/
#include <prrt/prrt_sockets.h>

#include <pthread.h>

 
namespace Adaptation {
  
  /** Class for the Adaptability of the PRRT.
   *
   * This class provides functions to adapting the PRRT on the fly, based on the feedback statistics
   * provided by the PRRT socket.
   */
  class Adaptability
  {
    public :
      
      // Use struct instead of class.
      class Statistics {
        public :
          
          uint32_t *sent_data;
          uint32_t *sent_red;
          uint32_t *lost;
          uint32_t *drop;
          uint32_t *miss;
          uint32_t *recv_data;
          uint32_t *recv_red;
          uint16_t *grtt;
          uint16_t *plr;
          uint16_t *gplr;
      };
      
      
      /** @name Constructor & Destructor
       */
      //@{
      /** Create a new Adaptability.
       *
       * This constructor creates an %Adaptability function for the PRRT socket.
       */
      Adaptability();

      /** Virtual Destructor.
       *
       * Releases the Adaptability.
       */
      virtual ~Adaptability();
      //@}
      
      /** @name Control methods
       */
      //@{
      /** Starts the adaptability.
       *
       * This method starts the adaptability for the PRRT socket.
       * If the adaptability is running, do nothing.
       */
      void start();

      /** Stops the adaptability.
       *
       * This method stops the adaptability of the PRRT socket.
       * If the adaptability is not running, do nothing.
       */
      void stop();
      
      /** Stores the pointer to the adapt PRRT socket.
       *
       * This method sets the adapt PRRT socket.
       *
       * @param sock a pointer to the adapt PRRT socket.
       */
      void setSock(media_transport_protocol::prrt_socket *sock);
      //@}

      
      /** Applies the optimized parameter set computed by the specific adaptation.
       *
       * Applies the current optimized parameter set on the PRRT socket and
       * saves it to the protected member m_opt_param_set.
       *
       * @param opt the parameter set to be applied which is computed by the specific adaptation.
       */
      virtual void applyOptParamSet(const OptimizedParameterSet *opt);
      

      /** Checks if the socket statistics of PRRT changes.
       *
       * Checks if the given statistics among all the socket statistics changes and
       * updates the given statistics if there is a change on them.
       * NOTE The subclasses may override this method with some purpose e.g. adding
       * threshold to determine the change.
       *
       * @param sent_data a pointer to the number of the packets having been sent.
       * @param sent_red a pointer to the number of the redundancy packets having been sent.
       * @param lost a pointer to the number of the lost packets on the transmission.
       * @param drop a pointer to the number of the packets that were dropped due to arriving too late.
       * @param miss a pointer to the number of the packets that can not be recovered by decoding.
       * @param recv_data a pointer to the number of the packets having been received.
       * @param recv_red a pointer to the number of the redundancy packets having been received.
       * @param grtt a pointer to the last global round trip time in millisecond.
       * @param plr a pointer to the last packet loss rate in 1/65535.
       * @param gplr a pointer to the last global packet loss rate in 1/65535.
       *
       * NOTE If the above statistics changes the value the pointer pointing to will be updated.
       * In the case of a null pointer, the change of the corresponding statistics is ignored.
       */
      virtual bool isStatsChanged(uint32_t *sent_data, uint32_t *sent_red, uint32_t *lost, uint32_t *drop, uint32_t *miss,
                          uint32_t *recv_data, uint32_t *recv_red, uint16_t *grtt, uint16_t *plr, uint16_t *gplr) const;

      /** Checks if the optimized parameters set is changed.
       *
       * Checks if the optimized parameters set is different to the current one m_opt_param_set.
       *
       * @param opt the optimized parameters to be compared with the current one.
       */
      virtual bool isOptParamSetChanged(const OptimizedParameterSet *opt) const;

      /** Set the update interval of the adaptation.
       * @param interval the update interval of the adaptation in milliseconds.
       */
      void setUpdateInterval(int interval);
      
      /** Get the update interval of the adaptation.
       * @return the update interval of the adaptation in milliseconds.
       */
      int getUpdateInterval();
      
    protected :
      media_transport_protocol::prrt_socket    *m_sock;
      
      bool              m_is_running;
      mutable pthread_mutex_t m_mutex;        ///< Mutex to protect the read/write access of the socket statistics. vs computing the parameters. vs socket setting vs state of adapting.
      mutable pthread_cond_t m_condition;    ///< Condition for notifying statistics updating thread

      int               m_update_interval;    ///< A time interval (in ms) to indicate how often to read out the statistics of PRRT socket.

      const OptimizedParameterSet  *m_opt_param_set;  ///< The optimized parameter set to be applied which is determined by the specific adaptation.
      
      media_transport_protocol::prrt_socket::SocketStatistics m_stats;
      static const int STATS_SCALE;             ///< A scale for the value of PRRT socket statistics.

      /** @name ThreadedObject methods
       */
      //@{
      /** Main thread method.
       * This method handles the computation of the parameters and applying them in a loop.
       * It runs in a separate thread owned by the Adapability.
       */
      virtual void run() = 0;
      //@}

      void updateStats();

    private :
      
      /** Internal helper method that starts the run method in its own thread.
      */
      static void* static_run(void* obj)
      {
        static_cast<Adaptability*>(obj)->run();
        return 0;
      }
      
      pthread_t m_thread;
      
      unsigned int          m_packet_length;              ///< The length of the packet (not a PRRT packet but the packet related to data rate.) in bytes.
      unsigned int          m_datarate;                   ///< The daterate of the packet stream in Mbps.
      unsigned int          m_packet_interval;            ///< The time interval of two consequent packets in milliseconds.
      
  };

}

#endif
