/* 
 * (c) 2011 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "OptimizedParameterSet.hpp"

#include <cerrno>       // for errnos


namespace Adaptation {

  OptimizedParameterSet::OptimizedParameterSet(unsigned int _k, unsigned int _r, unsigned int _n_p, unsigned int _rr,
                                               unsigned int _last_rr, unsigned int _d_fec)
  : m_k           ( _k )
  , m_r           ( _r )
  , m_n_p         ( _n_p )
  , m_rr          ( _rr )
  , m_last_rr     ( _last_rr )
  , m_v_rr        ( _rr, 1 )
  , m_delay_fec   ( _d_fec )
  {
    if (0 < m_rr && 0 < m_last_rr) {
      m_v_rr[m_rr -1] = m_last_rr;
    }
    
    m_num_opts++;
  }
  
  OptimizedParameterSet::OptimizedParameterSet(unsigned int _k, unsigned int _r, unsigned int _n_p, std::string _rr, unsigned int _d_fec)
  : m_k           ( _k )
  , m_r           ( _r )
  , m_n_p         ( _n_p )
  , m_rr          ( 0 )
  , m_last_rr     ( 0 )
  , m_v_rr        ()
  , m_delay_fec   ( _d_fec )
  {

    const char *rr = _rr.c_str();    
    unsigned int num_retrans = 0;
    
    // Convert string to a unsigned int vector and cut off all unnecessary spaces in the string if the given retransmission rounds is valid.
    unsigned int i = 0;
    while (i < _rr.size())
    {
      if (rr[i] == 32) {
        // The char is a space.
        if (num_retrans != 0) {
          m_v_rr.push_back( num_retrans );
        }
        num_retrans = 0;
      } // if (rr[i] == 32)
      else {
        // The char is not a space.
        int digit = rr[i] - '0';
        if (-1 < digit && digit < 10 ) {
          // The char is a digit.
          num_retrans = num_retrans * 10 + digit;
        } // if-else : The char is a digit.
        else {
          // This should never happen, the constructor is used only for creating the table of predefined parameters.
          std::cerr << "> OptimizedParameterSet Constructor > The given string is invalid : " << (int) rr[i] << " in " << i << "-th " << _rr << std::endl;
        }
      } // if-else : the char is not a space.
      
      i++;
    } // while (i < _rr.size())
    
    // Handle the last retransmission round.
    if (i == _rr.size() && num_retrans != 0) {
      m_v_rr.push_back( num_retrans );
    }
    
    m_rr = m_v_rr.size();
    m_last_rr = num_retrans;
    m_num_opts++;
  }


  OptimizedParameterSet::OptimizedParameterSet()
  : m_k           ( 12 )
  , m_r           ( 10 )
  , m_n_p         ( 0 )
  , m_rr          ( 0 )
  , m_last_rr     ( 0 )
  , m_v_rr        ()
  , m_delay_fec   ( 24 )
  {
    m_num_opts++;
  }

  OptimizedParameterSet::OptimizedParameterSet(const OptimizedParameterSet &opt)
  {
    *this = opt;
    m_num_opts++;
  }

  OptimizedParameterSet & OptimizedParameterSet::operator=(const OptimizedParameterSet &opt)
  {
    if (this == &opt) {
      // Identity test.
      return *this;
    }
    
    m_k = opt.getNumberOfDataPacketsPerBlock();
    m_r = opt.getNumberOfRedundancyPacketsPerBlock();
    m_n_p = opt.getNumberOfPresentRedundancyPacketsPerBlock();
    m_rr = opt.getNumberOfRetransmissionRounds();
    m_last_rr = opt.getNumberOfRetransmissionsInLastRound();
    m_delay_fec = opt.getDelayFEC();
    m_v_rr = opt.getRetransmissionRoundsVector();

    return *this;
  }

  OptimizedParameterSet::~OptimizedParameterSet()
  {
    m_num_opts--;  
  }
  
  bool OptimizedParameterSet::operator==(const OptimizedParameterSet &opt) const
  {
    if (opt.getNumberOfDataPacketsPerBlock() != m_k ||
        opt.getNumberOfRedundancyPacketsPerBlock() != m_r ||
        opt.getNumberOfPresentRedundancyPacketsPerBlock() != m_n_p ||
        opt.getNumberOfRetransmissionsInLastRound() != m_last_rr ||
        opt.getDelayFEC() != m_delay_fec ||
        opt.getNumberOfRetransmissionRounds() != m_rr) {
      return false;
    }
    
    for (unsigned int i=0; i<m_v_rr.size(); i++) {
      if (opt.getRetransmissionRoundsVector()[i] != m_v_rr[i]) {
        return false;
      }
    }
    
    return true;
  }

  std::string OptimizedParameterSet::toString() const
  {
    std::ostringstream s;
    s << "k=" << m_k << "; r=" << m_r << "; n_p=" << m_n_p << "; d_fec=" << m_delay_fec << "; v_rr=[";

    int size = m_v_rr.size();
    for (int i=0; i<size-1; i++) {
      s << m_v_rr[i] << " ";
    }

    if (0 < size) {
      s << m_v_rr[ size - 1 ];
    }

    s << "]";
    return s.str();
  }
  
  
  int OptimizedParameterSet::m_num_opts = 0;

}
