/*
 * (c) 2010 Saarland University, Telecommunications Lab
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 +
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 +
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 +
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 +
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 +
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device,
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */
#ifndef MAINCLASSPRRTRECV_H
#define	MAINCLASSPRRTRECV_H

#include <iostream>
#include <unistd.h>
#include <signal.h>
#include <cstdlib>
#include <string.h>
#include <sstream>
#include <sys/time.h>
#include <prrt/prrt_sockets.h>
#include <ReceiveSettings.h>
#include <CLAOptionWithParameter.h>
#include <SocketInitialiser.h>
#include <errno.h>

using namespace std;
using namespace CommandLineArgumentLib;
using namespace media_transport_protocol;

class mainClassPrrtRecv
{
    typedef void (*voidFunctionPointer)(int);
public:

    mainClassPrrtRecv()
    {
        forever = true;
    }

    int main(int argc, char** argv)
    {
        ReceiveSettings settings;
        CommandLineArgumentParser parser;

        InitRecvParser(parser, settings);

        parser.AddHelpFlag("-h");

        char* program = argv[0];

        try
        {
            parser.ParseArguments(argc - 1, &argv[1]);

            if (parser.IsHelpFlagSet())
            {
                show_usage(program, parser.GetOptionHelpMessage());
                exit(1);
            }

            prrt_socket prrtsocket;
            SocketInitialiser::InitReceiverSocket(settings, prrtsocket);

            const int payload_size = 1500;
            char payload[payload_size];

            int length = 1;

            while (length && forever)
            {
              try {
                length = prrtsocket.recv(payload, payload_size, 0);
	          	}
	          	catch (module_error &me)
	          	{
                if (me.GetErrorNumber() == -1 || me.GetErrorNumber() == -ETIMEDOUT) {
	          		  length = 0;
	          		  cerr << "WARNING: "<< me.what() << endl;
                } else {
                  cerr << "Could not catch receive error (no. " << me.GetErrorNumber() << ")" << endl;
                  throw;
                }
	          	}
              int ret = write(1, payload, length);  // returned value is used to solve warning.
              if (ret != length) {
                // solving warning.
              }
            }

        }
        catch (prrt_error &me)
        {
            cerr << "ERROR: " << me.what() << endl;
        }
        catch (CLAException error)
        {
            std::cerr << "Error when parsing options: " << error.what() << std::endl;
            show_usage(program, parser.GetOptionHelpMessage());
            exit(1);
        }

        return 0;
    }

    void shutdown() {
        forever = false;
    }

protected:

    void InitRecvParser(CommandLineArgumentParser &parser, ReceiveSettings &settings)
    {
        senderAddressValue.SetSetting(settings.GetSenderAddressSetting());
        parser.AddArgumentType(&senderAddressValue);

        stringstream helpMessage;

        helpMessage.str("");
        helpMessage << "\n   Receiver options:\n";
        helpMessage << "\t -rp <listen port>            \t\t[default: " << settings.GetPort() << "]\n";
        portOption.SetIdentifier("-rp");
        portOption.SetHelpMessage(helpMessage.str());
        portOption.SetSetting(settings.GetPortSetting());
        parser.AddArgumentType(&portOption);

        unicastOption.SetIdentifier("-ruc");
        unicastOption.SetHelpMessage("\t -ruc <unicast Address>		\tThe unicast address to receive on\n");
        unicastOption.SetSetting(settings.GetUnicastAddressSetting());
        parser.AddArgumentType(&unicastOption);

        multicastOption.SetIdentifier("-rmc");
        multicastOption.SetHelpMessage("\t -rmc <multicast Address> 	\tThe multicast address to receive on\n");
        multicastOption.SetSetting(settings.GetMulticastAddressSetting());
        parser.AddArgumentType(&multicastOption);

        helpMessage.str("");
        helpMessage << "\t -to <0|1>            \t\t\tUse timeout when receiving packets\n";
        timeoutOption.SetIdentifier("-to");
        timeoutOption.SetHelpMessage(helpMessage.str());
        timeoutOption.SetSetting(settings.GetTimeoutSetting());
        parser.AddArgumentType(&timeoutOption);

        helpMessage.str("");
        helpMessage << "\t -bw <0|1|2>            \t\tEnable bandwidth estimation [default: " << settings.GetBWESTMode() << "]\n";
        helpMessage << "\t\t\t\t\t\t0: no bandwidth estimation; 1: via TCP Friendly Rate Control; 2: based on the RTT.\n";
        bwEstOption.SetIdentifier("-bw");
        bwEstOption.SetHelpMessage(helpMessage.str());
        bwEstOption.SetSetting(settings.GetBWESTSetting());
        parser.AddArgumentType(&bwEstOption);

        helpMessage.str("");
        helpMessage << "\t -i <updateInterval>    \t\tThe update interval of bandwidth estimation in milliseconds [default: " << settings.GetUpdateInterval() << "]\n";
        helpMessage << "\t\t\t\t\t\tThe value doesn't effect, if the option of bandwidth estimation (-bw) is disabled;\n";
        updateIntervalOption.SetIdentifier("-i");
        updateIntervalOption.SetHelpMessage(helpMessage.str());
        updateIntervalOption.SetSetting(settings.GetUpdateIntervalSetting());
        parser.AddArgumentType(&updateIntervalOption);

        helpMessage.str("");
        helpMessage << "\t -fs <0|1>            \t\t\tEnable feedback suppression [default: " << settings.GetIsFeedbackSuppressionActive() << "]\n";
        fbSuppOption.SetIdentifier("-fs");
        fbSuppOption.SetHelpMessage(helpMessage.str());
        fbSuppOption.SetSetting(settings.GetFeedbackSuppressionSetting());
        parser.AddArgumentType(&fbSuppOption);

        helpMessage.str("");
        helpMessage << "\t -d <maxSentDelay>    \t\t\tThe maximum delay of sending feedback in milliseconds [default: " << settings.GetMaxFeedbackSentDelay() << "]\n";
        helpMessage << "\t\t\t\t\t\tThe value doesn't effect, if the option of feedback suppression (-fs) is disabled;\n";
        maxFbSentDelayOption.SetIdentifier("-d");
        maxFbSentDelayOption.SetHelpMessage(helpMessage.str());
        maxFbSentDelayOption.SetSetting(settings.GetMaxFeedbackSentDelaySetting());
        parser.AddArgumentType(&maxFbSentDelayOption);

        helpMessage.str("");
        helpMessage << "\t -dcc <delayCCParameters>        \t[default: \"" << settings.GetDelayCCParameters() << "\"]\n";
        helpMessage << "\t\t\t\t\t\tSample explanation : alpha (3.0), beta (5.0), gamma (0.5)\n";
        delayCCOption.SetIdentifier("-dcc");
        delayCCOption.SetHelpMessage(helpMessage.str());
        delayCCOption.SetSetting(settings.GetDelayCCSetting());
        parser.AddArgumentType(&delayCCOption);
    }

private:

    void show_usage(char* program, std::string optionMessage)
    {
        std::stringstream ss;
        ss << "\033[31mUsage: " << program << " [options] <sender>\033[0m\n\n";
        ss << "List of available options:\n" << optionMessage;

        cerr << ss.str();
    }

    bool forever;

    CLARequiredValue senderAddressValue;
    CLAOptionWithParameter portOption;
    CLAOptionWithParameter unicastOption;
    CLAOptionWithParameter multicastOption;
    CLAOptionWithParameter timeoutOption;
    CLAOptionWithParameter bwEstOption;
    CLAOptionWithParameter updateIntervalOption;
    CLAOptionWithParameter fbSuppOption;
    CLAOptionWithParameter maxFbSentDelayOption;
    CLAOptionWithParameter delayCCOption;
};

#endif	/* MAINCLASSPRRTRECV_H */

