/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#ifndef PRRT_OPTION_H
#define PRRT_OPTION_H

#include "prrt_option_base.h"
#include <string>

namespace media_transport_protocol {

// These are the different options that can be set for a socket of the prrt protocol.

/// This option controls the sending of feedback.
/// Through this option one can enable/disable the sending of feedback through a receiver.
class option_FEEDBACK : public prrt_option {
	public:
		enum option_value {
			ON,	///< Enable the sending of feedback.
			OFF	///< Disable the sending of feedback.
		};

		option_FEEDBACK(enum option_value val);
};


/// This option controls the sending of replacements packets.
/// Through this option one can enable/disable the sending of replacements through a sender.
class option_REPLACE : public prrt_option {
	public:
		enum option_value {
			ON,	///< Enable the sending of replacements.
			OFF	///< Disable the sending of replacements.
		};

		option_REPLACE(enum option_value val);
};


/// This option controls the coding activity of a socket.
/// Through this option one can enable/disable the coding of blocks.
/// Disabling this option will also disable the sending of replacements.
class option_CODING : public prrt_option {
	public:
		enum option_value {
			ON,	///< Enable the coding of blocks.
			OFF	///< Disable the coding of blocks.
		};

		option_CODING(enum option_value val);
};


/// This option states if the receiving should be stopped after a timeout occurs.
/// Disabling this option will make the socket trying to repeat the receiving
/// of packets after a timeout occurs. The socket will try to receive packets
/// until it is shut down.
class option_TIMEOUT : public prrt_option {
	public:
		enum option_value {
			ContinueAfterTimeout,	///< Socket continues receiving after a timeout.
			StopAfterTimeout	///< After a timeout the socket stops receiving.
		};

		option_TIMEOUT(enum option_value val);
};


/// This option is used to set or get the maximum socket receive buffer in bytes.
/// The kernel doubles this value (to allow space for bookkeeping overhead) when it is set using setsockopt(),
/// and this doubled value is returned by getsockopt().
/// The default value is set by the rmem_default sysctl and the maximum allowed value is set by the rmem_max sysctl.
/// The minimum (doubled) value for this option is 256.
/// There are three sockets used in PRRT transmission: PRRT_RECV, KERNEL_RECV and USER_RECV.
/// The buffer size 0 means that we don't explictly set it, i.e.: the socket just use a default/current value as it's created.
/// NOTE & HACK Since we don't support the querying of the option e.g.: getsockopt(), calling setsockopt() with option value 0 can get the current buffer size of the corresponding socket.
class option_RCVBUF : public prrt_option {
    public:
        enum Type {
            PRRT_RECV,  /// The option value is specified to the buffer size of the receive socket used in the transmission between a pair of PRRT sockets.
            KERNEL_RECV,    /// The option value is specified to the buffer size of the receive socket used in the kernel space of the kernel-user communication.
            USER_RECV       /// The option value is specified to the buffer size of the receive socket used in the user space of the kernel-user communication.
        };

        option_RCVBUF(enum Type type, int size = 0);
};


/// This option is used to set or get the adaptation mode and the update interval in milliseconds. (Only available for sender.)
/// There is only one mode at the moment -- simple analysis based on i.i.d.
/// The update interval has a default value 200 ms if we don't explictly set it.
class option_ADAPTATION : public prrt_option {
    public:
        enum Type {
            SIMPLE  /// The option value is specified to the buffer size of the receive socket used in the transmission between a pair of PRRT sockets.
        };

        option_ADAPTATION(enum Type type, int interval = 200);
};


/// This option is used to set or get the bandwidth estimation mode and the update interval in milliseconds. (Only available for the receiver.)
/// The bandwidth estimation has two modes : based on the TCP Friendly Rate Control or the RTT.
/// The update interval has a default value 200 ms if we don't explictly set it. This means that the bandwidth estimation is disabled.
class option_BWEST : public prrt_option {
    public:
        enum Type {
            TFRC = 1,     /// The option value specifies that the bandwidth is estimated via the TCP Friendly Rate Control.
            RTT = 2       /// The option value specifies that the bandwidth is estimated through calculating the RTT.
        };

        option_BWEST(enum Type type, int interval = 200);
};


/// This option is used to control the network traffic in order to optimize or guarantee performance, improve latency, and/or increase usable bandwidth for some kinds of packets by delaying other kinds of packets that meet certain criteria.
/// The value of the option is the data rate of the bandwidth in Mbps. Default value is 10Mbps.
class option_TRAFFICSHAPING : public prrt_option {
  public:
    option_TRAFFICSHAPING(const int datarate = 10);

};


/// This option is used to enable/disable the feedback suppression and the maximum feedback sent delay in milliseconds. (Only at the receiver.)
/// The maximum feedback sent delay has a default value 50 ms if we don't explictly set it.
class option_FBSUPP : public prrt_option {
    public:
        enum Option {
            ON,  /// The option value is specified to enable the feedback suppression.
            OFF  /// The option value is specified to disable the feedback suppression.
        };

        option_FBSUPP(enum Option opt, int delay = 50);
};



}

#endif /*PRRT_OPTION_H*/
