/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#ifndef PRRT_SOCKETS_H
#define PRRT_SOCKETS_H

#include <iostream>
#include <sys/time.h>
#include <pthread.h>

#include "proc_file.h"
#include "prrt_params.h"
#include "prrt_option.h"
#include "socket_list.h"
#include "prrt.h"
#include "tfrc.h"
#include "delay_cc.h"
#include "analysis.h"

namespace media_transport_protocol {

class prrt_relay;

/**
 * This should be the same type as in the module for the socket descriptor.
 */
class prrt_socket {
private:

  static const int DEFAULT_PAYLOAD_SIZE;  // Default payload size of PRRT packet. (188*7 bytes)

  std::string sockno;	///< The number of the socket as string.
  bool open;	///< Is this socket opened in the kernel?

  std::string boundSockAddr;
  std::string boundPeerAddr;
  unsigned short boundPort;

  std::string connectedSockAddr;
  std::string connectedHostAddr;
  unsigned short connectedPort;

  /// The following max buffer size is double to the size which is set using setsockopt(), because the kernel allows space for bookkeeping overhead.
  /// NOTE The kernel limits the max UDP receive buffer size. To correctly set the value as expecting, you need to consider the limit value given by net.core.rmem_max (default is 131071, max is 8388608 under linux).
  int rcv_buf_size;               ///< The max buffer size of the receive socket.
  int kui_kernel_sock_buf_size;   ///< The buffer size of the kernel-user-communication on the side of kernel-space.
  int kui_user_sock_buf_size;     ///< The buffer size of the kernel-user-communication on the side of user-space.

  int kui_port_read;
  int kui_port_write;
  int kui_sock_read; /*Netlink socket to kernel module*/
  int kui_sock_write; /*Netlink socket to kernel module*/

  int m_adaptation_mode;    ///< A flag indicates which mode of adaptation is used in PRRT.
  struct timeval interval;  ///< Update interval for bandwidth estimation.
  int m_update_interval;    ///< Update interval for adaptation.

  /**
   * AHEC Parameters including : D_target, D_FEC, k and N_p.
   */
  int m_delay_target;		///< The target delay.
  int m_delay_fec;		///< The delay of FEC.
  int m_delay_fb;       ///< The delay of receivers sending feedback.
  unsigned m_k;			///< The number of data packets per block.
  std::vector<unsigned> m_n_p;	///< The multiplier of parity packets per cycle (in vector form).

  /**
  * This list contains pointers to all sockets that are created.
  */
  static socket_list SocketList;

  /**
   * Bandwidth estimation options
   */
  int m_bw_est_mode;        ///< The mode for bandwidth estimation: 0 : disabled; 1 : estimated by a token bucket; 2 : estimated by the RTT.
  prrt_tfrc tfrc;           ///< A bandwidth estimation based on the TCP Friendly Rate Control.
  prrt_delay_cc m_delay_cc;   ///< A bandwidth estimation based on the RTT.

  /**
   * Help variables for analysis.
   */
  //std::string m_filename_analysis;   ///< The file name of the analysis file storing the goodput per second, is used to plot the goodput.
  //FILE *m_file_analysis;             ///< The pointer of file object for the analysis file.
  int m_est_bw;                     ///< The current estimated bandwidth. (in kbits/s)

  /**
  * The prrtAnalysis for UpdateCodingParameters.
  */
  /**
    * A separate thread for online computation.
    */
  pthread_t m_thread;
  bool              m_is_running;         ///< Flag indicates if the adaptation is running or not.
  mutable pthread_mutex_t m_mutex;        ///< Mutex for internal thread condition.
  mutable pthread_cond_t m_condition;     ///< Condition for waiting for the given time.

  prrt_analysis prrtAnalysis;

  timeval tv_last_param_update;
  unsigned int lastGRTT;

  int create_kui_socket()
  throw(prrt_error);

  double m_recv_rate;  ///< The acknowledged receiver data rate (in packets per second), used to limit the maximum estimated bandwidth. It should lighten the overshot at the begining of the bandwidth estimation and avoid the increase to infinity, if the sending rate is limited to a maximal value.
  double m_avg_recv_rate; ///< The average of acknowledged receiving date rate.
  void update_bw_est();
  void update_coding_params();

  /** Internal helper method that starts the run method in its own thread.
  */
  static void* static_update_coding_params(void* obj)
  {
    static_cast<prrt_socket*>(obj)->update_coding_params();
    return 0;
  }

  /** A mutex protects the following two variables during the computation of actual stream rate.
   */
  int m_send_payload_size;    ///< The average size of the packet payload for sending, computed by exponential smoothing. default 1316=188*7
  double m_streamrate;        ///< The data rate (kbits/s) of the actual streaming which can be used in the channel.
  mutable pthread_mutex_t m_mutex_streamrate;        ///< Mutex for the payload size and the stream rate.

  int m_recv_payload_size;    ///< The payload size of received packet, used to compute the receiving rate. default 1316=188*7.
                              ///< TODO this can be removed, replaced with m_send_payload_size instead, if send socket and recv socket can not exist at the same time.

  friend class prrt_relay;

public:
	class SocketStatistics {
    public:
      unsigned long deliveredData;
      unsigned long deliveredRedundancy;
      unsigned long lostPackets;
      unsigned long droppedPackets;
      unsigned long missedPackets;
      unsigned long receivedData;
      unsigned long receivedRedundancy;
      unsigned int lastGRTT;
      unsigned int lastFTT;
      unsigned int lastGFTT;
      unsigned int queueDelay;
      unsigned int lastPLR;
      unsigned int lastGPLR;
      unsigned int numGaps;
      unsigned int numBursts;
      unsigned int gapLength;
      unsigned int burstLength;
      unsigned int numGGaps;
      unsigned int numGBursts;
      unsigned int gapGLength;
      unsigned int burstGLength;
      unsigned int packetInterval;
      unsigned int bwEst;
      unsigned int buffb;
      std::string lossPattern;
  };

  /**
  * Constructor.
  */
  prrt_socket()
  throw(module_error, list_error, prrt_error);

  /**
  * Destructor.
  */
  ~prrt_socket();

  /**
  * Binds and prrt socket to an address and a port.
  * For multicast sockaddr is the address of the host and peeraddr is the multicast
  * address.
  * \param sockaddr address for the socket as a string (e.g. "127.0.0.1")
  * \param sockport the port the socket should be bound to
  * \param peeraddr the address the socket should send feedback to
    *
    * NOTE \param peeraddr is not just the address any more, to which the socket should send feedback.
    * Because the new bind()/connect() have been added in the PRRT module and the kui_control has been
    * changed. This parameter is also an option to specify which bind() (original/new) is used and
    * whether it binds the socket as a receiver or as a sender, respectively based on "as_recv" or "as_send".
    *
  */
  void bind(std::string &sockaddr, unsigned short sockport, std::string &peeraddr)
  throw(module_error, prrt_error);

  /**
  * Let the socket connect to a peer.
  * \param sockaddr address for the peer as a string (e.g. "127.0.0.1")
  * \param sockport the port of the peer.
  * \param hostaddr the address for the incoming feedback.
    *
    * NOTE \param hostaddr is not just the address for the incoming feedback any more.
    * Because the new bind()/connect() have been added in the PRRT module and the kui_control has been
    * changed. This parameter is also an option to specify which connect() (original/new) is used and
    * whether it connects the socket to a receiver or to a sender (feedback),
    * respectively based on "to_receiver" or "to_feedback".
    *
  */
  void connect(std::string &sockaddr, unsigned short sockport, std::string &hostaddr)
  throw(module_error, prrt_error);

  /**
  * Close a socket.
  */
  void close()
  throw();

  /**
  * Send data over a socket.
  * \param data the data that should be written to the socket.
  * \param len the number of bytes contained in data.
  * \param flags not implemented yet.
  * \return number of written bytes on success, -1 if an error occurs.
  */
  int send(char *data, int len, int flags)
  throw(prrt_error);

  /**
  * Receive data from a socket.
  * \param data the the buffer for the data that is read.
  * \param len the number of bytes data can hold.
  * \param flags not implemented yet.
  * \return number of bytes read on success, -1 if an error occurs.
  */
  int recv(char *data, int len, int flags)
  throw(module_error, prrt_error);

  /**
  * Set an option for a socket.
  * \param option option that should be applied.
  */
  void setsockopt(const prrt_option &option)
  throw(prrt_error, module_error);

  /**
  * Returns the option value for a socket with the given options.
  * \param option option whose value is to be returned.
  * \return the option value.
  * NOTE At the moment it just supports max buffer size of the sockets. Other options is not implemented yet.
  */
  int getsockopt(const prrt_option &option)
  throw(prrt_error, module_error);

  /**
  * Set a parameter for the prrt socket.
  * \param param the parameter that should be set.
  * \param value the value the parameter should be assigned.
  * \return ?
  */
  /*template<class T> */void set_param(const prrt_param/*<T>*/ &param)
  throw(prrt_error);

  /**
  * Retrieve a statistic over packet sending/receiving for the socket.
  */
  void getStats(SocketStatistics *sockStats)
  throw(prrt_error);

  /**
  * Starts / Stops the adaptation.
  */
  void startAdaptation();
  void stopAdaptation();

  /**
   * This function evaluates the stream rate based on the AHEC parameters and bandwidth estimation from the receiver.
   */
  double evaluateStreamrate();

  /**
   * The following methods concerns the setting of parameters in delay-based congestion control equation.
   */
  void setDelayCCAlpha(double alpha);
  double getDelayCCAlpha();
  void setDelayCCBeta(double beta);
  double getDelayCCBeta();
  void setDelayCCGamma(double gamma);
  double getDelayCCGamma();

/**
* Get a parameter from the prrt socket.
* If the parameter is successfully retrieved it's value is saved in param.
* \param param the parameter that should be retrieved.
*/
//		/*template<class T> */void get_param(prrt_param/*<T>*/ &param)
//		throw(prrt_error);

  /**
  * Equality Operator for sockets.
  */
  bool operator==(prrt_socket& s)
  throw(prrt_error);

  /**
  * This should close all open sockets.
  */
  static void delete_all_sockets();


  unsigned short getConnectedPort() const {
      return connectedPort;
  }
  std::string getConnectedHostAddr() const {
      return connectedHostAddr;
  }
  std::string getConnectedSockAddr() const {
      return connectedSockAddr;
  }
  unsigned short getBoundPort() const {
      return boundPort;
  }
  std::string getBoundPeerAddr() const {
      return boundPeerAddr;
  }
  std::string getBoundSockAddr() const {
      return boundSockAddr;
  }
  
  bool isOpen() const {
      return open;
  }
  
  int getSockfd() const {
      return kui_sock_read;
  }
  
  std::string getSockno() const {
      return sockno;
  }

  /**
   * Retrieve the number of data packets per block.
   */
  int getBlockSize() const { return m_k; }

  /**
   * Retrieve the target delay.
   */
  int getTargetDelay() const { return m_delay_target; }

  /**
   * Retrieve the delay of forward error correction.
   */
  int getDelayFEC() const { return m_delay_fec; }

  /**
   * Set and get the payload size of sending packet.
   * This two function is protected by a mutex, due to a separate thread.
   */
  int getSendPayloadSize();
  void setSendPayloadSize(int size);
 
 /**
  * Set and get the actual maximum availabel data rate for the applications (e.g. streaming).
  * This two function is protected by a mutex, due to a separate thread.
  */
 double getStreamrate();
 void setStreamrate( double rate );

  /**
   * Set and get the actual receiving rate (goodput) in bps (in average). It can be used for flow control in the application.
   * TODO This two function should be protected by a mutex.
   */
  double getReceivingRate();
  void setReceivingRate( double rate ); /// TODO should be private.
 
  /**
   * Set and get the buffer information of client application. It can be used for flow control or scheduling packet sending.
   * This buffer information will be sent to the server via PRRT feedback packet.
   * The unit can be various, depending on applications.
   */
  unsigned int getClientBufferInfo();
  void setClientBufferInfo( unsigned int buf_info );

  /**
   * Retrieve the multiplier of parity packets in each cycle (in vector form).
   */
  std::vector<unsigned> getParityPacketsPerCycle() const { return m_n_p; }
  
  std::string ToString() const {
      std::stringstream ss;

      ss << "Socket no: " << this->sockno << std::endl;
      ss << "fd: " << this->kui_sock_read << std::endl;
      ss << "Is open: " << this->open << std::endl;
      ss << "bound socket address: " << this->boundSockAddr << std::endl;
      ss << "bound peer address: " << this->boundPeerAddr << std::endl;
      ss << "bound port: " << this->boundPort << std::endl;
      ss << "connected socket address: " << this->connectedSockAddr << std::endl;
      ss << "connected host address: " << this->connectedHostAddr << std::endl;
      ss << "connected port: " << this->connectedPort << std::endl;
      ss << "buffer size of PRRT sock:" << this->rcv_buf_size << std::endl;
      ss << "buffer size of user sock:" << this->kui_user_sock_buf_size << std::endl;
      ss << "buffer size of kernel sock:" << this->kui_kernel_sock_buf_size << std::endl;
      //ss << "Adaptation mode:" << m_adaptation_mode << " with updating interval (ms):" << interval.tv_sec*1000 + interval.tv_usec/1000 << std::endl;
      ss << "Adaptation mode:" << m_adaptation_mode << " with updating interval (ms):" << m_update_interval << std::endl;
      
      return ss.str();
  };
};

}

#endif /*PRRT_SOCKETS_H*/

