/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "feedback_analyzer.h"
#include "parameters/param_code.h"

#include <linux/jiffies.h>

#define DEBUG_NAME "FB_ANALYZER"

static struct kmem_cache *fbana_fb_cache = 0;
static struct kmem_cache *fbana_recv_cache = 0;
static struct kmem_cache *fbana_repl_cache = 0;
static struct semaphore fbana_fb_cache_lock;
static struct semaphore fbana_repl_cache_lock;
//int fbentries = 0;
//int recventries = 0;
//int replentries = 0;

int prrt_fbana_caches_init(void)
{
    init_MUTEX(&fbana_fb_cache_lock);
    init_MUTEX(&fbana_repl_cache_lock);

    fbana_fb_cache = kmem_cache_create("fbana_fb_cache",
            sizeof (struct fbana_feedback_entry),
            0, 0,
            NULL);

    if (!fbana_fb_cache)
    {
        PERROR("%s: Could not create fbana_fb_cache\n", DEBUG_NAME);
        goto error;
    }

    fbana_recv_cache = kmem_cache_create("fbana_recv_cache",
            sizeof (struct fbana_receiver_entry),
            0, 0,
            NULL);

    if (!fbana_recv_cache)
    {
        PERROR("%s: Could not create fbana_recv_cache\n", DEBUG_NAME);
        goto error;
    }

    fbana_repl_cache = kmem_cache_create("fbana_repl_cache",
            sizeof (struct fbana_replacement_entry),
            0, 0,
            NULL);

    if (!fbana_repl_cache)
    {
        PERROR("%s: Could not create fbana_repl_cache\n", DEBUG_NAME);
        goto error;
    }

    return 0;

error:
    prrt_fbana_caches_uninit();
    return -1;
}

void prrt_fbana_caches_uninit(void)
{
    kmem_cache_destroy(fbana_fb_cache);
    kmem_cache_destroy(fbana_recv_cache);
    kmem_cache_destroy(fbana_repl_cache);

    fbana_fb_cache = 0;
    fbana_recv_cache = 0;
    fbana_repl_cache = 0;
}

int AllocFeedbackEntry(struct fbana_feedback_entry **fb_entry)
{
    PDEBUG(MSG_ALLOC, "%s: Allocating feedback entry\n", DEBUG_NAME);
    down(&fbana_fb_cache_lock);
    *fb_entry = kmem_cache_alloc(fbana_fb_cache, GFP_KERNEL);
    //fbentries++;
    //PDEBUG(MSG_FBANADEB, "%s: Allocating feedback entry (count %d)\n", DEBUG_NAME, fbentries);
    up(&fbana_fb_cache_lock);

    if (!(*fb_entry))
    {
        PERROR("%s: Could not allocate memory for feedback entry\n", DEBUG_NAME);
        return -1;
    }

    return 0;
}

void FreeFeedbackEntry(struct fbana_feedback_entry **fb_entry)
{
    if (!(*fb_entry))
        return;

    down(&fbana_fb_cache_lock);
    kmem_cache_free(fbana_fb_cache, *fb_entry);
    //fbentries--;
    up(&fbana_fb_cache_lock);
    *fb_entry = 0;
    PDEBUG(MSG_ALLOC, "%s: Feedback entry deleted\n", DEBUG_NAME);
}

int AllocReceiverEntry(struct fbana_receiver_entry **recv_entry)
{
    PDEBUG(MSG_ALLOC, "%s: Allocating receiver entry\n", DEBUG_NAME);
    down(&fbana_fb_cache_lock);
    *recv_entry = kmem_cache_alloc(fbana_recv_cache, GFP_KERNEL);
    //recventries++;
    //PDEBUG(MSG_FBANADEB, "%s: Allocating receiver entry (count %d)\n", DEBUG_NAME, recventries);
    up(&fbana_fb_cache_lock);

    if (!(*recv_entry))
    {
        PERROR("%s: Could not allocate memory for receiver entry\n", DEBUG_NAME);
        return -1;
    }

    return 0;
}

void FreeReceiverEntry(struct fbana_receiver_entry **recv_entry)
{
    if (!(*recv_entry))
        return;

    down(&fbana_fb_cache_lock);
    kmem_cache_free(fbana_recv_cache, *recv_entry);
    //recventries--;
    up(&fbana_fb_cache_lock);
    *recv_entry = 0;
    PDEBUG(MSG_ALLOC, "%s: Receiver entry deleted\n", DEBUG_NAME);
}

int AllocReplacementEntry(struct fbana_replacement_entry **repl_entry)
{
    PDEBUG(MSG_ALLOC, "%s: Allocating replacement entry\n", DEBUG_NAME);
    down(&fbana_repl_cache_lock);
    *repl_entry = kmem_cache_alloc(fbana_repl_cache, GFP_KERNEL);
    //replentries++;
    //PDEBUG(MSG_FBANADEB, "%s: Allocating replacement entry (count %d)\n", DEBUG_NAME, replentries);
    up(&fbana_repl_cache_lock);

    if (!(*repl_entry))
    {
        PERROR("%s: Could not allocate memory for replacement entry\n", DEBUG_NAME);
        return -1;
    }

    (*repl_entry)->max_n_cc = 0;

    return 0;
}

void FreeReplacementEntry(struct fbana_replacement_entry **repl_entry)
{
    if (!(*repl_entry))
        return;

    down(&fbana_repl_cache_lock);
    kmem_cache_free(fbana_repl_cache, *repl_entry);
    //replentries--;
    up(&fbana_repl_cache_lock);
    *repl_entry = 0;
    PDEBUG(MSG_ALLOC, "%s: Replacement entry deleted\n", DEBUG_NAME);
}

int prrt_fbana_init(struct prrt_fb_analyzer *fbana, struct prrt_time_state *time_state)
{
    INIT_LIST_HEAD(&fbana->fb_list);
    INIT_LIST_HEAD(&fbana->repl_list);
    init_MUTEX(&fbana->lock);
    init_MUTEX(&fbana->repl_lock);
    fbana->count = 0;
    fbana->time_state = time_state;

    return 0;
}

int prrt_fbana_reinit(struct prrt_fb_analyzer *fbana)
{
    prrt_fbana_free(fbana);
    fbana->count = 0;

    return 0;
}

static void free_feedback_entry(struct fbana_feedback_entry **fbe);
static void FreeReplacements(struct prrt_fb_analyzer *fbana);

void prrt_fbana_free(struct prrt_fb_analyzer *fbana)
{
    prrt_fbana_free_feedbacks(fbana);

    FreeReplacements(fbana);
}

void prrt_fbana_free_feedbacks(struct prrt_fb_analyzer *fbana)
{
    struct list_head *iter = 0;
    struct list_head *next = 0;
    struct fbana_feedback_entry *fbe = 0;

    //PDEBUG(MSG_FBANADEB, "%s: Freeing feedbacks\n", DEBUG_NAME);

    down(&fbana->lock);

    /*delete all feedbacks in the list*/
    list_for_each_safe(iter, next, &fbana->fb_list)
    {
        fbe = container_of(iter, struct fbana_feedback_entry, list);

        list_del(iter);
        free_feedback_entry(&fbe);

        fbana->count--;
    }

    up(&fbana->lock);
}

static void free_feedback_entry(struct fbana_feedback_entry **fbe)
{
    //struct list_head *iter = 0;
    //struct list_head *next = 0;
    //struct fbana_receiver_entry *re = 0;

    //PDEBUG(MSG_FBANADEB, "%s: Freeing feedback for packet %d\n", DEBUG_NAME, (*fbe)->packno);

    /*delete the receiver list in the entry first*/
    //list_for_each_safe(iter, next, &(*fbe)->recvs)
    //{
    //    re = container_of(iter, struct fbana_receiver_entry, list);

    //    list_del(iter);
    //    //PDEBUG(MSG_FBANADEB, "%s: Freeing receiver no %d\n", DEBUG_NAME, re->recv_ip);
    //    FreeReceiverEntry(&re);
    //}

    //PDEBUG(MSG_FBANADEB, "%s: Deleting feedback\n", DEBUG_NAME);
    FreeFeedbackEntry(fbe);
    //PDEBUG(MSG_FBANADEB, "%s: Deletion done\n", DEBUG_NAME);
}

static void FreeReplacements(struct prrt_fb_analyzer *fbana)
{
    struct list_head *iter = 0;
    struct list_head *next = 0;
    struct fbana_replacement_entry *rple = 0;

    down(&fbana->repl_lock);

    list_for_each_safe(iter, next, &fbana->repl_list)
    {
        rple = container_of(iter, struct fbana_replacement_entry, list);

        list_del(iter);
        prrt_fbana_free_replacement_entry(&rple);
    }

    up(&fbana->repl_lock);
}

void prrt_fbana_free_replacement_entry(struct fbana_replacement_entry **rple)
{
    FreeReplacementEntry(rple);
}

static int ResetInternalEntries(struct prrt_fb_analyzer *fbana);
//static int SearchFeedback(struct prrt_fb_analyzer *fbana, u16 miss);
//static int InsertAndUpdateReceiver(struct prrt_fb_analyzer *fbana, u32 receiver, int retr_round);
static int InsertAndUpdateFeedback(struct prrt_fb_analyzer *fbana, u16 miss, u32 receiver,
        int retr_round);

int prrt_fbana_insert_feedback(struct prrt_fb_analyzer *fbana, u16 miss, u32 receiver, int retr_round)
{
    int err = 0;
    u32 time = 0;
    time = prrt_get_prrttime(fbana->time_state);

    down(&fbana->lock);

    ResetInternalEntries(fbana);

    
    PDEBUG(MSG_FBANADEB, "%s: Trying to insert feedback for packet %d and receiver %u in round %d (at %u)\n",
    	DEBUG_NAME, miss, receiver, retr_round, time);
    //if(SearchFeedback(fbana, miss)) {
    //	/*we found an existing feedback, search for receiver in list of feedback*/
    //	err = InsertAndUpdateReceiver(fbana, receiver, retr_round);
    //} else {
    err = InsertAndUpdateFeedback(fbana, miss, receiver, retr_round);
    //}

    up(&fbana->lock);
    //PDEBUG(MSG_FBANADEB, "%s: Inserting feedback done.\n", DEBUG_NAME);

    return err;
}

static int ResetInternalEntries(struct prrt_fb_analyzer *fbana)
{
    fbana->fbe = 0;
    fbana->re = 0;
    fbana->rple = 0;

    return 0;
}

static int InsertFeedback(struct prrt_fb_analyzer *fbana, u16 miss, u8 retr_round);

static int InsertAndUpdateFeedback(struct prrt_fb_analyzer *fbana, u16 miss, u32 receiver,
        int retr_round)
{
    /*should be locked here*/
    int err = 0;

    /*did not find any entries for this miss, enter new entry*/
    AllocFeedbackEntry(&fbana->fbe);
    //AllocReceiverEntry(&fbana->re);
    if (!(fbana->fbe) /* || !(fbana->re)*/)
    {
        err = -ENOMEM;
        FreeFeedbackEntry(&fbana->fbe);
    //    FreeReceiverEntry(&fbana->re);
        return err;
    }

    PDEBUG(MSG_FBANADEB, "%s: Inserting new feedback\n", DEBUG_NAME);
    err = InsertFeedback(fbana, miss, (u8)retr_round);
    if (err < 0)
        return err;

    //PDEBUG(MSG_FBANADEB, "%s: Inserting new receiver\n", DEBUG_NAME);
//    err = InsertReceiver(fbana, receiver);
//    if (err < 0)
//        return err;

//    err = UpdateReceiver(fbana, retr_round);

    return err;
}

static int InitialiseFeedback(struct prrt_fb_analyzer *fbana, u16 miss, u8 retr_round);

/*insert the feedback entry re into the list of feedbacks in fbana*/
static int InsertFeedback(struct prrt_fb_analyzer *fbana, u16 miss, u8 retr_round)
{
    list_add_tail(&fbana->fbe->list, &fbana->fb_list);
    fbana->count++;

    InitialiseFeedback(fbana, miss, retr_round);

    PDEBUG(MSG_FBANADEB, "%s: Inserted feedback for packet no %u (count %d)\n", DEBUG_NAME, miss, fbana->count);

    return 0;
}

static int InitialiseFeedback(struct prrt_fb_analyzer *fbana, u16 miss, u8 retr_round)
{
    //INIT_LIST_HEAD(&fbana->fbe->recvs);
    fbana->fbe->count = 0;
    fbana->fbe->packno = miss;
    fbana->fbe->retr_round = retr_round;

    return 0;
}

/*insert the receiver entry re into the list of receivers in fbe*/
//static int InsertReceiver(struct prrt_fb_analyzer *fbana, u32 receiver)
//{
//    list_add_tail(&fbana->re->list, &fbana->fbe->recvs);
//
//    fbana->re->recv_ip = receiver;
//    fbana->fbe->count++;
//
//    return 0;
//}

//static int UpdateReceiver(struct prrt_fb_analyzer *fbana, int retr_round)
//{
//    fbana->re->retr_round = retr_round;
//
//    return 0;
//}

static int InsertReplacementForFeedback(struct prrt_fb_analyzer *fbana, struct prrt_block_list *blockList,
        struct prrt_receiver_list *rlist);

int prrt_fbana_compute_replacements(struct prrt_fb_analyzer *fbana, struct prrt_block_list *blockList,
        struct prrt_receiver_list *rlist)
{
    struct list_head *iter = 0;
    struct list_head *next = 0;
    int err = 0;

    FreeReplacements(fbana);

    down(&fbana->lock);
    down(&fbana->repl_lock);

    /*Search in the complete list of feedbacks*/
    list_for_each_safe(iter, next, &fbana->fb_list)
    {
        ResetInternalEntries(fbana);
        fbana->fbe = container_of(iter, struct fbana_feedback_entry, list);

        err = InsertReplacementForFeedback(fbana, blockList, rlist);
        if (err < 0)
        {
            PDEBUG(MSG_FBANADEB, "%s: Inserting replacement failed\n", DEBUG_NAME);
            continue;
        }

        /*this feedback is handled, delete it*/
        list_del(iter);
        free_feedback_entry(&fbana->fbe);
        fbana->count--;
        //PDEBUG(MSG_FBANADEB, "%s: First replacement done\n", DEBUG_NAME);
    }

    up(&fbana->repl_lock);
    up(&fbana->lock);

    return 0;
}

static int ReplacementReady(struct prrt_fb_analyzer *fbana, struct prrt_block_list *blockList);
//static int UpdateMax_n_cc(struct prrt_fb_analyzer *fbana, struct prrt_block *mblock);
//static int UpdateReplaceTimes(struct prrt_fb_analyzer *fbana, struct prrt_receiver_list *rlist,
//        struct prrt_packet_container *pcont);
static int SearchReplacement(struct prrt_fb_analyzer *fbana);
static int InsertReplacement(struct prrt_fb_analyzer *fbana, struct prrt_block *mblock);

static int InsertReplacementForFeedback(struct prrt_fb_analyzer *fbana, struct prrt_block_list *blockList,
        struct prrt_receiver_list *rlist)
{
    struct prrt_block *mblock = 0;
    //struct packet_list_entry *p = 0;
    int err = 0;
    char found_replacement = 0;
    int n_rr = 0;

    if (!ReplacementReady(fbana, blockList))
    {
        PDEBUG(MSG_FBANADEB, "%s: Replacements not yet ready\n", DEBUG_NAME);
        return -1;
    }

    err = prrtblist_get_block(blockList, fbana->fbe->packno, &mblock);
    if (err < 0){
	    PDEBUG(MSG_FBANADEB, "%s: Did not get block\n", DEBUG_NAME);
	    return err;
    }

    /*update the replace times for the different receivers if needed*/
//#if UNIT_TEST == 0
//    err = prrt_block_get_packet(mblock, fbana->fbe->packno, &p);
//    if (err == 0)
//    {
//        UpdateReplaceTimes(fbana, rlist, p->pcont);
//    }
//#endif

    n_rr = prrtcpar_get_n_rr(&mblock->cpars);
    if (fbana->fbe->retr_round > n_rr){
	    PDEBUG(MSG_FBANADEB, "%s: n_rr exceeded\n", DEBUG_NAME);
	    return 0;
    }
	
    found_replacement = SearchReplacement(fbana);

    if (!found_replacement)
    {
	    PDEBUG(MSG_SENDREPL, "%s: Have no entry for this block %d in round %d\n", DEBUG_NAME, fbana->fbe->packno, fbana->fbe->retr_round);
	    err = InsertReplacement(fbana, mblock);
    }
    else
    {
	    //PDEBUG(MSG_SENDREPL, "%s: Have already entry for this block %d in round %d\n", DEBUG_NAME, fbana->fbe->packno, fbana->fbe->retr_round);
	    /*we already have a replacement entry for this block*/
	    if (!TYPEII_HEC_MODE)
	    {
		    fbana->rple->no_repl++;
		    PDEBUG(MSG_SENDREPL, "%s: Incremented number of replacements to %d\n", DEBUG_NAME, fbana->rple->no_repl);
	    }
	    //err = UpdateMax_n_cc(fbana, mblock);
    }

    return err;
};

static int ReplacementReady(struct prrt_fb_analyzer *fbana, struct prrt_block_list *blockList)
{
    int last_pno = 0;
    int err = 0;

    err = prrtblist_get_last_packetno(blockList, &last_pno);
    if (err < 0)
    {
        PDEBUG(MSG_FBANADEB, "%s: Could not find last packet number in block list\n", DEBUG_NAME);
        return 0;
    }

    /*replacements are ready if the block containing this packetno is in the list*/
    if (prrt_comp_packetnos(fbana->fbe->packno, last_pno) > 0)
    {
        PDEBUG(MSG_FBANADEB, "%s: Block with this packet is not in list\n", DEBUG_NAME);
        return 0;
    }

    //PDEBUG(MSG_FBANADEB, "%s: Replacement for block is ready\n", DEBUG_NAME);

    return 1;
}

//static int UpdateMax_n_cc(struct prrt_fb_analyzer *fbana, struct prrt_block *mblock)
//{
//    struct list_head *iter = 0;
//    struct fbana_receiver_entry *re = 0;
//    int n_cc = 0;
//    int n_rr = 0;
//
////    list_for_each(iter, &fbana->fbe->recvs)
////    {
////        re = container_of(iter, struct fbana_receiver_entry, list);
////
////        n_rr = prrtcpar_get_n_rr(&mblock->cpars);
////        if (re->retr_round <= n_rr)
////            n_cc = prrtcpar_get_n_cc(&mblock->cpars, re->retr_round);
////        PDEBUG(MSG_SENDREPL, "%s: replacement multiplier of block is %d\n", DEBUG_NAME, n_cc);
////        PDEBUG(MSG_SENDREPL, "%s: retransmission round is %d\n", DEBUG_NAME, re->retr_round);
////
////        fbana->rple->max_n_cc = MAX(fbana->rple->max_n_cc, n_cc);
////	fbana->rple->retr_round = re->retr_round;
////    }
//        n_rr = prrtcpar_get_n_rr(&mblock->cpars);
//        if (re->retr_round <= n_rr)
//
//    return 0;
//}


//static int UpdateReplaceTimes(struct prrt_fb_analyzer *fbana, struct prrt_receiver_list *rlist,
//        struct prrt_packet_container *pcont)
//{
	//TODO: Move GRTT updating stuff here (from recv_feedback.c)
//    unsigned long first_transmit = 0;
//    unsigned long timestamp = prrt_get_prrttime();
//    unsigned long sampleReplace = 0;
    int err = 0;
//    struct list_head *iter = 0;
//    struct fbana_receiver_entry *re = 0;
//
//    list_for_each(iter, &fbana->fbe->recvs)
//    {
//        re = container_of(iter, struct fbana_receiver_entry, list);
//
//        if (re->retr_round == 1)
//        {
//            /*should only be called for feedbacks of the first transmission round*/
//            first_transmit = pcont->time.transmit_time;
//            PDEBUG(MSG_PARAM, "%s: Now = %lu\n", DEBUG_NAME, timestamp);
//            PDEBUG(MSG_PARAM, "%s: Old = %lu\n", DEBUG_NAME, first_transmit);
//            sampleReplace = prrt_diff_timestamps(timestamp, first_transmit);
//            if (rlist != 0)
//                prrtrlist_update_ReplaceTime(rlist, re->recv_ip, sampleReplace);
//        }
//    }
//
//    return err;
//}



static int FeedbackBelongsToReplacement(struct prrt_fb_analyzer *fbana,
        struct fbana_replacement_entry *rple);

/*test if there is an entry for a certain block in the list of replacements*/
static int SearchReplacement(struct prrt_fb_analyzer *fbana)
{
    struct list_head *iter = 0;
    struct fbana_replacement_entry *rple = 0;

    fbana->rple = 0;

    list_for_each(iter, &fbana->repl_list)
    {
        rple = container_of(iter, struct fbana_replacement_entry, list);

        if (FeedbackBelongsToReplacement(fbana, rple))
        {
            fbana->rple = rple;
            return 1;
        }
    }

    return 0;
}

static int FeedbackBelongsToReplacement(struct prrt_fb_analyzer *fbana,
        struct fbana_replacement_entry *rple)
{
//    return (prrt_comp_packetnos(rple->first_packno, fbana->fbe->packno) <= 0 &&
//            prrt_comp_packetnos(fbana->fbe->packno, rple->last_packno) <= 0);
    return (prrt_comp_packetnos(rple->first_packno, fbana->fbe->packno) <= 0 &&
            prrt_comp_packetnos(fbana->fbe->packno, rple->last_packno) <= 0 &&
	    (u8)fbana->fbe->retr_round == rple->retr_round);
}

static int InsertReplacementForRetransmission(struct prrt_fb_analyzer *fbana, struct prrt_block *mblock);
static int InsertReplacementForCodedBlock(struct prrt_fb_analyzer *fbana, struct prrt_block *mblock);

static int InsertReplacement(struct prrt_fb_analyzer *fbana, struct prrt_block *mblock)
{
    int err = 0;
    u8 block_retr_round = mblock->retr_round;
    u8 fb_retr_round = fbana->fbe->retr_round;
    u8 round = 0;

    u32 time = 0;
    time = prrt_get_prrttime(fbana->time_state);

    PDEBUG(MSG_FBANADEB, "%s: Inserting replacement for block starting with no %u (at %u) (seqnum %u)\n", DEBUG_NAME,
            (u16)prrt_block_get_first_packetno(mblock), time, fbana->fbe->packno);

    PDEBUG(MSG_FBANADEB, "%s: Inserting fb block_retr_round %d, fb_retr_round %d\n", DEBUG_NAME, block_retr_round, fb_retr_round);

    if (prrtcpar_get_r(&mblock->cpars) == 0 || prrtcpar_get_k(&mblock->cpars) == 1)
    {	
	    for (round = block_retr_round+1; round <= fb_retr_round; round++){
		    PDEBUG(MSG_FBANADEB, "%s: Inserting retransmission for round %d\n", DEBUG_NAME, round);
		    fbana->fbe->retr_round = round;
		    mblock->retr_round++;
		    err = InsertReplacementForRetransmission(fbana, mblock);
		    if (err < 0)
			    return err;
	    }
    }
    else
    {
	    for (round = block_retr_round+1; round <= fb_retr_round; round++){
		    PDEBUG(MSG_FBANADEB, "%s: Inserting parity for round %d\n", DEBUG_NAME, round);
		    fbana->fbe->retr_round = round;
		    mblock->retr_round++;
		    err = InsertReplacementForCodedBlock(fbana, mblock);
		    if (err < 0)
			    return err;
	    }
    }

    return err;
}

static int InsertReplacementForRetransmission(struct prrt_fb_analyzer *fbana, struct prrt_block *mblock)
{
    int err = 0;

    /*we need a new replacement entry for a block that can not be coded
     *=> retransmissions*/
    err = AllocReplacementEntry(&fbana->rple);
    if (err < 0)
        return err;

    /*later on get_block will search for the block containing this packno
     *get_next_replacement will give the packet with this number if
     *retransmissions are needed*/
    fbana->rple->first_packno = fbana->fbe->packno;
    fbana->rple->last_packno = fbana->fbe->packno;
    fbana->rple->no_repl = 1;
    fbana->rple->retr_round = fbana->fbe->retr_round;
    //UpdateMax_n_cc(fbana, mblock);
    fbana->rple->max_n_cc = prrtcpar_get_n_cc(&mblock->cpars, fbana->fbe->retr_round);
    //PDEBUG(MSG_SENDREPL, "%s: Inserted replacement: replacement multiplier of packet %d is %d\n", DEBUG_NAME, fbana->rple->max_n_cc, fbana->rple->first_packno);
    //PDEBUG(MSG_SENDREPL, "%s: retransmission round is %d\n", DEBUG_NAME, fbana->rple->retr_round);

    list_add(&fbana->rple->list, &fbana->repl_list);

    return 0;
}

static int InsertReplacementForCodedBlock(struct prrt_fb_analyzer *fbana, struct prrt_block *mblock)
{
    struct packet_list_entry *p = 0;
    int err = 0;

    /*we need a new replacement entry for this block*/
    err = AllocReplacementEntry(&fbana->rple);
    if (err < 0)
        return err;

    err = prrt_block_get_first_data(mblock, &p);
    if (err < 0)
    {
        prrt_fbana_free_replacement_entry(&fbana->rple);
        return err;
    }

    fbana->rple->first_packno = p->pcont->pack.data.header.seqnum;

    err = prrt_block_get_last_data(mblock, &p);
    if (err < 0)
    {
        prrt_fbana_free_replacement_entry(&fbana->rple);
        return err;
    }

    fbana->rple->last_packno = p->pcont->pack.data.header.seqnum;
    fbana->rple->no_repl = 1;
    //UpdateMax_n_cc(fbana, mblock);
    fbana->rple->retr_round = fbana->fbe->retr_round;
    fbana->rple->max_n_cc = prrtcpar_get_n_cc(&mblock->cpars, fbana->fbe->retr_round);
    //PDEBUG(MSG_SENDREPL, "%s: Inserted replacement: replacement multiplier of block %d is %d\n", DEBUG_NAME, fbana->rple->max_n_cc, fbana->rple->first_packno);
    //PDEBUG(MSG_SENDREPL, "%s: retransmission round is %d\n", DEBUG_NAME, fbana->rple->retr_round);

    list_add_tail(&fbana->rple->list, &fbana->repl_list);
    PDEBUG(MSG_FBANADEB, "%s: Inserted replacement for packets %d to %d (round %d)\n", DEBUG_NAME,
          fbana->rple->first_packno, fbana->rple->last_packno, fbana->rple->retr_round);

    return err;
}

int prrt_fbana_get_next_replacement(struct prrt_fb_analyzer *fbana, struct fbana_replacement_entry **rple)
{
    struct list_head *iter = 0;
    int err = 0;

    down(&fbana->repl_lock);

    if (list_empty(&fbana->repl_list))
    {
        err = -1;
        PDEBUG(MSG_SENDREPL, "%s: Replacement list is empty\n", DEBUG_NAME);
        goto done;
    }

    iter = fbana->repl_list.next;
    *rple = container_of(iter, struct fbana_replacement_entry, list);
    list_del(iter);

    PDEBUG(MSG_SENDREPL, "%s: Found replacement for packetnos %u to %u\n", DEBUG_NAME,
            (*rple)->first_packno, (*rple)->last_packno);

done:
    up(&fbana->repl_lock);

    return err;
}
