/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#ifndef FEEDBACK_ANALYZER_H
#define FEEDBACK_ANALYZER_H

#include "defines.h"

#include <linux/list.h>
#include <linux/version.h>
#include "kui/kui_option.h"
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26)
#include <asm/semaphore.h>
#else
#include <linux/semaphore.h>
#endif
#include <linux/slab.h>

#include "lists/receiver_list.h"
#include "lists/prrt_block_list.h"

/**
 * This struct is used to insert a receiver in the list of receivers that sent a feedback.
 */
struct fbana_receiver_entry {
	struct list_head list;
	u32 recv_ip;    /**< IP address of the receiver.*/
	int retr_round; /**< Round of the retransmission.*/
};

/**
 * This struct is used to insert a feedback into the list of feedbacks.
 */
struct fbana_feedback_entry {
	struct list_head list;  /**< This is used to build the list of feedbacks.*/
	//struct list_head recvs; /**< This is a list of receivers that send this feedback.*/
	u16 packno;             /**< The number of the missing packet.*/
	int count;              /**< Number of receivers that have send feedback.*/
	u8 retr_round;	 	/**< Round of the retransmission.*/
};

/**
 * This struct is used to save the computed replacements.
 */
struct fbana_replacement_entry {
	struct list_head list;
	u16 first_packno;   /**< The number of the first packet in the block.*/
	u16 last_packno;    /**< The number of the last packet in the block.*/
	int no_repl;        /**< The number of replacements for that block.*/
	int max_n_cc;       /**< Maximum multiplier for replacement sending.*/
	u8 retr_round;	    /**< Retransmission round of this replacement.*/
};

/**
 * This class gathers the feedback at the sender and analyzes it for replacement sending.
 * It saves the feedbacks in a list containing each feedback. Each feedback again contains a list
 * of receivers that sent it.
 * When replacements should be sent they are computed and saved in the repl_list where they can
 * easyly be retrieved from.
 */
__PCLASS __THREAD_SAFE struct prrt_fb_analyzer {
	struct list_head fb_list;   /**< Head of the list to gather feedback.*/
	struct list_head repl_list; /**< Head of the list for the computed replacements.*/
	struct semaphore lock;      /**< Mutex to synchronize access to fb_list.*/
	struct semaphore repl_lock; /**< Mutex to synchronize access to repl_list.*/
	int count;                  /**< Counter for the number of feedbacks in the list.*/
	struct prrt_time_state *time_state; /**< The state of time in this socket. */

	__LOCAL
	struct fbana_feedback_entry *fbe;
	struct fbana_receiver_entry *re;
	struct fbana_replacement_entry *rple;
};

/**
 * Initialise the cache structure that is used for packets.
 */
int prrt_fbana_caches_init(void);

/**
 * Uninitialise the cache structure for packets.
 */
void prrt_fbana_caches_uninit(void);

/**
 * Initialise a feedback analyzer.
 */
int prrt_fbana_init(struct prrt_fb_analyzer *fbana, struct prrt_time_state *time_state);
int prrt_fbana_reinit(struct prrt_fb_analyzer *fbana);

/**
 * Uninitialise a feedback analyzer and free allocated memory.
 */
void prrt_fbana_free(struct prrt_fb_analyzer *fbana);

/**
 * Delete all entries in the feedback list.
 */
void prrt_fbana_free_feedbacks(struct prrt_fb_analyzer *fbana);

/**
 * Insert a feedback request in the queue.
 * \param miss number of the missing packet.
 * \param receiver IP address of the receiver.
 * \param retr_round the number of times this feedback was transmitted.
 */
int prrt_fbana_insert_feedback(struct prrt_fb_analyzer *fbana, u16 miss, u32 receiver, int retr_round);

/**
 * Compute a list of replacement entries which can be retrieved afterwards.
 */
int prrt_fbana_compute_replacements(struct prrt_fb_analyzer *fbana, struct prrt_block_list *blockList,
				   struct prrt_receiver_list *rlist);

/**
 * Get a replacement entry from the list of replacements.
 * The retrieved replacement entry is deleted from the replacement list. Delete it after using it.
 * \param rple this will point to the replacement entry on success.
 * \return 0 on success, -1 on failure.
 */
int prrt_fbana_get_next_replacement(struct prrt_fb_analyzer *fbana, struct fbana_replacement_entry **rple);

/**
 * Free a replacement entry that was retrieved from the replacement list.
 */
void prrt_fbana_free_replacement_entry(struct fbana_replacement_entry **rple);

#endif  /*FEEDBACK_ANALYZER_H*/
