/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */


/*
 * This file holds the interface to control various functions of a socket.
 *
 * Return value table:
 * -------------------
 * see kui_error.h
 *
 * Functions:
 * ----------
 * 1. BIND:
 *    call with: bind <haddr> <hport> <paddr>
 *    For multicast the first address is the address of the host and the second one is the multicast
 *    address.
 *    params:
 *     - <haddr>: The address the socket should be bound to.
 *     - <hport>: The port the socket should be bound to.
 *     - <paddr>: The address the socket should send feedback to (feedback port is <hport>+1).
 *    returns: BIND_SUCC(success)
 *
 * 2. CONNECT:
 *    call with: connect <paddr> <pport> <haddr>
 *    For multicast there is only one address needed. The second address should be set to "0".
 *    params:
 *     - <paddr>: The address of the peer to connect to.
 *     - <pport>: The port of the peer to connect to.
 *     - <paddr>: The address where the host listens for feedback (feedback port is <pport>+1).
 *    returns: CONN_SUCC(success)
 *
 */

#include "kernel_user_interface.h"
#include "kui_error.h"

static const char *DEBUG_NAME = "KUI_CONTROL";

/*type for different events on this socket type*/
enum control_event_t {
	NO_EVENT,	/*this describes that no event has happened yet*/
	BIND,		/*bind the socket to an address supplied by the user*/
	CONNECT
};

struct control_event {
	enum control_event_t 	type;
	int			error;		/*this flag is set if an error
						 *occured*/
	Socketno_t 		sono;		/*for ACCEPT: the socketno of
						 *the new prrt socket*/
};

struct control_priv_data {
	Socketno_t		socketno;	/*socketno of the control file*/
	struct semaphore	mutex;		/*mutex to synchronize access
						 *to one file*/
	struct control_event	last_event;
};

/*from kernel_user_interface.c*/
Socketno_t kui_register_socket(int flags, struct prrt_socket *msock, int pid);
void kui_unregister_socket(Socketno_t socketno);

/*from socket_manager.c*/
int sm_get_prrt_sock(Socketno_t socketno, struct prrt_socket **msock);

/*from below*/
int kui_control_read_proc(char *buf, char **start, off_t offset,
				int count, int *eof, void *data);
int kui_control_write_proc(struct file *file, const char __user *buffer,
				unsigned long count, void *data);
static int checkInput(char *input);
// static int validate_sock_addr(char *addr, int addrlen);

/*
 * This function creates a new file in the procfs for controling a socket.
 */
void *kui_create_control_file(Socketno_t socketno)
{
	char name[NAME_SIZE];
	struct proc_dir_entry *procentry = 0;

	struct control_priv_data *priv_data = 0;

	priv_data = kmalloc(sizeof(*priv_data), GFP_KERNEL);

	priv_data->socketno = socketno;
	init_MUTEX(&(priv_data->mutex));
	priv_data->last_event.type = NO_EVENT;
	priv_data->last_event.error = 0;
	priv_data->last_event.sono = 0;

	snprintf(name, NAME_SIZE, "prrt/%d/control", socketno);

	procentry = create_proc_entry(name, S_IRUGO|S_IWUGO, NULL);
	if(procentry) {
		procentry->read_proc  = kui_control_read_proc;
		procentry->write_proc = kui_control_write_proc;
		procentry->data = priv_data;

		return priv_data;	/*SUCCESS*/
	}

	return 0;	/*FAILURE*/
}

/*
 * Function for deleting a procfs control file.
 */
void kui_delete_control_file(Socketno_t socketno)
{
	char name[NAME_SIZE];
	struct prrt_socket *msock = 0;
	struct control_priv_data *pd;

	snprintf(name, NAME_SIZE, "prrt/%d/control", socketno);

	while(sm_get_prrt_sock(socketno, &msock) == -ERESTARTSYS)
		;
	if(!msock)
		return;

	pd = (struct control_priv_data *)msock->control_pd;

	while(down_interruptible(&pd->mutex))
		;

	/*free private data of file*/
	kfree(msock->control_pd);

	remove_proc_entry(name, NULL);

	up(&pd->mutex);
}

/*
 * Read method for a control file.
 * Gives feedback when a user process tries to change options
 * or to close a socket.
 */
int kui_control_read_proc(char *buf, char **start, off_t offset,
				int count, int *eof, void *data)
{
	int len = 0;
	int ret = 0;
	struct control_priv_data *pd = (struct control_priv_data *)data;

	int interrupted = down_interruptible(&pd->mutex);
	if(interrupted)
		return -ERESTARTSYS;


	if(!pd->last_event.error) {
		switch(pd->last_event.type) {
		case NO_EVENT:
			len = snprintf(buf, PAGE_SIZE, "%da", NOEVENT);
			break;
		case BIND:
			len = snprintf(buf, PAGE_SIZE, "%da", BIND_SUCC);
			break;
		case CONNECT:
			//len = snprintf(buf, PAGE_SIZE, "%da", CONN_SUCC);
			len = snprintf(buf, PAGE_SIZE, "%da", pd->last_event.sono);
			break;
		default:
			/* TODO:Error handling*/
			break;
		}
	} else {
		len = snprintf(buf, PAGE_SIZE, "%da", pd->last_event.error);
	}

	if(len < 0 || len > PAGE_SIZE) {
		PERROR("%s: Could not write to proc file\n", DEBUG_NAME);
		goto free_mutex;
	}

	pd->last_event.type = NO_EVENT;
	pd->last_event.error = 0;
	pd->last_event.sono = 0;
	*eof = 1;
	ret = len;

free_mutex:
	up(&pd->mutex);

	return ret;
}

/*
 * Write method for a control file.
 * Here the input from the user process is analysed and the proper
 * action is done.
 */
int kui_control_write_proc(struct file *file, const char __user *buffer,
				unsigned long count, void *data)
{
	char *kernel_buffer = 0;
	char command[8];	/*this is either "bind" or "connect"*/
	char arg1[16];		/*this is an ip address of form x.x.x.x*/
	unsigned int arg2;	/*this is a port*/
	char arg3[16];		/*this is an ip address of form x.x.x.x*/
	char *arg3_pointer = arg3;	/*this will be NULL if multicast is demanded*/

	int not_copied = 0;
	struct prrt_socket *msock = 0;
	int err = 0;
	struct control_priv_data *pd = (struct control_priv_data *)data;
	int kui_port = 0;

	int interrupted = down_interruptible(&pd->mutex);
	if(interrupted) {
		pd->last_event.error = -EINTERR;
		return -ERESTARTSYS;
	}

	PDEBUG(MSG_KUICONTROL, "%s: Control for socket %d called...\n", DEBUG_NAME, pd->socketno);

	kernel_buffer = kmalloc(count+1, GFP_KERNEL);
	if(!kernel_buffer) {
		pd->last_event.error = -EALLOC;
		err = -ENOMEM;
		goto nomem;
	}

	/*TODO: what to do if something was not copied*/
	not_copied = copy_from_user(kernel_buffer, buffer, count);
	kernel_buffer[count] = '\0';

	err = checkInput(kernel_buffer);
	if(err < 0) {
		PERROR("%s: Malformed input: %s\nError no is %d\n", DEBUG_NAME,	kernel_buffer, err);
		pd->last_event.error = -EINVINP;
		goto error;
	}
	/*from this point on you can assume that the input is well formed*/

	sscanf(kernel_buffer, "%7s %15s %u %15s", command, arg1, &arg2, arg3);
	PDEBUG(MSG_KUICONTROL, "%s: Command: %s\n", DEBUG_NAME, command);
	PDEBUG(MSG_KUICONTROL, "%s: Command: %s\n", DEBUG_NAME, arg1);
	PDEBUG(MSG_KUICONTROL, "%s: Command: %u\n", DEBUG_NAME, arg2);
	PDEBUG(MSG_KUICONTROL, "%s: Command: %s\n", DEBUG_NAME, arg3);

	/*if multicast is demanded we don't need the last argument*/
	if(arg3[0] == '0' && strlen(arg3) == 1)
		arg3_pointer = 0;

	if(strncmp("bind", command, 4) == 0) {
		PDEBUG(MSG_KUICONTROL, "%s: Bind was called...\n", DEBUG_NAME);
		pd->last_event.type = BIND;
		/*bind called
		 *sockaddr in arg1
		 *portno in arg2
		 *feedback sockaddr in arg3*/
		while(sm_get_prrt_sock(pd->socketno, &msock) == -ERESTARTSYS)
			;
		if(!msock) {
			/*ERROR: socket not found*/
			pd->last_event.error = -ENOSOCK;
			goto error;
		}
       
        // Determine which prrt_bind function we use.
        if ( strncmp("as_recv", arg3, 7) == 0 ) {
          pd->last_event.error = prrt_bind_as_recv(msock, arg1, arg2);
//           pd->last_event.sono = BIND_SUCC;  // FIXME there is no socket port in this case. just inform the caller from the user space, the binding is successful.
        }
        else if ( strncmp("as_send", arg3, 7) == 0 ) {
          pd->last_event.error = prrt_bind_as_send(msock, arg1, arg2);
//           pd->last_event.sono = kui_port;
        }
        else {
          // The original bind() is used here, since arg3 is specified as an IP address.
          pd->last_event.error = prrt_bind(msock, arg1, arg2, arg3_pointer);
//           pd->last_event.sono = BIND_SUCC;  // FIXME there is no socket port in this case. just inform the caller from the user space, the binding is successful.
        }

		if(pd->last_event.error < 0)
			pd->last_event.error = -EGENERAL;


	} else if(strncmp("connect", command, 7)==0){
		PDEBUG(MSG_KUICONTROL, "%s: Connect was called...\n", DEBUG_NAME);
		pd->last_event.type = CONNECT;
		/*connect called
		 *peer address in arg1
		 *peer port in arg2
		 *feedback address in arg3*/
		while(sm_get_prrt_sock(pd->socketno, &msock) == -ERESTARTSYS)
			;
		if(!msock) {
			/*ERROR: socket not found*/
			pd->last_event.error = -ENOSOCK;
			goto error;
		}

        // Determine which prrt_connect function we use.
        if ( strncmp("to_receiver", arg3, 11) == 0 ) {
          pd->last_event.error = prrt_connect_to_receiver(msock, arg1, arg2, &kui_port);
          pd->last_event.sono = kui_port;
        }
        else if ( strncmp("to_feedback", arg3, 11) == 0 ) {
          pd->last_event.error = prrt_connect_to_feedback(msock, arg1, arg2);
          pd->last_event.sono = CONN_SUCC;  // FIXME there is no socket port in this case. just inform the caller from the user space, the connecting is successful.
        }
        else {
          // The original connect() is used here, since arg3 is specified as an IP address.
          pd->last_event.error = prrt_connect(msock, arg1, arg2, arg3_pointer, &kui_port);
          pd->last_event.sono = kui_port;
        }
 
		if(pd->last_event.error < 0)
			pd->last_event.error = -EGENERAL;

	} else {
		/*because of the new check above this should never happen*/
		PDEBUG(MSG_KUICONTROL, "%s: No valid option called: %s\n", DEBUG_NAME,
			kernel_buffer);
		pd->last_event.type = NO_EVENT;
	}

error:
	kfree(kernel_buffer);
	up(&pd->mutex);
	return count;

nomem:
	up(&pd->mutex);
	pd->last_event.error = -ENOMEM;
	return -ENOMEM;
}

static int checkCommand(char *command)
{
	int lcommand = strlen(command);
	int lcom;
	char *com;

	com = "bind";
	lcom = strlen(com);
	if(lcommand == lcom && strncmp(com, command, lcom) == 0)
		return 0;

	com = "connect";
	lcom = strlen(com);
	if(lcommand == lcom && strncmp(com, command, lcom) == 0)
		return 0;

	return -1;
}

/*
 *
 */
static int checkIP(char *ip)
{
	int i = 0;
// 	int no_blocks = 0;
	unsigned int blocks[4];
	int ret = 0;
	char end[2];

	ret = sscanf(ip, "%u.%u.%u.%u%1s", &blocks[0], &blocks[1], &blocks[2], &blocks[3], end);
	if(ret != 4)
		return -1;

	for(i=0; i<4; i++) {
// 		printk(KERN_ALERT "block%d: %d\n", i, blocks[i]);
		if(blocks[i]>255)
			return -1;
	}

	return 0;
}

static int checkPort(unsigned int port)
{
	if(0 < port && port < 65536)
		return 0;

	return -1;
}

/* check if we use the new bind() and connect().
 *
 */
static int checkOption(char *option)
{
  int loption = strlen(option);
  int lopt = 0;
  char *opt;

  opt = "as_recv";
  lopt = strlen(opt);
  if(loption == lopt && strncmp(opt, option, lopt) == 0)
    return 0;

  opt = "as_send";
  lopt = strlen(opt);
  if(loption == lopt && strncmp(opt, option, lopt) == 0)
    return 0;

  opt = "to_receiver";
  lopt = strlen(opt);
  if(loption == lopt && strncmp(opt, option, lopt) == 0)
    return 0;

  opt = "to_feedback";
  lopt = strlen(opt);
  if(loption == lopt && strncmp(opt, option, lopt) == 0)
    return 0;
  
  return -1;
}

/*
 * Checks if the input is valid.
 * \return * 0 on success
 *         * -1 on invalid format
 *         * -2 on invalid command
 *         * -3 on invalid first argument
 *         * -4 on invalid second argument
 *         * -5 on invalid third argument
 */
static int checkInput(char *input)
{
	char command[8];	/*this is either "bind" or "connect"*/
	char arg1[16];		/*this is an ip address of form x.x.x.x*/
	unsigned int arg2;	/*this is a port*/
    char arg3[16];      /*this is an ip address of form x.x.x.x or an option form for new bind()/connect() */
	int ret = 0;

	ret = sscanf(input, "%7s %15s %u %15s", command, arg1, &arg2, arg3);
	if(ret != 4) {
		return -1;
	}

	PDEBUG(MSG_COMMON, "%s: Checking command: %s\n", DEBUG_NAME, command);
	if(checkCommand(command) < 0)
		return -2;

	PDEBUG(MSG_COMMON, "%s: Checking arg1: %s\n", DEBUG_NAME, arg1);
	if(checkIP(arg1) < 0)
		return -3;

	PDEBUG(MSG_COMMON, "%s: Checking arg2: %u\n", DEBUG_NAME, arg2);
	if(checkPort(arg2) < 0)
		return -4;

	PDEBUG(MSG_COMMON, "%s: Checking arg3: %s\n", DEBUG_NAME, arg3);
    // Determine if arg3 is valid.
    // The original bind()/connect() is used if it's an IP address form.
    // The new bind()/connect() is used if it's an option form to specify which bind()/connect() is used.
    if(checkIP(arg3) < 0 && strncmp(arg3, "0", 1) != 0 && checkOption(arg3) < 0)
      return -5;

	return 0;
}

