/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "kernel_user_interface.h"
#include "../prrt_socket.h"
#include "kui_error.h"
#include "string_helper.h"

static const char *DEBUG_NAME = "KUI_OPTION";

struct option_event
{
    enum option_t type;
    int error; /*this flag is set if an error
						 *occured*/
};

struct option_priv_data
{
    Socketno_t socketno; /*socketno of the control file*/
    struct semaphore mutex; /*mutex to synchronize access
						 *to one file*/
    struct option_event last_event;
};

/*from socket_manager.c*/
int sm_get_prrt_sock(Socketno_t socketno, struct prrt_socket **msock);

/*from below*/
int kui_option_read_proc(char *buf, char **start, off_t offset,
        int count, int *eof, void *data);
int kui_option_write_proc(struct file *file, const char __user *buffer,
        unsigned long count, void *data);

void *kui_create_option_file(Socketno_t socketno)
{
    char name[NAME_SIZE];
    struct proc_dir_entry *procentry = 0;

    struct option_priv_data *priv_data = 0;

    priv_data = kmalloc(sizeof (*priv_data), GFP_KERNEL);

    priv_data->socketno = socketno;
    init_MUTEX(&(priv_data->mutex));
    priv_data->last_event.type = NONE;
    priv_data->last_event.error = 0;

    snprintf(name, NAME_SIZE, "prrt/%d/option", socketno);

    procentry = create_proc_entry(name, S_IRUGO | S_IWUGO, NULL);
    if (procentry)
    {
        procentry->read_proc = kui_option_read_proc;
        procentry->write_proc = kui_option_write_proc;
        procentry->data = priv_data;

        return priv_data; /*SUCCESS*/
    }

    return 0; /*FAILURE*/
}

void kui_delete_option_file(Socketno_t socketno)
{
    char name[NAME_SIZE];
    struct prrt_socket *msock = 0;
    struct option_priv_data *pd;

    snprintf(name, NAME_SIZE, "prrt/%d/option", socketno);

    while (sm_get_prrt_sock(socketno, &msock) == -ERESTARTSYS)
        ;
    if (!msock)
        return;

    pd = (struct option_priv_data *) msock->option_pd;

    while (down_interruptible(&pd->mutex))
        ;

    /*free private data of file*/
    kfree(msock->option_pd);

    remove_proc_entry(name, NULL);

    up(&pd->mutex);
}

static int checkVal(char *option, int val)
{
    int loption = strlen(option);
//     int lval = strlen(val);
//     int lv;
//     char *v;

    if (stringEqualsTestString(option, loption, "prrt_rcvbuf")
      || stringEqualsTestString(option, loption, "kernel_rcvbuf")
      || stringEqualsTestString(option, loption, "user_rcvbuf")
      || stringEqualsTestString(option, loption, "fb_supp_on")
      || stringEqualsTestString(option, loption, "trafficshaping")) {
      if (0 <= val)
        return 0;
    }
    else {
      if (0 == val || 1 == val)
        return 0;
    }

//     v = "0";
//     lv = strlen(v);
//     if (lval == lv && strncmp(v, val, lval) == 0)
//         return 0;
// 
//     v = "1";
//     lv = strlen(v);
//     if (lval == lv && strncmp(v, val, lval) == 0)
//         return 0;

    return -1;
}

static int checkOption(char *option)
{
    int loption = strlen(option);

    if (stringEqualsTestString(option, loption, "feedback"))
        return 0;

    if (stringEqualsTestString(option, loption, "replace"))
        return 0;

    if (stringEqualsTestString(option, loption, "coding"))
        return 0;

    if (stringEqualsTestString(option, loption, "timeout"))
        return 0;

    if (stringEqualsTestString(option, loption, "fixrepl"))
        return 0;

    if (stringEqualsTestString(option, loption, "prrt_rcvbuf"))
        return 0;

    if (stringEqualsTestString(option, loption, "kernel_rcvbuf"))
        return 0;

    if (stringEqualsTestString(option, loption, "user_rcvbuf"))
        return 0;

    if (stringEqualsTestString(option, loption, "trafficshaping"))
        return 0;

    if (stringEqualsTestString(option, loption, "fb_supp_on"))
        return 0;

    if (stringEqualsTestString(option, loption, "fb_supp_off"))
        return 0;

    return -1;
}

static int checkInput(char *input)
{
//     char option[9];
    char option[14];
//     char val[20];
    int val = 0;
    int ret = 0;

//     ret = sscanf(input, "%8s %20s", option, val);
    ret = sscanf(input, "%14s %d", option, &val);
    if (ret != 2)
    {
        return -1;
    }

    if (checkOption(option) < 0)
        return -2;

    if (checkVal(option, val) < 0)
        return -3;

    return 0;
}

int kui_option_read_proc(char *buf, char **start, off_t offset,
        int count, int *eof, void *data)
{
    int len = 0;
    int ret = 0;
    struct option_priv_data *pd = (struct option_priv_data *) data;

    int interrupted = down_interruptible(&pd->mutex);
    if (interrupted)
        return -ERESTARTSYS;

    // 	PDEBUGG(MSG_COMMON, "%s: In option read..\n", DEBUG_NAME);

    if (!pd->last_event.error)
    {
        switch (pd->last_event.type)
        {
        case NONE:
            len = snprintf(buf, PAGE_SIZE, "0a");
            break;
        case FEEDBACK: case REPLACE: case CODING:
            len = snprintf(buf, PAGE_SIZE, "1a");
            break;
        default:
            /* TODO:Error handling*/
            break;
        }
    }
    else
    {
        // NOTE & HACK we return the default/current value of buffer size to the user space by err flag, in the case of PRRT_RCVBUF and KERNEL_RCVBUF.
        len = snprintf(buf, PAGE_SIZE, "%da", pd->last_event.error);
    }

    if (len < 0 || len > PAGE_SIZE)
    {
        PERROR("%s: Could not write to proc file\n", DEBUG_NAME);
        goto free_mutex;
    }

    pd->last_event.type = NONE;
    pd->last_event.error = 0;
    *eof = 1;
    ret = len;

free_mutex:
    up(&pd->mutex);

    return ret;
}

int kui_option_write_proc(struct file *file, const char __user *buffer,
        unsigned long count, void *data)
{
    char *kernel_buffer = 0;
    char option[20]; /*this is the option that is set*/
    // NOTE shine : After reviewing, all option value is an integer, in my opinion, we don't need to parse it in a string form, and change the string form back to an integer again.
    // So we just parse the option value in an integer form. And the coresponding code is also changed, e.g.: checkInput(), checkVal() and checkOption(), also this function.
//     char val[20]; /*this is the value of the option that is set*/
    int value = 0;
    int err = count;

    int not_copied = 0;
    struct prrt_socket *msock = 0;
    struct option_priv_data *pd = (struct option_priv_data *) data;

    int interrupted = down_interruptible(&pd->mutex);
    if (interrupted)
    {
        pd->last_event.error = -EINTERR;
        return -ERESTARTSYS;
    }

    PDEBUG(MSG_COMMON, "%s: Option for socket %d called...\n", DEBUG_NAME, pd->socketno);

    kernel_buffer = kmalloc(count + 1, GFP_KERNEL);
    if (!kernel_buffer)
    {
        pd->last_event.error = -EALLOC;
        err = -ENOMEM;
        goto done;
    }

    /*TODO: what to do if something was not copied*/
    not_copied = copy_from_user(kernel_buffer, buffer, count);
    kernel_buffer[count] = '\0';

    err = checkInput(kernel_buffer);
    if (err < 0)
    {
        PERROR("%s: Malformed input for option: %s\nError no is %d\n", DEBUG_NAME,
                kernel_buffer, err);
        pd->last_event.error = -EINVINP;
        err = -EINVINP;
        goto done;
    }
    /*from this point on you can assume that the input is well formed*/

//     sscanf(kernel_buffer, "%20s %20s", option, val);
    sscanf(kernel_buffer, "%20s %d", option, &value);
//     PDEBUG(MSG_COMMON, "%s: Option: %s\n", DEBUG_NAME, option);
//     PDEBUG(MSG_COMMON, "%s: Value: %s\n", DEBUG_NAME, val);
//     PDEBUG(MSG_COMMON, "%s: Value: %d\n", DEBUG_NAME, value);

    err = sm_get_prrt_sock(pd->socketno, &msock);
    if (err < 0)
    {
        pd->last_event.error = -ENOSOCK;
        err = -ENOSOCK;
        goto done;
    }


    if (strncmp("feedback", option, 8) == 0)
    {
        pd->last_event.type = FEEDBACK;
//         sscanf(val, "%d", &value);
        PDEBUGG(MSG_COMMON, "%s: Value %d\n", DEBUG_NAME, value);
        if (value)
        {
            PDEBUG(MSG_COMMON, "%s: Feedback is now turned on...\n", DEBUG_NAME);
        }
        else
        {
            PDEBUG(MSG_COMMON, "%s: Feedback is now turned off...\n", DEBUG_NAME);
        }

    }
    else if (strncmp("replace", option, 6) == 0)
    {
        pd->last_event.type = REPLACE;
//         sscanf(val, "%d", &value);
        if (value)
        {
            PDEBUG(MSG_COMMON, "%s: Replacements are now turned on...\n", DEBUG_NAME);
        }
        else
        {
            PDEBUG(MSG_COMMON, "%s: Replacements are now turned off...\n", DEBUG_NAME);
        }

    }
    else if (strncmp("coding", option, 5) == 0)
    {
        pd->last_event.type = CODING;
//         sscanf(val, "%d", &value);
        if (value)
        {
            PDEBUG(MSG_COMMON, "%s: Coding is now turned on...\n", DEBUG_NAME);
        }
        else
        {
            PDEBUG(MSG_COMMON, "%s: Coding is now turned off...\n", DEBUG_NAME);
        }

    }
    else if (strncmp("timeout", option, 7) == 0)
    {
        pd->last_event.type = TIMEOUT;
//         sscanf(val, "%d", &value);
        if (value)
        {
            PDEBUG(MSG_COMMON, "%s: Timeout is now turned on...\n", DEBUG_NAME);
        }
        else
        {
            PDEBUG(MSG_COMMON, "%s: Timeout is now turned off...\n", DEBUG_NAME);
        }

    }
    else if (strncmp("prrt_rcvbuf", option, 11) == 0)
    {
        pd->last_event.type = PRRT_RCVBUF;
        PDEBUG(MSG_COMMON, "%s: rcv_bufsize of PRRT receive socket is set to %d.\n", DEBUG_NAME, value);
    }
    else if (strncmp("kernel_rcvbuf", option, 13) == 0)
    {
        pd->last_event.type = KERNEL_RCVBUF;
        PDEBUG(MSG_COMMON, "%s: rcv_bufsize in the kernel space of KUI is set to %d.\n", DEBUG_NAME, value);
    }
    else if (strncmp("user_rcvbuf", option, 11) == 0)
    {
        pd->last_event.type = USER_RCVBUF;
        PDEBUG(MSG_COMMON, "%s: rcv_bufsize in the user space of KUI is set to %d.\n", DEBUG_NAME, value);
    }
    else if (strncmp("trafficshaping", option, 14) == 0)
    {
        pd->last_event.type = TRAFFICSHAPING;
        PDEBUG(MSG_COMMON, "%s: datarate of traffic shaping in the KUI is set to %d.\n", DEBUG_NAME, value);
    }
    else if (strncmp("fb_supp_on", option, 10) == 0)
    {
        pd->last_event.type = FB_SUPP_ON;
        PDEBUG(MSG_COMMON, "%s: max feedback sent delay in feedback suppression in the KUI is set to %d.\n", DEBUG_NAME, value);
    }
    else if (strncmp("fb_supp_off", option, 11) == 0)
    {
        /* Receiving feedback thread is still running. but we set delay to 0. */
        // TODO we may just stop the thread.
        pd->last_event.type = FB_SUPP_ON;
        value = 0;
        PDEBUG(MSG_COMMON, "%s: max feedback sent delay in feedback suppression in the KUI is set to %d.\n", DEBUG_NAME, value);
    }
    else
    {
        PERROR("%s: The option %s was not recognized\n", DEBUG_NAME, option);
        pd->last_event.error = -EINVINP;
        err = -EINVINP;
        goto done;
    }

    err = prrt_set_option(msock, pd->last_event.type, value);
    pd->last_event.error = err; // NOTE & HACK we return the default/current value of buffer size to the user space by err flag, in the case of PRRT_RCVBUF and KERNEL_RCVBUF.
    if (err < 0)
    {
        pd->last_event.error = -EGENERAL;
        err = -EGENERAL;
    }

done:
    up(&pd->mutex);

    // out:
    kfree(kernel_buffer);
    if (err == 0)
        err = count;
    PDEBUGG(MSG_COMMON, "%s: Returning %d from option call\n", DEBUG_NAME, err);
    return err;
}
