/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "kernel_user_interface.h"
#include "kui_param.h"

static const char *DEBUG_NAME = "KUI_PARAM";

struct param_priv_data
{
    Socketno_t socketno; /*socketno of the read file*/
    struct semaphore mutex; /*mutex to synchronize access
						 *to one file*/
    enum Param_t param;
};

/*from socket_manager.c*/
int sm_get_prrt_sock(Socketno_t socketno, struct prrt_socket **msock);

/*from below*/
int kui_param_read_proc(char *buf, char **start, off_t offset,
        int count, int *eof, void *data);
int kui_param_write_proc(struct file *file, const char __user *buffer,
        unsigned long count, void *data);

/*
 * This function creates a new file in the procfs for controling a socket.
 */
void *kui_create_param_file(Socketno_t socketno, enum Param_t param)
{
    char name[NAME_SIZE];
    struct proc_dir_entry *procentry;
    struct param_priv_data *priv_data = 0;

    priv_data = kmalloc(sizeof (*priv_data), GFP_KERNEL);

    /*create string for parameter*/
    switch (param)
    {
    case PLR:
        snprintf(name, NAME_SIZE, "prrt/%d/plr", socketno);
        priv_data->param = param;
        break;
    case AHEC:
        snprintf(name, NAME_SIZE, "prrt/%d/ahec", socketno);
        priv_data->param = param;
        break;
    case BWEST:
        snprintf(name, NAME_SIZE, "prrt/%d/bwest", socketno);
        priv_data->param = param;
        break;
    case BUFFB:
        snprintf(name, NAME_SIZE, "prrt/%d/buffb", socketno);
        priv_data->param = param;
        break;

    default:
        kfree(priv_data);
        PERROR("%s: Parameter not recognized\n", DEBUG_NAME);
        return 0;
        break;
    }

    priv_data->socketno = socketno;
    init_MUTEX(&(priv_data->mutex));

    procentry = create_proc_entry(name, S_IRUGO | S_IWUGO, NULL);
    if (procentry)
    {
        procentry->read_proc = kui_param_read_proc;
        procentry->write_proc = kui_param_write_proc;
        procentry->data = priv_data;

        return priv_data; /*SUCCESS*/
    }

    return 0; /*FAILURE*/
}

/*
 * Function for deleting a procfs control file.
 */
void kui_delete_param_file(Socketno_t socketno, enum Param_t param)
{
    char name[NAME_SIZE];
    struct prrt_socket *msock = 0;

    while (sm_get_prrt_sock(socketno, &msock) == -ERESTARTSYS)
        ;
    if (!msock)
        return;

    switch (param)
    {
    case PLR:
        snprintf(name, NAME_SIZE, "prrt/%d/plr", socketno);
        kfree(msock->plr_pd);
        break;
    case AHEC:
        snprintf(name, NAME_SIZE, "prrt/%d/ahec", socketno);
        kfree(msock->ahec_pd);
        break;
    case BWEST:
        snprintf(name, NAME_SIZE, "prrt/%d/bwest", socketno);
        kfree(msock->bwest_pd);
        break;
    case BUFFB:
        snprintf(name, NAME_SIZE, "prrt/%d/buffb", socketno);
        kfree(msock->buffb_pd);
        break;

    default:
        PERROR("%s: Parameter not recognized\n", DEBUG_NAME);
        return;
        break;
    }

    remove_proc_entry(name, NULL);
}

int kui_param_read_proc(char *buf, char **start, off_t offset,
        int count, int *eof, void *data)
{
    int ret = 0;
    int len = 0;
    struct param_priv_data *pd = 0;
    int interrupted = 0;
    struct prrt_socket *msock = 0;
    int val = 0;

    pd = (struct param_priv_data *) data;

    interrupted = down_interruptible(&pd->mutex);
    if (interrupted)
        return -ERESTARTSYS;

    if (sm_get_prrt_sock(pd->socketno, &msock) == -ERESTARTSYS)
        return -ERESTARTSYS;
    if (!msock)
        goto free_mutex;

    switch (pd->param)
    {
    case PLR:
        //val = prrtlpar_get_PLR_target(&msock->loc_params_out);
        val = prrtcpar_get_PLR_target(&msock->cod_params);
        len = snprintf(buf, PAGE_SIZE, "%da", val);
        break;
    case AHEC:
        // TODO we may retrieve the AHEC parameters in a string form.
        
        break;
    case BWEST:
        //val = prrtlpar_get_BW_est(&msock->loc_params_out);
        val = prrtcpar_get_BW_est(&msock->cod_params);
        len = snprintf(buf, PAGE_SIZE, "%da", val);
        break;
    case BUFFB:
        val = prrtlpar_get_BUF_fb(&msock->loc_params_in);
        len = snprintf(buf, PAGE_SIZE, "%da", val);
        break;
    default:
        /*should not happen*/
        goto free_mutex;
        break;
    }

    if (len < 0 || len > PAGE_SIZE)
    {
        PERROR("%s: Could not write to proc file\n", DEBUG_NAME);
        goto free_mutex;
    }

    *eof = 1;
    ret = len;

free_mutex:
    up(&pd->mutex);

    return ret;
}

static int CheckInputValue(enum Param_t paramType, char *input);
static int GetMultipleIntegers(char *input, int numberInts, unsigned *numbers);
void SetFixedTReplace(struct prrt_socket *msock, int n_rr);

int kui_param_write_proc(struct file *file, const char __user *buffer,
        unsigned long count, void *data)
{
    char *kernel_buffer = 0;
    int not_copied = 0;
    struct prrt_socket *msock = 0;
    int temp = 0;
    int *temp_arr = 0;
    int interrupted = 0;
    struct param_priv_data *pd = 0;
    int err = 0;
    int n_cc = 0;
    int n_p = 0;
    int *n_ccs = 0;

    pd = (struct param_priv_data *) data;

    interrupted = down_interruptible(&pd->mutex);
    if (interrupted)
        return -ERESTARTSYS;

    PDEBUG(MSG_KUIPARAM, "%s: Param change for socket %d called...\n",
            DEBUG_NAME, pd->socketno);

    kernel_buffer = kmalloc(count + 1, GFP_KERNEL);
    if (!kernel_buffer)
    {
        goto nomem;
    }
    kernel_buffer[count] = '\0';

    /*TODO: what to do if something was not copied*/
    not_copied = copy_from_user(kernel_buffer, buffer, count);

    PDEBUG(MSG_KUIPARAM, "%s: Kernel buffer for parameter value: %s (size: %lu)\n", DEBUG_NAME,
            kernel_buffer, count);

    if (sm_get_prrt_sock(pd->socketno, &msock) == -ERESTARTSYS)
    {
        kfree(kernel_buffer);
        return -ERESTARTSYS;
    }
    if (!msock)
    {
        /*ERROR: socket not found*/
        goto error;
    }

    if (CheckInputValue(pd->param, kernel_buffer) == -1)
        goto error;

    switch (pd->param)
    {
    case PLR:
        sscanf(kernel_buffer, "%d", &temp);
        //prrtlpar_set_PLR_target(&msock->loc_params_out, temp);
        prrtcpar_set_PLR_target(&msock->cod_params, temp);
        prrtlpar_set_PLR_target(&msock->loc_params_in, temp);
        PDEBUG(MSG_KUIPARAM, "%s: Set param PLR to %d\n", DEBUG_NAME, temp);
        break;

    case AHEC :
        // TODO parse a string from the user space and set them.
        // Parameters in order: #parameters d_target, d_fec, d_c, d_req, d_fb, k, n_p. (example : 10 200 20 2 40 0 1 1 1 3)
        // D_FEC
        sscanf(kernel_buffer, "%d", &temp);
        PDEBUG(MSG_KUIPARAM, "%s: There are %d numbers in AHEC parameters\n", DEBUG_NAME, temp);

        // AHEC parameters.
        temp_arr = kmalloc(sizeof (*temp_arr) * temp, GFP_KERNEL);
        if (!temp_arr)
            goto error;

        GetMultipleIntegers(kernel_buffer, temp, temp_arr);

        n_cc = temp - 7;   // The number of cycles of parity packets.
        n_p = n_cc > 0 ? temp_arr[7] : 0;
        n_cc--;
        if (n_cc < 0) {
          n_cc = 0;
        }
        n_ccs = n_cc > 0 ? temp_arr + 8 : 0;

        err = prrtcpar_set_ahec_params(&msock->cod_params, temp_arr[1], temp_arr[2], temp_arr[3], temp_arr[4], temp_arr[5], temp_arr[6], n_p, n_cc, n_ccs);
        if (err < 0)
        {
            kfree(temp_arr);
            goto error;
        }

        kfree(temp_arr);

        //PDEBUG(MSG_KUIPARAM, "%s: There are %d numbers in AHEC parameters, set is done with err=%d\n", DEBUG_NAME, temp, err);
        break;

    case BWEST:
        sscanf(kernel_buffer, "%d", &temp);
	      prrtlpar_set_BW_est(&msock->loc_params_in, temp);
        PDEBUG(MSG_KUIPARAM, "%s: Set param bwest to %d\n", DEBUG_NAME, temp);
        break;

    case BUFFB:
        sscanf(kernel_buffer, "%d", &temp);
	      prrtlpar_set_BUF_fb(&msock->loc_params_in, temp);
        PDEBUG(MSG_KUIPARAM, "%s: Set param buffb to %d\n", DEBUG_NAME, temp);
        break;

    default:
        PERROR("%s: Parameter not recognized\n", DEBUG_NAME);
        break;
    }

error:
    kfree(kernel_buffer);
    up(&pd->mutex);
    return count;

nomem:
    up(&pd->mutex);
    return -ENOMEM;
}

static int CheckSingleUnsigned(char *input);
static int CheckRate(char *input);
static int CheckAHEC(char *input);

static int CheckInputValue(enum Param_t paramType, char *input)
{
    switch (paramType)
    {
    case BUFFB:
        return CheckSingleUnsigned(input);
        break;
    case BWEST:
        return CheckSingleUnsigned(input);
        break;
    case PLR:
        return CheckRate(input);
        break;
    case AHEC :
        return CheckAHEC(input);
        break;
    default:
        PERROR("%s: No valid parameter file\n", DEBUG_NAME);
        return -1;
        break;
    }
}

static int CheckSingleUnsigned(char *input)
{
    int argsRead = 0;
    unsigned i = 0;
    char s[2];

    argsRead = sscanf(input, "%u %1s", &i, s);
    if (argsRead != 1)
    {
        PERROR("%s: No integer or more than an integer found for param\n", DEBUG_NAME);
        return -1;
    }

    return 0;
}

static int CheckRate(char *input)
{
    int err = 0;
    unsigned rate = 0;

    err = CheckSingleUnsigned(input);
    if (err < 0)
        return -1;

    sscanf(input, "%u", &rate);

    if (rate < 0 || rate > 100)
    {
        PERROR("%s: A rate must lie between 0 and 100\n", DEBUG_NAME);
        return -1;
    }

    return 0;
}


static int CheckMultipleUnsigneds(char *input, int numberInts)
{
    int err = 0;
    unsigned *numbers = kmalloc(sizeof (*numbers) * numberInts, GFP_KERNEL);
    if (!numbers)
    {
        PERROR("%s: Not enough memory for allocating numbers array\n", DEBUG_NAME);
        return -ENOMEM;
    }

    err = GetMultipleIntegers(input, numberInts, numbers);

    kfree(numbers);

    return err;
}


static int CheckAHEC(char *input)
{
    int argsRead = 0;
    int cnt = 0;
    int err = 0;

    PDEBUG(MSG_KUIPARAM, "%s: Parsing input: %s\n", DEBUG_NAME, input);

    argsRead = sscanf(input, "%d", &cnt);
    if (argsRead != 1)
    {
        PERROR("%s: Invalid input for number of AHEC parameters(%d args)\n",
                DEBUG_NAME, argsRead);
        return -1;
    }

    PDEBUG(MSG_KUIPARAM, "%s: %d numbers found\n", DEBUG_NAME, cnt);

    err = CheckMultipleUnsigneds(input, cnt);
    if (err < 0)
    {
        PERROR("%s: Could not parse AHEC parameters\n", DEBUG_NAME);
    }

    return err;
}

static int GetMultipleIntegers(char *input, int count, unsigned *numbers)
{
    int argsRead = 0;
    int i = 0;
    int numbersRead = 0;

    PDEBUG(MSG_KUIPARAM, "%s: Searching %d numbers in: %s\n", DEBUG_NAME, count, input);

    for (i = 0; i < strlen(input); ++i)
    {

        if (input[i] != ' ' && input[i] != '0' && input[i] != '1' && input[i] != '2' &&
                input[i] != '3' && input[i] != '4' && input[i] != '5' && input[i] != '6' &&
                input[i] != '7' && input[i] != '8' && input[i] != '9')
        {
            break;
        }
        else if (input[i] == ' ')
        {
            if (numbersRead >= count)
            {
                PERROR("%s: Found too many numbers\n", DEBUG_NAME);
                return -1;
            }
        }

        if (i == 0 || input[i] == ' ')
        {
            argsRead = sscanf(input + i, "%u", &(numbers[numbersRead++]));
            if (argsRead != 1)
            {
                PERROR("%s: Could not read in number\n", DEBUG_NAME);
                return -1;
            }
        }
    }

    if (numbersRead < count)
    {
        PERROR("%s: Found not enough numbers\n", DEBUG_NAME);
        return -1;
    }

    return 0;
}

void SetFixedTReplace(struct prrt_socket *msock, int n_rr)
{
    int t_replace = 0;

    //t_replace = prrtlpar_get_D_target(&msock->loc_params_out) / n_rr;
    t_replace = prrtcpar_get_D_target(&msock->cod_params) / n_rr;

    prrtrlist_setWorstTReplace(&msock->out_packet_list.recv_list, t_replace);
    PDEBUG(MSG_KUIPARAM, "%s: Set fixed t_replace to %d\n", DEBUG_NAME, t_replace);
}

