/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "kernel_user_interface.h"
#include "../prrt_socket.h"

static const char *DEBUG_NAME = "KUI_READ";

/*from socket_manager.c*/
int sm_get_prrt_sock(Socketno_t socketno, struct prrt_socket **msock);

/*from below*/
static int kui_read_open(struct inode *inode, struct file *file);
static int kui_read_release(struct inode *inode, struct file *file);

struct read_priv_data {
	Socketno_t socketno;	/*socketno of the read file*/
};

/*
 * Function to create a read file.
 */
int kui_create_read_file(Socketno_t socketno)
{
	char name[NAME_SIZE];
	struct proc_dir_entry *procentry;

	static struct file_operations read_file_ops = {
		.owner   = THIS_MODULE,
		.open    = kui_read_open,
		.read    = seq_read,
		.llseek  = seq_lseek,
		.release = kui_read_release
	};

	snprintf(name, NAME_SIZE, "prrt/%u/read", socketno);

	procentry = create_proc_entry(name, 0, NULL);
	if (procentry) {
		procentry->proc_fops = &read_file_ops;

		return 1;	/*SUCCESS*/
	}

	return 0;	/*FAILURE*/
}

/*
 * Function for deleting a procfs read file.
 */
void kui_delete_read_file(Socketno_t socketno)
{
	char name[NAME_SIZE];

	snprintf(name, NAME_SIZE, "prrt/%d/read", socketno);

	remove_proc_entry(name, NULL);
}

static int WriteBinaryDataToSeqfile(struct seq_file *seq, const void *data, size_t len);

static int kui_read_show(struct seq_file *sfile, void *v)
{
	struct prrt_socket *msock = 0;
	int bytes_recv = 0;
	int err = 0;
	struct read_priv_data *priv = (struct read_priv_data *)sfile->private;
	struct prrt_packet_container *pcont = 0;

	PDEBUG(MSG_COMMONRECV, "%s: Show called!\n", DEBUG_NAME);

	while(sm_get_prrt_sock(priv->socketno, &msock) == -ERESTARTSYS)
		;
	if(!msock)
		goto out;

	bytes_recv = prrt_recv(msock, &pcont);
	if(bytes_recv < 0) {
		PERROR("%s: Show failed with error %d!\n", DEBUG_NAME, bytes_recv);
		err = bytes_recv;
		goto out;
	}

	PDEBUG(MSG_PRRTRECV, "%s: Returning %d bytes of data!\n", DEBUG_NAME, bytes_recv);
	if(WriteBinaryDataToSeqfile(sfile, prrt_packet_get_data(pcont), bytes_recv) < 0) {
		PERROR("%s: Could not print everything to seq_file\n", DEBUG_NAME);
	}

	prrt_packet_destroy(&pcont);

out:
	return err;
}

static int WriteBinaryDataToSeqfile(struct seq_file *seq, const void *data, size_t len)
{
	if (seq->count + len < seq->size) {
		memcpy(seq->buf + seq->count, data, len);
		seq->count += len;
		return 0;
	}

	seq->count = seq->size;
	return -1;
}

/*open a proc file for reading from sockets*/
static int kui_read_open(struct inode *inode, struct file *file)
{
	int ret = 0;
	struct read_priv_data *priv;

	/*allocate space for private data*/
	priv = kmalloc(sizeof(*priv), GFP_KERNEL);
	if(!priv)
		/*ERROR: no memory*/
		goto nomem;

	/*open seq_file for single show*/
	ret = single_open(file, kui_read_show, (void*)priv);

	/*find out socketno of the read file and initialise private data*/
	priv->socketno = (Socketno_t)simple_strtoul(file->f_path.dentry->d_parent->d_iname, NULL, 10);

	PDEBUG(MSG_COMMONRECV, "%s: Opening read file, socketno=%u\n", DEBUG_NAME, priv->socketno);

	return ret;

nomem:
	return -ENOMEM;
}

static int kui_read_release(struct inode *inode, struct file *file)
{
	struct seq_file *sfile = (struct seq_file *)file->private_data;

	kfree(sfile->private);
	sfile->private = 0;
	return single_release(inode, file);
}
