/*
 * (c) 2010 Saarland University, Telecommunications Lab
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 +
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 +
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 +
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 +
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 +
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device,
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "kernel_user_interface.h"
#include "kui_error.h"
#include <linux/in.h>

static const char *DEBUG_NAME = "KUI_RELAY";

struct relay_state {
    struct semaphore mutex;     /*mutex to lock access to the read file*/
    unsigned char mutex_init;   /*is the mutex initialised*/
};

static struct relay_state state;

//from relay_manager.c
int rm_get_all_relays(struct prrt_relay ***mrelays);

// from below
static int kui_relay_read_proc(char *buf, char **start, off_t offset,
        int count, int *eof, void *data);
static void BuildRelaysString(struct prrt_relay **mrelays, int relay_count, char *buffer);

/*
 * This function creates a new file in the procfs for retrieving relay information.
 */
int kui_create_relay_file(void) {
    struct proc_dir_entry *procentry;

    if (!state.mutex_init) {
        init_MUTEX(&state.mutex);
        state.mutex_init = 1;
        PDEBUG(MSG_COMMON, "%s: semaphore initialised\n", DEBUG_NAME);
    }

    procentry = create_proc_entry("prrt/relays", S_IRUGO, NULL);
    if (procentry) {
        procentry->read_proc = kui_relay_read_proc;
        procentry->data = NULL;

        return 1; /*SUCCESS*/
    }

    PERROR("%s: failure when creating prrt/relays\n", DEBUG_NAME);

    return 0; /*FAILURE*/
}

/*
 * Function for deleting a procfs relay file.
 */
void kui_delete_relay_file(void) {
    remove_proc_entry("prrt/relays", NULL);
}

/*
 * Read method for a relay file.
 * Returns infos about currently active relays.
 */
static int kui_relay_read_proc(char *buf, char **start, off_t offset,
        int count, int *eof, void *data) {

    int len = 0;
    int ret = 0;
    int relay_count = 0;
    struct prrt_relay **mrelays = 0;
    //char buffer[PAGE_SIZE];
    char *buffer = 0;
    int interrupted = 0;
    
    /* Sovled warning of larger than 1024 bytes in frame. */
    buffer = (char *) kmalloc(PAGE_SIZE, GFP_KERNEL);
    
    interrupted = down_interruptible(&state.mutex);
    if (interrupted)
        return -ERESTARTSYS;

    *eof = 1;

    relay_count = rm_get_all_relays(&mrelays);
    if (relay_count == 0) {
        goto free_relays;
    }

    BuildRelaysString(mrelays, relay_count, buffer);

    len = snprintf(buf, PAGE_SIZE, "%s", buffer);

    if (len < 0 || len > PAGE_SIZE) {
        PERROR("%s: Could not write to proc file\n", DEBUG_NAME);
        goto free_relays;
    }

    ret = len;

free_relays:
    kfree(mrelays);
    kfree(buffer);
    up(&state.mutex);

    return ret;
}

char *in_ntoa(__u32 in, char *buff) {
    char *p;

    p = (char *) &in;
    sprintf(buff, "%d.%d.%d.%d",
            (p[0] & 255), (p[1] & 255), (p[2] & 255), (p[3] & 255));
    return (buff);
}

static void BuildRelaysString(struct prrt_relay **mrelays, int relay_count, char *buffer) {
    int charsWritten = 0;
    int i = 0;
    struct prrt_relay *mrelay = NULL;
    struct prrt_socket *msock_recv = NULL;
    struct prrt_socket *msock_send = NULL;
    char addr1[32];
    char addr2[32];

    for (i=0; i<relay_count; i++){
        mrelay = mrelays[i];
        msock_recv = mrelay->msock_in;
        msock_send = mrelay->msock_out;

	      charsWritten += sprintf(buffer+charsWritten, "[%lu]%s: receiversocket %lu, sendersocket %lu, originIp %s, targetIp %s\n",
                                (unsigned long)mrelay->relayno, mrelay->relayname, (unsigned long)msock_recv->socketno,(unsigned long) msock_send->socketno,
                in_ntoa( msock_recv->peerfb->sin_addr.s_addr, addr1 ),
                in_ntoa( msock_send->peer->sin_addr.s_addr, addr2) );
    }
}
