/*
 * (c) 2010 Saarland University, Telecommunications Lab
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 +
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 +
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 +
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 +
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 +
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device,
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "relay_manager.h"
#include "kernel_user_interface.h"
#include "../prrt_relay.h"

#define SOCKET_STRING_MEM 50	/*this is the memory allocated for printing an entry in the relay list*/

static const char *DEBUG_NAME = "RM";

struct relay_entry {
    struct list_head list;
    Relayno_t relayno;
    struct prrt_relay *mrelay;
};

static struct semaphore mutex; /*this mutex controls the access to the relay_list*/
static Socketno_t next_relayno = 1; /*counter for relay numbers*/
static struct kmem_cache *relay_entry_cache; /*this lookaside cache contains objects for relay_list entries*/
static struct list_head relay_list; /*list containing the relays*/
static int relay_count = 0;         /*the number of relays in the relay_list*/

void rm_init(void) {
    init_MUTEX(&mutex);
    INIT_LIST_HEAD(&relay_list);
    relay_entry_cache = kmem_cache_create("relay_entry_cache",
            sizeof (struct relay_entry), 0, 0,
            NULL);

    if (!relay_entry_cache)
        PERROR("%s: Could not create socket cache\n", DEBUG_NAME);
}

void rm_uninit(void) {
    /*TODO: assure all sockets are deleted, else memory leak*/
    kmem_cache_destroy(relay_entry_cache);
}

/**
 * Register a relay in the list.
 * @return Relaynumber(>0) on success, <0 on failure.
 */
Relayno_t rm_register_relay(char *relayname) {
    int err = 0;
    struct relay_entry *new_relay = kmem_cache_alloc(relay_entry_cache, GFP_KERNEL);

    if (down_interruptible(&mutex))
        return -ERESTARTSYS; /*this means caller has to try again*/

    /*new socket was registered by user*/
    err = prrt_relay(&new_relay->mrelay, relayname);

    if (err == -ENOMEM) {
        PDEBUG(MSG_COMMON, "ERROR %s: No mem to create new relay\n",
                DEBUG_NAME);
        goto nomem;
    } else if (err < 0) {
        PDEBUG(MSG_COMMON, "ERROR %s: Relay creation failed\n",
                DEBUG_NAME);
        goto nomem;
    }

    /*TODO: number wrapping and exclusion of zero*/
    new_relay->mrelay->relayno = next_relayno++;

    list_add_tail(&(new_relay->list), &relay_list);
    relay_count++;

    PDEBUG(MSG_COMMON, "%s: New relay created, no %d\n", DEBUG_NAME,
            new_relay->mrelay->relayno);

    up(&mutex);

    return new_relay->mrelay->relayno;

nomem:
    kmem_cache_free(relay_entry_cache, new_relay);
    up(&mutex);
    return -ENOMEM;
}

/**
 * Try to unregister a relay from the list.
 * @param relayno The relay number that should be unregistered.
 */
int rm_unregister_relay(Relayno_t relayno) {
    struct list_head *iter = NULL;
    struct relay_entry *entry = NULL;

    if (down_interruptible(&mutex))
        return -ERESTARTSYS;

    if (list_empty(&relay_list))
        return 0;

    list_for_each(iter, &relay_list) {
        entry = list_entry(iter, struct relay_entry, list);

        if (entry->mrelay->relayno == relayno)
            break;
    }

    if (entry->mrelay->relayno == relayno) {
        /*relay was found*/
        PDEBUG(MSG_COMMON, "%s: Deleting relay no %d\n", DEBUG_NAME, entry->mrelay->relayno);

        prrt_shutdown_relay(entry->mrelay);

        relay_count--;
        list_del(iter);
        kmem_cache_free(relay_entry_cache, entry);
    } else {
        PDEBUG(MSG_COMMON, "%s: Could not find relay no %d\n", DEBUG_NAME, entry->mrelay->relayno);
    }

    up(&mutex);

    return 0;
}

/**
 * Get a specific relay from the list.
 * @param relayno The number of the relay to search for.
 * @param mrelay The relay that was found on success.
 * @return 0 on success, <0 on failure.
 */
int rm_get_prrt_relay(Relayno_t relayno, struct prrt_relay **mrelay) {
    struct list_head *iter = NULL;
    struct relay_entry *entry = NULL;

    *mrelay = 0;

    if (down_interruptible(&mutex))
        return -ERESTARTSYS;

    list_for_each(iter, &relay_list) {
        entry = list_entry(iter, struct relay_entry, list);

        // 		PDEBUG(MSG_COMMON, "%s: %d == %d ?\n", DEBUG_NAME, entry->socketno, socketno);

        if (entry->mrelay->relayno == relayno)
            break;
    }

    // 	PDEBUG(MSG_COMMON, "%s: %d == %d ?\n", DEBUG_NAME, (unsigned int)iter, (unsigned int)&socket_list);

    if (iter != &relay_list) {
        *mrelay = entry->mrelay;
    }

    up(&mutex);

    return 0;
}

/**
 * Socket should no longer be used by any relay.
 * @param msock socket that becomes unusable.
 */
void rm_invalidate_socket_in_relays(struct prrt_socket *msock) {
    struct list_head *iter = NULL;
    struct relay_entry *entry = NULL;
    struct prrt_relay *mrelay = NULL;

    list_for_each(iter, &relay_list) {
        entry = list_entry(iter, struct relay_entry, list);
        mrelay = entry->mrelay;

        down(&mrelay->socket_lock);
        if(mrelay->msock_in == msock){
            mrelay->msock_in = NULL;
        }
        if(mrelay->msock_out == msock){
            mrelay->msock_out = NULL;
        }
        up(&mrelay->socket_lock);
    }
}

int rm_get_all_relays(struct prrt_relay ***mrelays){
    int i = 0;
    struct list_head *iter = NULL;
    struct relay_entry *entry = NULL;
    struct prrt_relay *mrelay = NULL;

    if(relay_count == 0)
        return 0;

    *mrelays = kmalloc(sizeof(*mrelays)*relay_count, GFP_KERNEL);

    list_for_each(iter, &relay_list) {
        if(i>=relay_count){
            PERROR("%s: More entries in relay_list than relay_count\n", DEBUG_NAME);
            break;
        }

        entry = list_entry(iter, struct relay_entry, list);
        mrelay = entry->mrelay;

        (*mrelays)[i] = mrelay;
        i++;
    }

    return i;
}

Relayno_t rm_get_relayno(void)
{
	struct relay_entry *entry;
	Relayno_t ret = 0;

	if(down_interruptible(&mutex))
		return -ERESTARTSYS;

	if(!list_empty(&relay_list)) {
		entry = list_entry(relay_list.next, struct relay_entry, list);
		ret = entry->mrelay->relayno;
	}

	up(&mutex);

	return ret;
}
