/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "socket_manager.h"
#include "kernel_user_interface.h"
#include "../prrt_socket.h"

#define SOCKET_STRING_MEM 50	/*this is the memory allocated for printing an entry in the socket list*/

/*included from relay_manager*/
void rm_invalidate_socket_in_relays(struct prrt_socket *msock);

static const char *DEBUG_NAME = "SM";

struct socket_entry {
	struct list_head list;
	Socketno_t socketno;
	struct prrt_socket *msock;
};

static struct semaphore mutex;			/*this mutex controls the access to the socket_list*/
static Socketno_t next_socketno = 1;		/*counter for socket numbers*/
static struct kmem_cache *socket_entry_cache;	/*this lookaside cache contains objects for socket_list entries*/
static struct list_head socket_list;		/*list containing the sockets*/

void sm_init(void)
{
	init_MUTEX(&mutex);
	INIT_LIST_HEAD(&socket_list);
	socket_entry_cache = kmem_cache_create("socket_entry_cache",
					sizeof(struct socket_entry), 0, 0,
					NULL);

	if(!socket_entry_cache)
		PERROR("%s: Could not create socket cache\n", DEBUG_NAME);
}

void sm_uninit(void)
{
	/*TODO: assure all sockets are deleted, else memory leak*/
	kmem_cache_destroy(socket_entry_cache);
}

/*
 * flags: 0 normal registering
 *        1 when accepting a new socket (DOES NOT EXIST ANYMORE)
 * msock: 0 (not used) when creating new socket
 *        pointer to the prrt socket that accepts the new socket
 */
Socketno_t sm_register_socket(int flags, struct prrt_socket *msock, int kui_port)
{
	int err = 0;
	struct socket_entry *new_socket = kmem_cache_alloc(socket_entry_cache, GFP_KERNEL);

	if(down_interruptible(&mutex))
		return -ERESTARTSYS;	/*this means caller has to try again*/

//	struct list_head *iter = NULL;
//	struct socket_entry *entry = NULL;
//
//	list_for_each(iter, &socket_list) {
//		entry = list_entry(iter, struct socket_entry, list);
//
//
//		if((entry->msock->host->sin_port == msock->host->sin_port &&
//				entry->msock->host->sin_addr.s_addr == msock->host->sin_addr.s_addr)||
//(entry->msock->peer->sin_port == msock->peer->sin_port &&
//				entry->msock->peer->sin_addr.s_addr == msock->peer->sin_addr.s_addr)){
//			PDEBUG(MSG_COMMON, "ERROR %s: Socket exists\n", DEBUG_NAME);
//			//goto nomem;
//break;
//		}
//	}

	/*TODO: mem management*/
	if(flags == 0) {
		/*new socket was registered by user*/
		err=prrt_socket(&new_socket->msock, kui_port);
	} else {
		/*NO SUCH ACTION: accepting a new socket*/
// 		err=prrt_accept(msock, &(new_socket->msock));
	}

	if(err == -ENOMEM) {
		PDEBUG(MSG_COMMON, "ERROR %s: No mem to create new socket\n",
			DEBUG_NAME);
		goto nomem;
	} else if (err < 0) {
		PDEBUG(MSG_COMMON, "ERROR %s: Socket creation failed\n",
			DEBUG_NAME);
		goto nomem;
	}

	/*TODO: number wrapping and exclusion of zero*/
	new_socket->msock->socketno = next_socketno++;

	list_add_tail(&(new_socket->list), &socket_list);

	PDEBUG(MSG_COMMON, "%s: New socket created, no %d\n", DEBUG_NAME,
		new_socket->msock->socketno);

	up(&mutex);

	return new_socket->msock->socketno;

nomem:
	kmem_cache_free(socket_entry_cache, new_socket);
	up(&mutex);
	return -ENOMEM;
}

int sm_unregister_socket(Socketno_t socketno)
{
	struct list_head *iter = NULL;
	struct socket_entry *entry = NULL;

	if(down_interruptible(&mutex))
		return -ERESTARTSYS;

	if(list_empty(&socket_list))
		return 0;

	list_for_each(iter, &socket_list) {
		entry = list_entry(iter, struct socket_entry, list);

		if(entry->msock->socketno == socketno)
			break;
	}

	if(entry->msock->socketno == socketno) {
		/*socket was found*/
		PDEBUG(MSG_COMMON, "%s: Deleting socket no %d\n", DEBUG_NAME, entry->msock->socketno);

		prrt_shutdown(entry->msock);

                rm_invalidate_socket_in_relays(entry->msock);

		list_del(iter);
		kmem_cache_free(socket_entry_cache, entry);
	} else {
		PDEBUG(MSG_COMMON, "%s: Could not find socket no %d\n", DEBUG_NAME, entry->msock->socketno);
	}

	up(&mutex);

	return 0;
}

int sm_set_socket_privdata(Socketno_t socketno, void **priv_datas, int len)
{
	struct list_head *iter = NULL;
	struct socket_entry *entry = NULL;

	if(down_interruptible(&mutex))
		return -ERESTARTSYS;

	/*Search for right socket_entry*/
	list_for_each(iter, &socket_list) {
		entry = list_entry(iter, struct socket_entry, list);

		if(entry->msock->socketno == socketno)
			break;
	}

	/*Socket found, set pointer*/
	if(iter != &socket_list) {
		entry->msock->control_pd = priv_datas[0];
		entry->msock->write_pd = priv_datas[1];
		entry->msock->read_pd = priv_datas[2];
		entry->msock->option_pd = priv_datas[3];
		entry->msock->stats_pd = priv_datas[4];
		//entry->msock->d_target_pd = priv_datas[5];
		entry->msock->plr_pd = priv_datas[5];
    //entry->msock->d_fec_pd = priv_datas[7];
		//entry->msock->codp_pd = priv_datas[8];
		//entry->msock->rtr_pd = priv_datas[9];
		entry->msock->bwest_pd = priv_datas[6];
		entry->msock->ahec_pd = priv_datas[7];
		entry->msock->buffb_pd = priv_datas[8];
	}

	up(&mutex);

	return 0;
}

int sm_socket_list_empty(void) {
	int ret = 0;

	if(down_interruptible(&mutex))
		return -ERESTARTSYS;

	if(list_empty(&socket_list))
		ret = 1;

	up(&mutex);

	return ret;
}

Socketno_t sm_get_socketno(void)
{
	struct socket_entry *entry;
	Socketno_t ret = 0;

	if(down_interruptible(&mutex))
		return -ERESTARTSYS;

	if(!list_empty(&socket_list)) {
		entry = list_entry(socket_list.next, struct socket_entry, list);
		ret = entry->msock->socketno;
	}

	up(&mutex);

	return ret;
}

int sm_get_prrt_sock(Socketno_t socketno, struct prrt_socket **msock)
{
	struct list_head *iter = NULL;
	struct socket_entry *entry = NULL;

	*msock = 0;

	if(down_interruptible(&mutex))
		return -ERESTARTSYS;

	list_for_each(iter, &socket_list) {
		entry = list_entry(iter, struct socket_entry, list);

// 		PDEBUG(MSG_COMMON, "%s: %d == %d ?\n", DEBUG_NAME, entry->socketno, socketno);

		if(entry->msock->socketno == socketno)
			break;
	}

// 	PDEBUG(MSG_COMMON, "%s: %d == %d ?\n", DEBUG_NAME, (unsigned int)iter, (unsigned int)&socket_list);

	if(iter != &socket_list) {
		*msock = entry->msock;
	}

	up(&mutex);

	return 0;
}

#ifdef DEBUG
/**
 * Creates a string containing information about all sockets.
 * \param list the list of all sockets
 * \param sinfo the address of the string containing the information about the sockets
 *		has to be freed by the caller
 */
int sm_get_socket_info(char **sinfo)
{
	struct list_head *iter;
	struct socket_entry *entry;
	int 	count = 0,
		written = 0,
		alloc_size = 0;
	char* info = NULL;	/*string where the information is saved*/
	char socketinfo[20];
	int ret = 0;

	PDEBUG(MSG_COMMON, "%s: Counting members in socket list\n", DEBUG_NAME);

	if(down_interruptible(&mutex))
		return -ERESTARTSYS;

	/*count elements to determine memory need*/
	list_for_each(iter, &socket_list) {
		count++;
	}

	PDEBUG(MSG_COMMON, "%s: Allocating memory for info buffer\n", DEBUG_NAME);

	alloc_size = SOCKET_STRING_MEM*count;
	info = kmalloc(alloc_size, GFP_KERNEL);
	if(!info) {
		PDEBUG(MSG_COMMON, "%s: Could not allocate enough memory for printing sockets\n", DEBUG_NAME);
		*sinfo = NULL;
		ret = 0;
		goto free_mutex;
	}

	*sinfo = info;

	PDEBUG(MSG_COMMON, "%s: Gathering socket information\n", DEBUG_NAME);

	count = 0;
	list_for_each(iter, &socket_list) {
		entry = list_entry(iter, struct socket_entry, list);
		if(!entry) {
			PDEBUG(MSG_COMMON, "%s: Could not find container of list item at %u\n", DEBUG_NAME, (unsigned)iter);
			goto free_mutex;
		}
// 		PDEBUG(MSG_COMMON, "SM: Entry found, socketno is %u\n", entry->socketno);
		written = snprintf(socketinfo, 20, "%d\n", entry->msock->socketno);
		if(count+written > alloc_size) {
			PDEBUG(MSG_COMMON, "%s: Could not write everything into info buffer\n", DEBUG_NAME);
			goto free_mutex;
		}
		strncpy(&info[count], socketinfo, written);
		count += written;
// 		PDEBUG(MSG_COMMON, "SM: Socket information of socket no %u gathered\n", entry->socketno);
	}

free_mutex:
	up(&mutex);

	info[count++] = '\0';

	return count;
}
#endif /*DEBUG*/

