/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#ifndef PRRT_BLOCK_LIST_H
#define PRRT_BLOCK_LIST_H

#include "../prrt_block.h"

/**
 * This class implements a list that holds blocks of struct prrt_block.
 * This list is only used on sender side as the receiver does not really
 * know about blocks.
 * INVARIANTS:
 * - The blocks are ordered in the list by the numbers of the contained packets.
 */
__NOT_THREAD_SAFE __PCLASS struct prrt_block_list {
	__PROTECTED
	struct list_head head;		/**< The head of the list with the blocks (see kernel lists)*/
	struct list_head *next_packet;	/**< The next packet to send.*/
	struct list_head *next_packet_block; /**< The block which contains the next packet.*/
	long next_packet_no;		/**< The packetno of the next packet to send.*/
	struct semaphore packet_count;	/**< Semaphore that counts the data packets that need to be sent.*/

	atomic_t block_count;		/**< Counter for the number of blocks in the queue.*/
};

/**
 * Initialise a block list.
 */
void prrtblist_init(struct prrt_block_list *blist);

/**
 * Reinitialise a block list.
 * Needed when connecting the sender to a new ip address.
 */
int prrtblist_reinit(struct prrt_block_list *blist);

/**
 * Destroy all blocks in the list.
 * This frees all blocks and the packets they contain. It also removes the packets
 * from the lists they are in.
 */
void prrtblist_destroy_all_blocks(struct prrt_block_list *blist);

/**
 * Insert a block into the list.
 * The block may contain packets but you may only insert blocks that contain
 * packets that were already added to the list.
 */
int prrtblist_insert_block(struct prrt_block_list *blist, struct prrt_block *mblock);

/**
 * May we insert a packet in the last block of the list.
 * If not a new block must be inserted.
 */
int prrtblist_can_insert_packet(struct prrt_block_list *blist);

/**
 * Insert a data packet into the last block of the list.
 */
//int prrtblist_insert_packet(struct prrt_block_list *blist, struct prrt_packet_container *pcont, struct params_code *cpars, struct params_local *lpars);
int prrtblist_insert_packet(struct prrt_block_list *blist, struct prrt_packet_container *pcont, struct params_code *cpars);

/**
 * All data packets of this block were sent.
 */
int prrtblist_block_was_send_completely(struct prrt_block_list *blist, struct prrt_block *mblock);

/**
 * Get the block that contains the packet number.
 */
int prrtblist_get_block(struct prrt_block_list *blist, u16 packetno, struct prrt_block **mblock);

/**
 * Get the last block in the list.
 */
int prrtblist_get_last_block(struct prrt_block_list *blist, struct prrt_block **mblock);

/**
 * Get the next packet that must be sent.
 * Pay attention to confirm the sending of a packet on success.
 */
int prrtblist_get_next_packet(struct prrt_block_list *blist, struct prrt_packet_container **pcont);

/**
 * After a packet was sent one must confirm the sending.
 */
void prrtblist_confirm_packet_sending(struct prrt_block_list *blist);

/**
 * Get the packet number of the last packet in the block list.
 */
int prrtblist_get_last_packetno(struct prrt_block_list *blist, int *packetno);

/**
 * Get timestamp of last packet in the block list.
 */
int prrtblist_get_timestamp_of_last_packet(struct prrt_block_list *blist, u32 *timestamp);

/**
 * Retrieve the first block in the block list.
 */
int prrtblist_get_first_block(struct prrt_block_list *blist, struct prrt_block **mblock);

/**
 * Check if the first block must be deleted and do so if necessary.
 */
int prrtblist_try_to_delete_first_block(struct prrt_block_list *blist);

/**
 * Print some statistics for the block list.
 */
void prrtblist_print_stats(struct prrt_block_list *blist);

#endif /*PRRT_BLOCK_LIST_H*/

