/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#ifndef _PRRT_DATA_LIST_H
#define	_PRRT_DATA_LIST_H

#include "../defines.h"
#include "../stats/list_stats.h"
#include "prrt_list.h"

#include <linux/list.h>
#include <linux/slab.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26)
#include <asm/semaphore.h>
#else
#include <linux/semaphore.h>
#endif

/**
 * This list holds the data packets that were received by the receiver so far.
 * Decoded redundancy is inserted into this list with a changed sequence number
 * but the old header structure.
 * Packets that were read by the application or were too old are deleted.
 * INVARIANT:
 * - packets are ordered according to their sequence number.
 */
__NOT_THREAD_SAFE __PCLASS struct prrt_data_list {
	struct list_head head;		/**<The head of the list containing the data packets.*/
	struct semaphore count;		/**<A semaphore used to count the elements of the list.*/
	struct prrt_list_stats stats;	/**<Holds statistics concerning this list.*/
	long int last_read_pno;		/**<The number of the last packet that was removed/read from
                                         *  the list. */
	char last_packet_inserted_at_the_end;
	
	// Variables for the function prrtdlist_get_data_packet_by_block.
	long WorkingBlockStart;
  long WorkingBlockEnd;
  struct list_head *WorkingBlocksLastData;
		
	struct prrt_time_state *time_state; /**< The state of time in this socket. */
};

/**
 * Initialise the data list.
 */
int prrtdlist_init(struct prrt_data_list *dlist, struct prrt_time_state *time_state);
int prrtdlist_reinit(struct prrt_data_list *dlist);

/**
 * Free the packets contained in the data list.
 */
void prrtdlist_free(struct prrt_data_list *dlist);

/**
 * Is the data list empty.
 */
int prrtdlist_is_empty(struct prrt_data_list *dlist);

/**
 * Get the number of the last packet that was removed/read from the list.
 */
int prrtdlist_get_last_read_packetnumber(struct prrt_data_list *dlist, u16 *pno);

/**
 * Insert data packet at the right position in the list.
 * \param pcont the packet that should be inserted.
 */
int prrtdlist_insert_datapacket(struct prrt_data_list *dlist, struct prrt_packet_container *pcont);

/**
 * States if the last packet that was inserted into the list went to the end of the list.
 */
int prrtdlist_last_packet_inserted_at_the_end(struct prrt_data_list *dlist);

/**
 * Waits at most POLL_MAXWAITTIME for a data packet to be inserted into the list and tries to get
 * ownership of it.
 */
int prrtdlist_wait_for_and_grab_next_datapacket(struct prrt_data_list *dlist);

/**
 * Return the ownership of a packet to the list.
 */
void prrtdlist_release_a_grabbed_packet(struct prrt_data_list *dlist);

/**
 * May the next data packet in the list be read and removed.
 * \param d_target the target delay used for this socket.
 */
int prrtdlist_next_datapacket_is_readable(struct prrt_data_list *dlist, u32 d_target);

/**
 * Retrieve next data packet.
 * Only do this after obtaining ownership of it.
 * \param pcont will point to the packet on success.
 */
int prrtdlist_get_next_datapacket(struct prrt_data_list *dlist, struct prrt_packet_container **pcont);

int prrtdlist_show_next_datapacket(struct prrt_data_list *dlist, struct prrt_packet_container **pcont);
/**
 * Delete the first packet in the list if it is too old.
 * \param d_target the target delay used for this socket.
 */
long int prrtdlist_delete_old_packet(struct prrt_data_list *dlist, u32 d_target);

/**
 * Retrieve data packets of a block.
 * This function must be called until it returns -1 because it holds an internal state that is
 * resetted on failure.
 * It will retrieve the different data packets of a block until it can not find anymore packets of
 * the block.
 * \param p the list entry for a packet in the list.
 * \param BlockStart the first sequence number in this block.
 * \param BlockEnd the last sequence number in this block.
 */
int prrtdlist_get_data_packet_by_block(struct prrt_data_list *dlist, struct packet_list_entry **p,
				      u16 BlockStart, u16 BlockEnd);



#endif	/* _PRRT_DATA_LIST_H */

