/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#ifndef PRRT_INC_LIST_H
#define PRRT_INC_LIST_H

#include "prrt_list.h"
#include "../kui/kui_option.h"
#include "../vdmcode/prrt_block_code.h"
#include "../loss_gatherer.h"
#include "../parameters/param_network.h"
#include "prrt_data_list.h"
#include "prrt_redundancy_list.h"

/**
 * This Pseudo Class implements a list for incoming packets at the receiver.
 * It contains a list for data packets and a list for redundancy packets. They are both locked
 * separately with the locks data_lock and red_lock.
 * The public objects are thread safe by themself and are completely separated from the private
 * objects (they are only used by them).
 * Functions for this class start with prrtilist_.
 */
__THREAD_SAFE __PCLASS struct prrt_inc_list {
	__PRIVATE
	struct prrt_data_list data_list;		/**< This is the list with data packets.*/
	struct semaphore data_lock;		/**< Synchronizes access to the data list.*/
	struct prrt_redundancy_list red_list;	/**< This is the list with redundancy packets.*/
	struct semaphore red_lock;		/**< Synchronizes access to the redundancy list.*/

	struct prrt_coder *decoder;		/**< The coder object that is used to decode.*/

	long biggest_inserted_data_packet;	/**< The number of the data packet with the biggest
						 *  sequence number that was inserted.*/
	long current_sn_base;			/**< Current block's first sequence number.*/
	int dec_timeout_min;
	int dec_timeout_max;

	/*options for this sockets receiving part(1=on)*/
	int feedback;	/*should the receiving part send feedback?*/
  int decoding;	/*should the receiving part decode replacements?*/
  struct semaphore opt_lock;

	__LOCAL
	long first_packet_fb_gathering;		/**< The first number encountered when gathering
						 *   feedback.*/

	/*packet stats*/
	__TCONFINED unsigned long received_datapackets;	/**< #Data packets received in this list.*/
	__TCONFINED unsigned long received_redpackets;	/**< #Red. packets received in this list.*/
	__TCONFINED unsigned long received_overall;	/**< #Data and red. packets received.*/
	__TCONFINED unsigned long ommited_at_start_packets; /**< #Data packets ommited before start of reading.*/
	__TCONFINED unsigned long old_packets;		/**< #Packets deleted from this list.*/
	__TCONFINED unsigned long lost_packets;		/**< #Packets that were lost.*/
	__TCONFINED unsigned long missed_packets;	/**< #Packets that were not delivered.*/
	__TCONFINED unsigned long used_redpackets;	/**< #Red. Packets that were not ommited on arrival.*/
	__TCONFINED unsigned long useful_redpackets;	/**< #Red. Packets that were decoded and inserted.*/

	__TCONFINED unsigned long feedbacks_sent;	/**< #Feedbacks that was sent.*/
	__TCONFINED unsigned long fbpackets_sent;	/**< #Feedback packets that was sent.*/

	__TCONFINED unsigned long lastPacketDelivered;  /**< Last packet delivered to application.*/
	atomic_t lastMeasured_t_replace;		/**< last t_replace that was measured.*/
	atomic_t lastDecodingDelay;			/**< last time between receiving first data and decoding*/
	atomic_t lastPacketTs;				/**< last incoming packet time*/
	atomic_t lastPacketTimeout;			/**< last packet timeout*/
	atomic_t lastPacketRecvTs;			/**< last incoming packet receive time in microseconds*/
	atomic_t lastPacketRecvTsLocal;			/**< last incoming packet receive time in microseconds*/
	atomic_t lastFBSentTs;				/**< last feedback sent time*/
	atomic_t readingAlreadyStarted;

	// variables for the function prrtilist_get_first_packet
	unsigned long last_retrieved_packetno_plus_one; /**< The sequence number of the last retrieved packet plus 1*/
	
	__PUBLIC
	/*these objects are even thread safe if public*/
	struct prrt_loss_gatherer lossgath;	/**< This is the object that manages feedback.*/

	/*parameters for this receiver*/
	struct params_network net_params;	/**< The parameters for the network connection.*/
	
	struct prrt_time_state *time_state; /**< The state of time in this socket. */
};

/**
 * Initialise a prrt_inc_list class object.
 */
void prrtilist_init(struct prrt_inc_list *ilist, struct prrt_time_state *time_state);

/**
 * Destroy all packets in this list structure.
 */
void prrtilist_uninit(struct prrt_inc_list *ilist);

/**
 * Insert a data packet into list.
 * The packet is inserted at the right position in the ordered list data_list.
 * \param pcont the packet that should be inserted.
 * \return 0 if inserted, -1 if packet is to old, -2 if packet is already in list, -ENOMEM if no
 *         memory is available
 */
int prrtilist_insert_datapacket(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont);

/**
 * Retrieve the first packet in the list.
 * After the packet is retrieved, it is removed from the list. The functions waits POLL_MAXWAITTIME
 * if the list is empty. When no packets are present at the end of the wait time it returns.
 * \param pcont this will point to the packet after it was retrieved.
 * \param d_target the target delay that is used for this socket.
 * \return 0 if retrieved, -1 if timeout happened, -ERESTARTSYS if interrupted
 */
int prrtilist_get_first_packet(struct prrt_inc_list *ilist, struct prrt_packet_container **pcont,
			      u32 d_target);

/**
 * Delete old packets in the list.
 * This function will delete all old packets in the list.
 * See prrt_packet_may_delete for the conditions of deletion.
 * \param d_target The target delay used for this socket.
 */
void prrtilist_delete_old_packets(struct prrt_inc_list *ilist, u32 d_target);

/**
 * Builds a feedback packet from the losses gathered so far.
 * \param mfbpack on success, points to the feedback packet build (free this after use).
 * \param d_target The target delay used for this socket.
 * \return 0 on success, <0 if no feedback packet could be build.
 */
int prrtilist_gather_feedback_packet(struct prrt_inc_list *ilist, struct prrt_fbpacket_container **mfbcont,
				    int d_target, u16 *repetitions);

/**
 * This function tries to use a redundancy packet to replace a lost packet.
 * The function searches all packets of the block that the replacement packet belongs to. If only
 * one packet is missing it decodes the packet. Else it will search in the list of incoming
 * redundancy packets for other replacement packets belonging to this block. If enough packets are
 * found it will decode the missing packets. If not the replacement packet is added to the list.
 * By giving pcont to this function the ownership is given to it also.
 * \param pcont the replacement packet received.
 * \return 0 on success
 */
int prrtilist_try_decode(struct prrt_inc_list *ilist, struct prrt_packet_container *pcont);

/**
 * Set a option of the receiver side.
 */
int prrtilist_set_option(struct prrt_inc_list *ilist, enum option_t opt, int val);

/**
 * Retrieve the value of an option on receiver side.
 */
int prrtilist_get_option(struct prrt_inc_list *ilist, enum option_t opt);

#endif

