/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "prrt_list.h"

#define DEBUG_NAME "PRRT_LIST"

static struct kmem_cache *list_entry_cache;
static struct semaphore list_entry_cache_lock;
//int listentries = 0;

int prrt_entry_cache_init(void)
{
	init_MUTEX(&list_entry_cache_lock);
	list_entry_cache = kmem_cache_create("list_entry_cache",
					  sizeof(struct packet_list_entry),
					  0, 0,
					  NULL);

	if(!list_entry_cache) {
		PERROR("%s: Could not create list entry cache\n", DEBUG_NAME);
		return -1;
	}

	return 0;
}

void prrt_entry_cache_uninit(void)
{
	kmem_cache_destroy(list_entry_cache);
}

int prrt_list_entry_alloc(struct packet_list_entry **p)
{
	PDEBUG(MSG_ALLOC, "%s: Allocating list entry\n", DEBUG_NAME);
	down(&list_entry_cache_lock);
	*p = kmem_cache_alloc(list_entry_cache, GFP_KERNEL);
	//listentries++;
	//PDEBUG(MSG_TRANSPORT, "%s: Allocating list entry (count %d)\n", DEBUG_NAME, listentries);
	up(&list_entry_cache_lock);

	PDEBUG(MSG_ALLOC, "%s: List entry allocated\n", DEBUG_NAME);
	/*zero out packet for coding purposes*/
	if(!(*p)) {
		PERROR("%s: Could not allocate memory for packet_list_entry\n", DEBUG_NAME);
		return -1;
	} else {
		bzero(*p, sizeof(**p));
	}

	return 0;
}

void prrt_list_entry_free(struct packet_list_entry **p)
{
	if(!(*p))
		return;

	down(&list_entry_cache_lock);
	kmem_cache_free(list_entry_cache, *p);
	//listentries--;
	up(&list_entry_cache_lock);
	*p = 0;
	PDEBUG(MSG_ALLOC, "%s: List entry deleted\n", DEBUG_NAME);
}

void prrt_list_delete_packet_from_list(struct list_head *iter)
{
	struct packet_list_entry *p = 0;

	p = container_of(iter, struct packet_list_entry, list);

	list_del(iter);
	prrt_packet_destroy(&p->pcont);
	prrt_list_entry_free(&p);
}

void prrt_list_delete_all_packets_in_list(struct list_head *head)
{
	struct list_head *iter = 0;
	struct list_head *next = 0;

	list_for_each_safe(iter, next, head) {
		prrt_list_delete_packet_from_list(iter);
	}
}

int PacketIsAlreadyInListAtThisPosition(u16 packetNumber, struct packet_list_entry *p);
int ThisIsThePositionForThisPacket(u16 packetNumber, struct packet_list_entry *p);

int prrt_list_find_position_for_sorted_insertion(struct list_head *head, u16 packetNumber,
						struct list_head **position)
{
	struct packet_list_entry *p = 0;
	struct list_head *iter = 0;

	*position = 0;

	PDEBUG(MSG_DATALIST, "%s: Search for right position in list\n", DEBUG_NAME);

	list_for_each_prev(iter, head) {
		p = list_entry(iter, struct packet_list_entry, list);

		if(PacketIsAlreadyInListAtThisPosition(packetNumber, p)) {
			return -2;
		}

		if(ThisIsThePositionForThisPacket(packetNumber, p)) {
			*position = iter;
			return 0;
		}
	}

	/*new packet needs to go to the start of the list*/
	*position = head;

	return 0;
}

int PacketIsAlreadyInListAtThisPosition(u16 packetNumber, struct packet_list_entry *p)
{
	return (prrt_comp_packetnos(p->packetnumber, packetNumber) == 0);
}

int ThisIsThePositionForThisPacket(u16 packetNumber, struct packet_list_entry *p)
{
	return (prrt_comp_packetnos(p->packetnumber, packetNumber) < 0);
}

int prrt_list_insert_packet_behind_position(struct prrt_packet_container *pcont,
					   struct list_head *position)
{
	int err = 0;
	struct packet_list_entry *new = 0;
	u32 packetnumber = 0;

	prrt_list_entry_alloc(&new);
	if(!new) {
		PERROR("%s: No memory for new data packet entry\n", DEBUG_NAME);
		return -ENOMEM;
	}

	PDEBUG(MSG_DATALIST, "%s: Entry allocated, setting content\n", DEBUG_NAME);
	new->pcont = pcont;

	if(prrt_packet_is_redundancy(pcont)) {
		packetnumber = (u32)pcont->pack.red.red_header.sn_base;
		packetnumber += (u32)pcont->pack.red.header.index;
		new->packetnumber = packetnumber;
	}
	else {
		new->packetnumber = pcont->pack.data.header.seqnum;
	    //PDEBUG(MSG_TRANSPORT, "%s: Inserting seqnum %u\n", DEBUG_NAME, new->packetnumber);
	}

	list_add(&new->list, position);

	return err;
}
