#ifndef PRRT_OUT_LIST_H
#define PRRT_OUT_LIST_H

#include "prrt_list.h"
#include "../kui/kui_option.h"
#include "prrt_block_list.h"
#include "receiver_list.h"
#include "../prrt_time.h"
#include "../feedback_analyzer.h"

#include <linux/net.h>
#include <linux/in.h>

/**
 * This class contains all the functionality for the sender side of a connection.
 * 
 */
__THREAD_SAFE __PCLASS struct prrt_out_list {
	__PRIVATE
	struct prrt_block_list block_list;	/**< The block list that governs the blocks to send.*/
	struct list_head *next_block;		/**< The block whose redundancy should be send next.*/
	struct semaphore list_lock;		/**< Mutex to guard access to the lists.*/
	struct semaphore coded_block_count;	/**< This mutex is increased when coding a block.*/

	struct prrt_coder *out_cod;              /**< The coder used to code packets on demand.*/

	u16 out_pno_red;                        /**< Packet number for next redundancy packet.*/
	struct semaphore out_pno_lock;          /**< Lock for incrementing out_pno_red.*/

	/*options for this sockets sending part(1=on)*/
	int replace;                            /*should the sender send replacements*/
	int coding;                             /*should packets be encoded?(turns replacement off)*/
  int gather_feedback;                    /*should the feedback of the receivers be gathered*/
  struct semaphore opt_lock;

	/*stats for sending and deleting elements*/
	__TCONFINED unsigned long send_packets;		/**< #packets that were send*/
	__TCONFINED unsigned long send_replacements;	/**< #replacements send*/
	__TCONFINED unsigned long old_blocks;		/**< #blocks deleted*/

	__PUBLIC
	struct prrt_receiver_list recv_list; /**< This list holds all known receivers of the stream.*/
	struct prrt_fb_analyzer fbana;       /**< This object holds the functionality to schedule
                                            *    controlled replacement sending.*/
	
	struct prrt_time_state *time_state; /**< The state of time in this socket. */
	u32 last_sent;				/*timestamp of last sending activity*/
	u32 send_tokens;			/*number of available send tokens*/

  int traffic_shaping_datarate;   /** The datarate for the traffic shaping **/
};

/**
 * Initialise a prrt_inc_list class object.
 */
void prrtolist_init(struct prrt_out_list *olist, struct prrt_time_state *time_state);
int prrtolist_reinit(struct prrt_out_list *olist);

/**
 * Free all allocated memory.
 */
void prrtolist_free(struct prrt_out_list *olist);

/**
 * Insert packet into out list for sending.
 */
int prrtolist_insert_packet(struct prrt_out_list *olist, struct prrt_packet_container *pcont, struct params_code *cpars);

/**
 * Try to code the next uncoded block.
 * \param cpars the coding parameters that should be used for coding.
 */
int prrtolist_try_to_code_next_block(struct prrt_out_list *olist, struct params_code *cpars);

/**
 * Wait for the next data packet, not interruptible.
 */
int prrtolist_wait_for_next_packet(struct prrt_out_list *olist);

/**
 * Sends the next packet that must be sent.
 * \param sock the socket to send the packets on.
 * \param paddr the addr to send the packets to.
 * \return 0 on success, <0 on failure (see kernel errors)
 */
int prrtolist_send_next_packet(struct prrt_out_list *olist, struct socket *sock,
			      struct sockaddr_in *paddr);

/**
 * Sends the redundancy of the 0th round of next block that is ready.
 * \param sock the socket to send the packets on.
 * \param paddr the addr to send the packets to.
 * \return 0 on success, <0 on failure (see kernel errors)
 */
int prrtolist_try_to_send_redundancy_of_next_coded_block(struct prrt_out_list *olist, struct socket *sock,
							struct sockaddr_in *paddr);

/**
 * Delete old blocks from the list.
 * This removes blocks from the block list, the packets of the blocks from the packet list
 * and frees the memory for them.
 * A block is deleted if its newest packet is to old.
 * \param delay maximum delay between newest and oldest packet.
 * \return 0 on success, -1 if interrupted
 */
int prrtolist_delete_old_blocks(struct prrt_out_list *olist, u32 d_target);

/**
 * This function computes the replacements that should be send and sends them.
 * \param sock the socket where the message should be send on
 * \param paddr the addr where the message should be send to
 */
int prrtolist_send_replacements(struct prrt_out_list *olist, struct socket *sock,
				struct sockaddr_in *paddr);

/**
* Insert a feedback into the feedback analyser.
* \param miss number of the missing packet.
* \param receiver IP address of the receiver.
* \param retr_round the number of times this feedback was transmitted.
*/
int prrtolist_insert_feedback(struct prrt_out_list *olist, u16 miss, u32 receiver, int retr_round);

int prrtolist_set_option(struct prrt_out_list *olist, enum option_t opt, int val);

int prrtolist_get_option(struct prrt_out_list *olist, enum option_t opt);

#endif /*PRRT_OUT_LIST_H*/

