/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "prrt_redundancy_list.h"

#define DEBUG_NAME "PRRT_RED_LIST"

int prrtredlist_init(struct prrt_redundancy_list *redlist)
{
	INIT_LIST_HEAD(&redlist->head);
	redlist->count = 0;
	redlist->last_deleted_packno = -1;
	prrtstats_init_list_stats(&redlist->stats);
	
	redlist->lastBlocksFirstDataNo = -1;
	redlist->lastBlocksNextRedundancy = 0;

	return 0;
}

void prrtredlist_free(struct prrt_redundancy_list *redlist)
{
	prrt_list_delete_all_packets_in_list(&redlist->head);
}

static int RedundancyPacketMayBeInserted(struct prrt_redundancy_list *redlist,
					 struct prrt_packet_container *pcont);

int prrtredlist_insert_redundancy_packet(struct prrt_redundancy_list *redlist,
					struct prrt_packet_container *pcont)
{
	int err = 0;
	struct list_head *position = 0;
	// redundancy list requires sequence number just for ordered packet insertion.
	u32 packetnumber = (u32)pcont->pack.red.red_header.sn_base;
	packetnumber += (u32)pcont->pack.red.header.index;

	//PDEBUG(MSG_TRANSPORT, "%s: Inserting parity with packet number %lu\n", DEBUG_NAME, (u16)packetnumber);
	if(!RedundancyPacketMayBeInserted(redlist, pcont)){
		//PDEBUG(MSG_TRANSPORT, "%s: Error inserting parity: May not be inserted\n", DEBUG_NAME);
		return -1;
	}

	err = prrt_list_find_position_for_sorted_insertion(&redlist->head, (u16)packetnumber, &position);
	if(err < 0){
		//PDEBUG(MSG_TRANSPORT, "%s: Error inserting parity: did not find position\n", DEBUG_NAME);
		return err;
	}

	err = prrt_list_insert_packet_behind_position(pcont, position);
	if(err < 0){
		//PDEBUG(MSG_TRANSPORT, "%s: Error inserting parity: could not insert\n", DEBUG_NAME);
		return err;
	}

	redlist->count++;

	return err;
}

static int NoDataPacketsWereRemovedUntilNow(struct prrt_redundancy_list *redlist);
static int FirstDataPacketInBlockAlreadyGone(struct prrt_redundancy_list *redlist, u16 first_packno);

static int RedundancyPacketMayBeInserted(struct prrt_redundancy_list *redlist,
					 struct prrt_packet_container *pcont)
{
	int first_data_packetnumber_in_block = pcont->pack.red.red_header.sn_base;

	if(NoDataPacketsWereRemovedUntilNow(redlist))
		return 1;

	if(FirstDataPacketInBlockAlreadyGone(redlist, first_data_packetnumber_in_block))
		return 0;

	return 1;
}

static int NoDataPacketsWereRemovedUntilNow(struct prrt_redundancy_list *redlist)
{
	return (redlist->last_deleted_packno == -1);
}

static int FirstDataPacketInBlockAlreadyGone(struct prrt_redundancy_list *redlist, u16 first_packno)
{
	return (prrt_comp_packetnos(first_packno, (u16)(redlist->last_deleted_packno)) <= 0);
}

int FindARedundancyPacketForBlock(struct prrt_redundancy_list *redlist,
				  u16 lastBlocksFirstDataNo, struct list_head **nextRedundancy);

int prrtredlist_get_redundancy_packet_by_block_start(struct prrt_redundancy_list *redlist,
						    struct packet_list_entry **p,
						    u16 firstDataPacketno)
{
	int err = 0;

	if(firstDataPacketno != redlist->lastBlocksFirstDataNo) {
		redlist->lastBlocksFirstDataNo = firstDataPacketno;
		redlist->lastBlocksNextRedundancy = 0;
		//PDEBUG(MSG_TRANSPORT, "%s: lastBlocksFirstDataNo %u\n", DEBUG_NAME, redlist->lastBlocksFirstDataNo);
	}

	PDEBUG(MSG_TRANSPORT, "%s: lastBlocksFirstDataNo %u\n", DEBUG_NAME, redlist->lastBlocksFirstDataNo);
	err = FindARedundancyPacketForBlock(redlist, redlist->lastBlocksFirstDataNo, &redlist->lastBlocksNextRedundancy);
	if(err < 0) {
		redlist->lastBlocksFirstDataNo = -1;
		redlist->lastBlocksNextRedundancy = 0;
		*p = 0;
	} else {
		*p = container_of(redlist->lastBlocksNextRedundancy, struct packet_list_entry, list);
	}



	return err;
}

static int BlockNotSearchedYet(struct list_head *nextRedundancy);
int SearchCompleteListForBlockRedundancy(struct prrt_redundancy_list *redlist,
					 u16 lastBlocksFirstDataNo, struct list_head **nextRedundancy);
int TryToGetNextRedundancyPacketOfBlock(struct prrt_redundancy_list *redlist,
					 u16 lastBlocksFirstDataNo, struct list_head **nextRedundancy);

int FindARedundancyPacketForBlock(struct prrt_redundancy_list *redlist,
				  u16 lastBlocksFirstDataNo, struct list_head **nextRedundancy)
{
	int err = 0;

	if(BlockNotSearchedYet(*nextRedundancy)){
		err = SearchCompleteListForBlockRedundancy(redlist, lastBlocksFirstDataNo, nextRedundancy);
		//PDEBUG(MSG_TRANSPORT, "%s: SearchedCompleteList %d\n", DEBUG_NAME, err);
	}
	else {
		err = TryToGetNextRedundancyPacketOfBlock(redlist, lastBlocksFirstDataNo, nextRedundancy);
		PDEBUG(MSG_TRANSPORT, "%s: TriedToGetNextRedundancy %d\n", DEBUG_NAME, err);
	}
	return err;
}

static int BlockNotSearchedYet(struct list_head *nextRedundancy)
{
	return ((nextRedundancy) == 0);
}

int ListEntryBelongsToBlock(struct list_head *list_entry, u16 firstDataNumber);

int SearchCompleteListForBlockRedundancy(struct prrt_redundancy_list *redlist,
					 u16 lastBlocksFirstDataNo, struct list_head **nextRedundancy)
{
	struct list_head *iter = 0;

	list_for_each(iter, &redlist->head) {
		if(ListEntryBelongsToBlock(iter, lastBlocksFirstDataNo)) {
			*nextRedundancy = iter;
			return 0;
		}
	}

	*nextRedundancy = 0;
	return -1;
}

int ListEntryBelongsToBlock(struct list_head *list_entry, u16 firstDataNumber) 
{
	struct packet_list_entry *p = 0;
	
	p = container_of(list_entry, struct packet_list_entry, list);

	if(p->pcont->pack.red.red_header.sn_base == firstDataNumber){
		return 1;
	}
	
	return 0;
}

static int ReachedLastPacket(struct prrt_redundancy_list *redlist, struct list_head *nextRedundancy);
int TryFollowingPacket(struct list_head **nextRedundancy, u16 lastBlocksFirstDataNo);
int TryPreviousPacket(struct list_head **nextRedundancy, u16 lastBlocksFirstDataNo);

int TryToGetNextRedundancyPacketOfBlock(struct prrt_redundancy_list *redlist,
					 u16 lastBlocksFirstDataNo, struct list_head **nextRedundancy)
{
	int err = 0;

	if(ReachedLastPacket(redlist, *nextRedundancy)) {
		PDEBUG(MSG_TRANSPORT, "%s: Reached last packet\n", DEBUG_NAME);
		goto error;
	}

	err = TryFollowingPacket(nextRedundancy, lastBlocksFirstDataNo);
	if(err == 0) {
		*nextRedundancy = (*nextRedundancy)->next;
		return err;
	}

	/*this means we have found all packets of the block*/
/*
	err = TryPreviousPacket(nextRedundancy, lastBlocksFirstDataNo);
	if(err < 0)
		return err;
*/

error:
	*nextRedundancy = 0;
	return -1;
}

static int ReachedLastPacket(struct prrt_redundancy_list *redlist, struct list_head *nextRedundancy)
{
	return (nextRedundancy->next == &redlist->head);
}

int TryFollowingPacket(struct list_head **nextRedundancy, u16 lastBlocksFirstDataNo)
{
	struct list_head *iter = 0;

	iter = (*nextRedundancy)->next;
	if(ListEntryBelongsToBlock(iter, lastBlocksFirstDataNo)) {
		return 0;
	}
		PDEBUG(MSG_TRANSPORT, "%s: Does not belong to Block\n", DEBUG_NAME);

	return -1;
}

int TryPreviousPacket(struct list_head **nextRedundancy, u16 lastBlocksFirstDataNo)
{
	struct list_head *iter = 0;

	iter = (*nextRedundancy)->prev;
	if(ListEntryBelongsToBlock(iter, lastBlocksFirstDataNo)) {
		return 0;
	}

	return -1;
}

void prrtredlist_delete_redundancy_packet(struct prrt_redundancy_list *redlist,
					 struct packet_list_entry *p)
{
	prrt_list_delete_packet_from_list(&p->list);

	redlist->count--;
}

void prrtredlist_delete_redundancy_packets_by_block_start(struct prrt_redundancy_list *redlist,
							 u16 firstDataPacketno)
{
	struct list_head *iter = 0;
	struct list_head *next = 0;
	char alreadyDeletedSomeThing = 0;

	list_for_each_safe(iter, next, &redlist->head) {

		if(ListEntryBelongsToBlock(iter, firstDataPacketno)) {
			prrt_list_delete_packet_from_list(iter);

			redlist->count--;
			alreadyDeletedSomeThing = 1;
		} else {
			if(alreadyDeletedSomeThing)
				break;	/*we have deleted all of the block, stop now*/
		}

	}
}

void prrtredlist_remove_entry_from_list(struct prrt_redundancy_list *redlist,
				       struct packet_list_entry *p)
{
	list_del(&p->list);
	prrt_list_entry_free(&p);

	redlist->count--;
}

void prrtredlist_unlink_entry_from_list(struct prrt_redundancy_list *redlist,
				       struct packet_list_entry *p)
{
	list_del(&p->list);

	redlist->lastBlocksFirstDataNo = -1;
	redlist->lastBlocksNextRedundancy = 0;

	redlist->count--;
}

void prrtredlist_inform_of_packet_deletion(struct prrt_redundancy_list *redlist, u16 deletedPacketNo)
{
	if(prrt_comp_packetnos((u16)(redlist->last_deleted_packno), deletedPacketNo) < 0)
		redlist->last_deleted_packno = deletedPacketNo;
}
