/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#ifndef _PRRT_REDUNDANCY_LIST_H
#define	_PRRT_REDUNDANCY_LIST_H

#include "../defines.h"
#include "../stats/list_stats.h"
#include "prrt_list.h"

#include <linux/list.h>
#include <linux/slab.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26)
#include <asm/semaphore.h>
#else
#include <linux/semaphore.h>
#endif

/**
 * This list contains redundancy packets that are safed till they are needed or discarded.
 * Redundancy packets need to be discarded when no packets are missing in their block anymore
 * or the first data packet of their block is removed from the data list.
 * INVARIANTS:
 * - The blocks are ordered in the list by the numbers of the contained packets.
 */
__NOT_THREAD_SAFE __PCLASS struct prrt_redundancy_list {
	    __PROTECTED
	    struct list_head head;	/**<The head of the list containing the redundancy packets.*/
	    int count;			/**<Number of packets in the list.*/
	    struct prrt_list_stats stats;/**<Holds statistics concerning this list.*/
	    long last_deleted_packno;	/**<This is the packetnumber of the last deleted data packet.*/
		  long lastBlocksFirstDataNo;  /**<The sequence number of the first data packet in the last block.*/
		  struct list_head *lastBlocksNextRedundancy;  /**< Pointer to the next redundancy packet in the last block.*/
};

/**
 * Initialise the structure.
 */
int prrtredlist_init(struct prrt_redundancy_list *redlist);

/**
 * Free all packets contained in the list.
 */
void prrtredlist_free(struct prrt_redundancy_list *redlist);

/**
 * Insert redundancy packet at the right position in the list.
 * \param pcont redundancy packet to insert.
 */
int prrtredlist_insert_redundancy_packet(struct prrt_redundancy_list *redlist, 
					struct prrt_packet_container *pcont);

/**
 * Retrieve the redundancy packets of one block one after another.
 * This function must be called until it returns -1 because it holds an internal state that is
 * resetted on failure.
 * \param p list entry to the retrieved packet.
 * \param firstDataPacketno the first sequence number of the data packets in the block.
 */
int prrtredlist_get_redundancy_packet_by_block_start(struct prrt_redundancy_list *redlist,
						    struct packet_list_entry **p,
						    u16 firstDataPacketno);

/**
 * Delete the given redundancy packet from the list and free its memory.
 * \param p the redundancy packet to delete from the list.
 */
void prrtredlist_delete_redundancy_packet(struct prrt_redundancy_list *redlist,
					 struct packet_list_entry *p);

/**
 * Delete all redundancy packets of a block from the list.
 * \param firstDataPacketno the first sequence number of the data packets in the block.
 */
void prrtredlist_delete_redundancy_packets_by_block_start(struct prrt_redundancy_list *redlist,
							 u16 firstDataPacketno);

/**
 * Only remove the given redundancy packet from the list without freeing it.
 * \param p the redundancy packet to delete from the list.
 */
void prrtredlist_remove_entry_from_list(struct prrt_redundancy_list *redlist,
				       struct packet_list_entry *p);

void prrtredlist_unlink_entry_from_list(struct prrt_redundancy_list *redlist,
				       struct packet_list_entry *p);

/**
 * Inform the class object of the deletion of a data packet.
 * \param deletedPacketNo the sequence number of the deleted data packet.
 */
void prrtredlist_inform_of_packet_deletion(struct prrt_redundancy_list *redlist, u16 deletedPacketNo);

#endif	/* _PRRT_REDUNDANCY_LIST_H */

