/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#ifndef PARAM_CODE_H
#define PARAM_CODE_H

#include "../defines.h"
#include "param_network.h"
//#include "param_local.h"
#include "../prrt_fbpacket.h"
#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26)
#include <asm/semaphore.h>
#else
#include <linux/semaphore.h>
#endif
#include <asm/atomic.h>

/**
 * This structure holds parameters for the coding behavior.
 * The values are roughly divided into two parts:
 * 1. The parameters of the code itself (k,n,r,n_p,code_start)
 * 2. The parameters for the retransmission rounds (n_rr, n_rr_max, n_cc)
 *
 * INVARIANTS:
 * - n = k+r
 * - n_p <= n-k
 * - code_start >= n_p
 * - code_start <= n-k
 *
 * - n_rr <= n_rr_max
 * - sizeof the n_cc array = n_rr_max
 */
__NOT_THREAD_SAFE __PCLASS struct params_code {
	__PRIVATE
	struct semaphore lock;	/**<Lock to synchronize access.*/

	// 1. part
	int k;		/**<#Data packets in one block.*/
	int r;		/**<#Redundancy packets in one block.*/
	int n;		/**<#Overall packets in one block.*/
	int n_p;	/**<#Redundancy packets that are send with the data in the first round.*/
	int code_start;	/**<#Redundancy packets coded when block is created.*/

	// 2. part
	int n_rr;	/**<#Retransmission rounds for this code.*/
	int n_rr_max;	/**<Maximum # of retransmission rounds for this network.*/
	int *n_cc;	/**<Multiplier for the redundancy send in each round.(array with at most
			 *  n_rr_max entries)*/

	// 3. part
  /*This are not really network parameters rather constraints for them.*/
  atomic_t D_target;  	/**<The given maximum target delay of a receiver in ms.*/
  atomic_t PLR_target;  	/**< The PLR that has to be achieved for a receiver.*/
  atomic_t D_Fec;         /**< Maximum FEC delay.*/
  atomic_t D_C;         /**< Maximum block coding delay.*/
  atomic_t D_Req;         /**< Request timer for loss notifications*/
  atomic_t D_fb;         /**< Delay of receivers sending the feedback.*/
  atomic_t BW_est;        /**< Estimated available bandwidth*/
  atomic_t Scheduled_RI;  /**< Redundancy bytes scheduled for sending*/

};

/**
 * Initialise coding parameters.
 */
int prrtcpar_init_codeparams(struct params_code *params);

/**
 * Zero out the coding parameters, so they can be seen as not set yet.
 */
void prrtcpar_zero_codeparams(struct params_code *params);
void prrtcpar_free_codeparams(struct params_code *params);

/**
 * Are the coding parameters already set(means not zero).
 */
int prrtcpar_already_set(struct params_code *params);

/**
 * Copy a set of coding parameters.
 * \param params the coding parameters to copy to.
 * \param cpar the coding parameters that are copied.
 */
int prrtcpar_copy_code(struct params_code *params, struct params_code *cpar);

/**
 * This function can be used to update the coding parameters (NOT READY TO USE).
 * To implement this functionality you need to implement the function update_params in param_code.c.
 * \param netparams the set of network parameters that is needed to compute a code.
 * \param locparams the set of local parameters that is needed to compute a code.
 */
void prrtcpar_update_params(struct params_code *params, struct params_network *netparams);

/**
 * Set the values of the parameters in the first part of the structure.
 * The function will check the values for the invariants stated above.
 * \return 0 on success, -1 on failure
 */
int prrtcpar_set_coding_params(struct params_code *params, int k, int r, int n_p, int code_start);

/**
 * Set the values of the parameters in the second part of the structure.
 * The function can checks for the validity of the parameters where possible. n_rr should also be
 * the size of the array n_ccs.
 * These values are strongly related to t_replace so remember to reset t_replace when changing theses
 * values.
 * \return 0 on success, -1 on failure
 */
int prrtcpar_set_retransmission_rounds(struct params_code *params, int n_rr, int *n_ccs);

/**
 * Set the values of the coding parameters, retransmission rounds and D_target as well as D_FEC,
 * D_REQ and D_fb together in a block.
 * The function will check the values for the invariants stated above.
 * The function can checks for the validity of the parameters where possible. n_rr should also be
 * the size of the array n_ccs.
 * These values are strongly related to t_replace so remember to reset t_replace when changing theses
 * values.
 * \return 0 on success, -1 on failure
 */
int prrtcpar_set_ahec_params(struct params_code *params,
                             int d_target, int d_fec, int d_c, int d_req, int d_fb, int k, int n_p, int n_cc, int *n_ccs);

/**
 * Returns the value for the number of data packets for a block.
 */
int prrtcpar_get_k(struct params_code *params);

/**
* Returns the number of redundancy packets for one block.
*/
int prrtcpar_get_r(struct params_code *params);

/**
 * Returns the overall number of packets for one block.
 */
int prrtcpar_get_n(struct params_code *params);

/**
 * Returns the number of redundancy packets that should be send in the first transmission.
 */
int prrtcpar_get_n_p(struct params_code *params);

/**
 * Returns the number of retransmission rounds for this code.
 */
int prrtcpar_get_n_rr(struct params_code *params);

/**
 * Get the multiplyer for the number of replacements in a certain round.
 * The rounds start with 1 (n_cc for round zero, the data transmission does not exist).
 */
int prrtcpar_get_n_cc(struct params_code *params, int retr_round);

/**
 * Returns the number of redundancy packets that should be coded on block creation.
 */
int prrtcpar_get_code_start(struct params_code *params);


/**
 * Initialise the parameter values.
 */
void prrtcpar_reinit_locparams(struct params_code *params);



/**
 * Retrieve the maximum target delay of a receiver.
 */
int prrtcpar_get_D_target(struct params_code *params);

/**
* Set the maximum target PLR of a receiver.
*/
void prrtcpar_set_PLR_target(struct params_code *params, int plr_target);

/**
* Retrieve the maximum target PLR of a receiver.
*/
int prrtcpar_get_PLR_target(struct params_code *params);


/**
 * Retrieve the maximum FEC delay.
 */
int prrtcpar_get_D_Fec(struct params_code *params);

/**
 * Retrieve the maximum block coding delay.
 */
int prrtcpar_get_D_C(struct params_code *params);

/**
 * Retrieve the request timer.
 */
int prrtcpar_get_D_Req(struct params_code *params);

/**
 * Retrieve the Delay of receivers sending feedbacks.
 */
int prrtcpar_get_D_fb(struct params_code *params);

/**
 * Set the estimated available bandwidth.
 */
void prrtcpar_set_BW_est(struct params_code *params, int bw_est);

/**
 * Retrieve the estimated available bandwidth.
 */
int prrtcpar_get_BW_est(struct params_code *params);

/**
 * Set the estimated available bandwidth.
 */
void prrtcpar_update_Scheduled_RI(struct params_code *params, int RI_increment);

/**
 * Retrieve the estimated available bandwidth.
 */
int prrtcpar_get_Scheduled_RI(struct params_code *params);

#endif /*PARAM_CODE_H*/

