/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "prrt_packet.h"
#include <linux/slab.h>
#include <linux/jiffies.h>
#include "defines.h"

#define DEBUG_NAME "PRRT_PACKET"

static struct kmem_cache *packet_cache;
static struct semaphore packet_cache_lock;
//int prrtpackets = 0;

inline int prrt_packet_is_preredundancy(struct prrt_packet_container *pcont)
{
  if((pcont->pack.data.header.packtype & 0xF) == ((u8)packtype_prered & 0xF))
    return 1;

  return 0;
}

inline int prrt_packet_is_redundancy(struct prrt_packet_container *pcont)
{
  if((pcont->pack.data.header.packtype & 0xF) == ((u8)packtype_red & 0xF) ||
     prrt_packet_is_preredundancy(pcont))
    return 1;

  return 0;
}

inline int prrt_packet_is_data(struct prrt_packet_container *pcont)
{
  if((pcont->pack.data.header.packtype & 0xF) == ((u8)packtype_data & 0xF))
    return 1;

  return 0;
}

inline int prrt_packet_is_redata(struct prrt_packet_container *pcont)
{
  if((pcont->pack.data.header.packtype & 0xF) == ((u8)packtype_redata & 0xF))
    return 1;

  return 0;
}

inline int prrt_packet_is_replacement(struct prrt_packet_container *pcont)
{
  return (prrt_packet_is_redata(pcont) || prrt_packet_is_redundancy(pcont));
}

inline int prrt_packet_may_delete(struct prrt_packet_container *pcont, u32 d_target, struct prrt_time_state *time_state)
{
  u32 timestamp = 0;
  u32 now = prrt_get_prrttime(time_state);
  u32 sum = 0;
  char comp = 0;

  u32 delay = pcont->pack.data.data_header.pack_timeout + MAYDELPACK_DELAY_ADD;

  /*decoded packets dont have the right timestamp for this decision
   *just delete them*/
  if(!prrt_packet_is_data(pcont))
    return 0;

  //timestamp = pcont->time.creation_time;
  timestamp = pcont->pack.data.data_header.timestamp;
  sum = (timestamp+delay)%MAX_TIMESTAMP;
  comp = prrt_compare_timestamps(time_state, sum, now);

//  PDEBUG(MSG_COMMON, "%s: cmp(%u, %u) = %d\n", DEBUG_NAME, sum,
//  now, comp);
//  PDEBUG(MSG_COMMON, "%s: arrival time: %d\n", DEBUG_NAME,
//    timestamp);
//  PDEBUG(MSG_COMMON, "%s: now: %d\n", DEBUG_NAME,
//    now);
//  PDEBUG(MSG_COMMON, "%s: comp timestamp: %d\n", DEBUG_NAME,
//    sum);
//  PDEBUG(MSG_COMMON, "%s: delay: %d\n", DEBUG_NAME, 3*d_target);

  if(comp == -1){
		PDEBUG(MSG_FINAL, "%s: May delete packet with timestamp %u, at prrt time %u\n", DEBUG_NAME, 
			pcont->pack.data.data_header.timestamp, prrt_get_prrttime(time_state));
    return 1;
  }else
    return 0;
}

inline int prrt_packet_may_read(struct prrt_packet_container *pcont, u32 d_target, struct prrt_time_state *time_state)
{
  u32 timestamp = 0;
  u32 now = prrt_get_prrttime(time_state);
  u32 sum = 0;
  char comp = 0;
  
  u32 delay = pcont->pack.data.data_header.pack_timeout + MAYREAD_DELAY_ADD;

  if(!prrt_packet_is_data(pcont))
    return 1;

  //timestamp = pcont->time.creation_time;
  timestamp = pcont->pack.data.data_header.timestamp;
  sum = (timestamp+delay)%MAX_TIMESTAMP;
  comp = prrt_compare_timestamps(time_state, sum, now);

  if(comp == -1){
		PDEBUG(MSG_FINAL, "%s: May read packet with timestamp %u, at prrt time %u\n", DEBUG_NAME, 
			pcont->pack.data.data_header.timestamp, prrt_get_prrttime(time_state));
    return 1;
  }else
    return 0;
}
int prrt_packet_cache_init(void)
{
  init_MUTEX(&packet_cache_lock);
  packet_cache = kmem_cache_create("packet_cache",
          MAX_PRRT_PACKET_CONT_SIZE,
          0, 0,
          NULL);

  if(!packet_cache) {
    PERROR("%s: Could not create packet cache\n", DEBUG_NAME);
    return -1;
  }

  return 0;
}

void prrt_packet_cache_uninit(void)
{
  kmem_cache_destroy(packet_cache);
}

void *prrt_packet_alloc(struct prrt_time_state *time_state)
{
  struct prrt_packet_container *pcont;

  PDEBUG(MSG_ALLOC, "%s: Allocating packet\n", DEBUG_NAME);
  down(&packet_cache_lock);
  pcont = kmem_cache_alloc(packet_cache, GFP_KERNEL);
  //prrtpackets++;
  //PDEBUG(MSG_TRANSPORT, "%s: Allocating packet (count %d)\n", DEBUG_NAME, prrtpackets);
  up(&packet_cache_lock);

  PDEBUG(MSG_ALLOC, "%s: Packet allocated\n", DEBUG_NAME);
  /*zero out packet for coding purposes*/
  if(!pcont) {
    PERROR("%s: Could not allocate memory for packet\n", DEBUG_NAME);
  } else {
    bzero(pcont, MAX_PRRT_PACKET_CONT_SIZE);
    PDEBUG(MSG_ALLOC, "%s: Setting creation time\n", DEBUG_NAME);
    pcont->time.creation_time = prrt_get_prrttime(time_state);
  }

  return pcont;
}

int prrt_packet_set_fields(struct prrt_packet_container *pcont, int packetno, int index, unsigned long timestamp, int group_rtt,
    int dlen, int packtype)
{
  /*timestamps of the packet need to be in real time(e.g. ms)*/
  //pcont->packet_status = NOTUSED;
  pcont->pack.data.header.packtype = (pcont->pack.data.header.packtype & ~0xF) | ((u8)packtype & 0xF);
  //pcont->pack.data.header.index = (u8)index;
  pcont->pack.data.header.priority = (pcont->pack.data.header.priority & ~0xF);
  //pcont->pack.data.header.flags = 0;

  pcont->pack.data.header.seqnum = (u16)packetno;
  //pcont->pack.data.data_header.pack_timeout = 0;
  //pcont->pack.data.data_header.fb_timeout = 0;
  pcont->pack.data.data_header.timestamp = (u32)timestamp;
  pcont->pack.data.data_header.group_rtt = (u16)group_rtt;

  //packet length just set within packet container, use UDP length field for transmission
  pcont->payload_size = (u16)dlen;

  return PRRT_HEADER_SIZE+PRRT_DATA_HEADER_SIZE+dlen;
}

void prrt_packet_destroy(struct prrt_packet_container **pcont)
{
  if(!(*pcont))
    return;

  down(&packet_cache_lock);
  kmem_cache_free(packet_cache, *pcont);
  //prrtpackets--;
  up(&packet_cache_lock);
  *pcont = 0;
  PDEBUG(MSG_ALLOC, "%s: Packet deleted\n", DEBUG_NAME);
}

char *prrt_packet_get_data(struct prrt_packet_container *pcont)
{
  if(prrt_packet_is_redundancy(pcont)) {
    return &pcont->pack.red.data;
  } else {
    return &pcont->pack.data.data;
  }
}

void prrt_packet_hton(struct prrt_packet_container *pcont)
{
  if(prrt_packet_is_redundancy(pcont)) {
    pcont->pack.red.header.seqnum = htons(pcont->pack.red.header.seqnum);
    pcont->pack.red.red_header.sn_base = htons(pcont->pack.red.red_header.sn_base);
    //pcont->pack.red.red_header.t_replace = htons(pcont->pack.red.red_header.t_replace);
  } else {
  //pcont->pack.data.header.index = htons(pcont->pack.data.header.index);
  pcont->pack.data.header.seqnum = htons(pcont->pack.data.header.seqnum);
  pcont->pack.data.data_header.group_rtt = htons(pcont->pack.data.data_header.group_rtt);
  pcont->pack.data.data_header.pack_timeout = htons(pcont->pack.data.data_header.pack_timeout);
  pcont->pack.data.data_header.dec_timeout = htons(pcont->pack.data.data_header.dec_timeout);
  pcont->pack.data.data_header.fb_timeout = htons(pcont->pack.data.data_header.fb_timeout);
  pcont->pack.data.data_header.timestamp = htonl(pcont->pack.data.data_header.timestamp);
  //pcont->pack.data.data_header.payload_size = htons(pcont->pack.data.data_header.payload_size);
  }
}

void prrt_packet_ntoh(struct prrt_packet_container *pcont)
{
  if(prrt_packet_is_redundancy(pcont)) {
    pcont->pack.red.header.seqnum = ntohs(pcont->pack.red.header.seqnum);
    pcont->pack.red.red_header.sn_base = ntohs(pcont->pack.red.red_header.sn_base);
    //pcont->pack.red.red_header.t_replace = ntohs(pcont->pack.red.red_header.t_replace);
  } else {
  //pcont->pack.data.header.index = ntohs(pcont->pack.data.header.index);
  pcont->pack.data.header.seqnum = ntohs(pcont->pack.data.header.seqnum);
  pcont->pack.data.data_header.group_rtt = ntohs(pcont->pack.data.data_header.group_rtt);
  pcont->pack.data.data_header.pack_timeout = ntohs(pcont->pack.data.data_header.pack_timeout);
  pcont->pack.data.data_header.dec_timeout = ntohs(pcont->pack.data.data_header.dec_timeout);
  pcont->pack.data.data_header.fb_timeout = ntohs(pcont->pack.data.data_header.fb_timeout);
  pcont->pack.data.data_header.timestamp = ntohl(pcont->pack.data.data_header.timestamp);
  //pcont->pack.data.header.payload_size = ntohs(pcont->pack.data.header.payload_size);
  }
}

inline u16 prrt_increment_packetno_locked(u16 *packetno, u16 i, struct semaphore *lock)
{
  u16 oldno = 0;

  down(lock);
  oldno = *packetno;
  *packetno += i;
  up(lock);

  return oldno;
}

inline u16 prrt_increment_packetno(u16 packetno, u16 i)
{
//  int j = 0;
  /*exclude zero as packetnumber, wrap arounds are intended*/
//  for(j=0; j<i; j++) {
//    packetno++;
// //     if(packetno == 0)
// //       packetno++;
//  }
  packetno += i;

  return packetno;
}

inline u16 prrt_decrement_packetno(u16 packetno, u16 i)
{
//  int j = 0;
  /*exclude zero as packetnumber, wrap arounds are intended*/
//  for(j=0; j<i; j++) {
//    packetno--;
// //     if(packetno == 0)
// //       packetno--;
//  }
  packetno -= i;

  return packetno;
}

inline int prrt_comp_packetnos(u16 packetno1, u16 packetno2)
{
  s16 diff = 0;
  diff = (s16)packetno1 - (s16)packetno2;

  if(diff < 0) {
    /*packetno2 > packetno1*/
    return -1;
  } else if(diff == 0) {
    /*packetno2 == packetno1*/
    return 0;
  } else {
    /*packetno2 < packetno1*/
    return 1;
  }
}

inline u16 prrt_get_max_packetno(u16 packetno1, u16 packetno2) {
  if(prrt_comp_packetnos(packetno1, packetno2) < 0) {
    return packetno2;
  } else {
    return packetno1;
  }
}

inline u16 prrt_dist_packetnos(u16 packetno1, u16 packetno2)
{
  u16 diff = abs(packetno1-packetno2);

  if(diff > HALF_PACKETNO)
    diff = MAX_PACKETNO-diff+1;

  return diff;
}

inline int prrt_packetno_lies_in_limits(u16 packetno, u16 start, u16 end)
{
  return (prrt_comp_packetnos(start, packetno) <= 0 &&
    prrt_comp_packetnos(packetno, end) <= 0);
}

inline void prrt_packet_print_header(struct prrt_packet_container *pcont, int level)
{
  PDEBUG(level, "%s: Packet Header :\n", DEBUG_NAME);
  PDEBUG(level, "%s: ---------------\n", DEBUG_NAME);
  PDEBUG(level, "%s: packtype = %u\n", DEBUG_NAME, pcont->pack.data.header.packtype & 0xF);
  PDEBUG(level, "%s: priority= %u\n", DEBUG_NAME, pcont->pack.data.header.priority & 0xF);
  PDEBUG(level, "%s: index = %u\n", DEBUG_NAME, pcont->pack.data.header.index);
  PDEBUG(level, "%s: seqnum = %u\n", DEBUG_NAME, pcont->pack.data.header.seqnum);
  PDEBUG(level, "%s: timestamp = %f\n", DEBUG_NAME, pcont->pack.data.data_header.timestamp);
  PDEBUG(level, "%s: group_rtt = %u\n", DEBUG_NAME, pcont->pack.data.data_header.group_rtt);
  PDEBUG(level, "%s: pack_timeout = %u\n", DEBUG_NAME, pcont->pack.data.data_header.pack_timeout);
  PDEBUG(level, "%s: dec_timeout = %u\n", DEBUG_NAME, pcont->pack.data.data_header.dec_timeout);
  PDEBUG(level, "%s: fb_timeout = %u\n", DEBUG_NAME, pcont->pack.data.data_header.fb_timeout);
//  PDEBUG(level, "%s: data_size = %u\n", DEBUG_NAME, pcont->pack.data.header.payload_size);
//  PDEBUG(MSG_PACK, "%s: timestamp = %X", DEBUG_NAME, pcont->pack.data.header.timestamp);
  if(prrt_packet_is_redundancy(pcont)) {
    PDEBUG(level, "%s: sn_base = %u\n", DEBUG_NAME, pcont->pack.red.red_header.sn_base);
    PDEBUG(level, "%s: k = %u\n", DEBUG_NAME, pcont->pack.red.red_header.k);
    PDEBUG(level, "%s: n = %u\n", DEBUG_NAME, pcont->pack.red.red_header.n);
    //PDEBUG(level, "%s: index = %u\n", DEBUG_NAME, pcont->pack.red.header.index);
    //PDEBUG(level, "%s: t_replace = %u\n", DEBUG_NAME, pcont->pack.red.red_header.t_replace);
  }
}

void prrt_packet_redundancy_to_data(struct prrt_packet_container *pcont) {
    char *dataDataAddress = &pcont->pack.data.data;
    char *redDataAddress = &pcont->pack.red.data;
    u16 dataSize = pcont->payload_size;
    int i = 0;

    if(!prrt_packet_is_redundancy(pcont))
        return;

    for(i = 0; i<dataSize; i++){
        dataDataAddress[i] = redDataAddress[i];
    }

    pcont->pack.data.header.packtype = (pcont->pack.data.header.packtype & ~0xF) | ((u8)packtype_data & 0xF);
}

void LogRedundancyPacket(struct prrt_packet_container *pcont);
void LogDataPacket(struct prrt_packet_container *pcont);

void prrt_packet_log_sending(struct prrt_packet_container *pcont)
{
  if(prrt_packet_is_redundancy(pcont)) {
    PLOGPACK("Sending Redundancy Packet:\n");
    LogRedundancyPacket(pcont);
  } else if(prrt_packet_is_data(pcont)){
    PLOGPACK("Sending Data Packet:\n");
    LogDataPacket(pcont);
  } else {
    PLOGPACK("Sending Retransmission Data Packet:\n");
    LogDataPacket(pcont);
  }
}

void LogRedundancyPacketHeader(struct prrt_red_header *red_header);

void LogRedundancyPacket(struct prrt_packet_container *pcont)
{
  if(!prrt_packet_is_redundancy(pcont))
    return;

  prrt_header_log_general(&pcont->pack.red.header);

  LogRedundancyPacketHeader(&pcont->pack.red.red_header);
}

void prrt_header_log_general(struct prrt_header *header)
{
  PLOGPACK("------------------------------------------\n");
  PLOGPACK("Packettype: %d\n", header->packtype & 0xF);
  PLOGPACK("Packet Index: %d\n", header->index);
  //PLOGPACK("Payload Size: %d\n", header->payload_size);
  PLOGPACK("------------------------------------------\n");
}

void LogRedundancyPacketHeader(struct prrt_red_header *red_header)
{
  PLOGPACK("First Data Packet: %d\n", red_header->sn_base);
  PLOGPACK("k: %d\n", red_header->k);
  PLOGPACK("n: %d\n", red_header->n);
  //PLOGPACK("index: %d\n", red_header->index);
  //PLOGPACK("t_replace: %d\n", red_header->t_replace);
  PLOGPACK("------------------------------------------\n");
}

void LogDataPacket(struct prrt_packet_container *pcont)
{
  if(!prrt_packet_is_data(pcont) && !prrt_packet_is_redata(pcont)) {
    PERROR("Trying to log packet with wrong function\n");
    return;
  }

  prrt_header_log_general(&pcont->pack.data.header);
}

void prrt_packet_log_receiving(struct prrt_packet_container *pcont)
{
  if(prrt_packet_is_redundancy(pcont)) {
    PLOGPACK("Receiving Redundancy Packet:\n");
    LogRedundancyPacket(pcont);
  } else if(prrt_packet_is_data(pcont)){
    PLOGPACK("Receiving Data Packet:\n");
    LogDataPacket(pcont);
  } else {
    PLOGPACK("Receiving Retransmission Data Packet:\n");
    LogDataPacket(pcont);
  }
}

void prrt_packet_print_hexdata(struct prrt_packet_container *pcont)
{
  unsigned short *data = 0;
  int i = 0;

  data = (unsigned short *)prrt_packet_get_data(pcont);
  PDEBUG(PRINT_PACKETHEX, "Packet Number %u\n", pcont->pack.data.header.seqnum);
  PDEBUG(PRINT_PACKETHEX, "------------------------------------\n");
  for(i=0; i<pcont->payload_size/2; i+=8) {
      PDEBUG(PRINT_PACKETHEX, "%4x %4x %4x %4x %4x %4x %4x %4x\n",
        data[i], data[i+1], data[i+2], data[i+3],
        data[i+4], data[i+5], data[i+6], data[i+7]);
  }
  PDEBUG(PRINT_PACKETHEX, "------------------------------------\n");
}

void prrt_packet_log_omission(struct prrt_packet_container *pcont)
{
  if(prrt_packet_is_redundancy(pcont)) {
    PDEBUG(MSG_PACKOMMIT, "%s: Omitting Redundancy Packet:\n", DEBUG_NAME);
    prrt_packet_print_header(pcont, MSG_PACKOMMIT);
  } else if(prrt_packet_is_data(pcont)){
    PDEBUG(MSG_PACKOMMIT, "%s: Omitting Data Packet:\n", DEBUG_NAME);
    prrt_packet_print_header(pcont, MSG_PACKOMMIT);
  } else {
    PDEBUG(MSG_PACKOMMIT, "%s: Omitting Retransmission Data Packet:\n", DEBUG_NAME);
    prrt_packet_print_header(pcont, MSG_PACKOMMIT);
  }
}

struct tsPacket {
  u8 syncByte;
  u16 bitsAndPid;
  u8 TscAfcCounter;
  u8 data;
};

int prrt_packet_data_starts_with_sync_byte(struct prrt_packet_container *pcont)
{
  struct tsPacket *tspack = (struct tsPacket *)prrt_packet_get_data(pcont);

  return (tspack->syncByte == 0x47);
}
