/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#ifndef PRRT_PACKET_H
#define PRRT_PACKET_H

#include <linux/types.h>
#include "prrt_time.h"

#define packtype_data	0
#define packtype_redata	1
#define packtype_red	2
#define packtype_fb	3
#define packtype_prered	4
#define packtype_chanfb	5

/**
 * This is a header for a prrt packet.
 * These header are used by both the data and redundancy packets. To discriminate them the packtype
 * field is used.
 * The header also includes the number of a packet where number spaces for data and redundancy
 * packets are different.
 * The header also holds the size of the data fields in the different packet types.
 */
struct prrt_header {
	u8	packtype:4,	/**< This does for example the job of marking redundancy packets.*/
		priority:4;	/**< Packet priority*/
	u8	index;		/**< Packet index within coding block (0 <= index < k)*/
	u16	seqnum;
};

struct prrt_data_header {
	u32	timestamp;	/**< Time when this packet was allocated (in ms).*/
	u16	group_rtt;	/**< Group RTT estimated by sender (in ms)*/
	u16	pack_timeout;	/**< Time after which this packet expires (in ms).*/
	u16	dec_timeout;	/**< First feedback is to be scheduled at decoding timeout (in ms).*/
	u16	fb_timeout;      /**< Time after which receiver feedback for this packet is expected (in ms).*/
};

#define MAX_PACKETNO 65535
#define HALF_PACKETNO 32767

/**
 * Additional header for a redundancy packet.
 * Additional to the normal header a redundancy packet contains this header that holds information
 * about the block it is contained in.
 */
struct prrt_red_header {
//	u16	t_replace;	/**<The worst replace time that the sender estimates.*/
	u16	sn_base;	/**<The number of the first data packet in the block.*/
	u8	n;		/**<The overall number of packets in the block.*/
	u8	k;		/**<The number of information packets in the block.*/
//	u8	index;		/**<This is the index of the packet in the block.*/
//	u8	round;		/**<The retransmission round this red_packet belongs to.*/
};

#define PRRT_HEADER_SIZE sizeof(struct prrt_header)
#define PRRT_DATA_HEADER_SIZE sizeof(struct prrt_data_header)
#define PRRT_RED_HEADER_SIZE sizeof(struct prrt_red_header)
#define MAX_DATA_SIZE 1500 /*maximum size of incoming packet data in bytes*/
#define MAX_PRRT_PACKET_SIZE PRRT_HEADER_SIZE + PRRT_DATA_HEADER_SIZE + MAX_DATA_SIZE
#define MAX_PRRT_PACKET_CONT_SIZE sizeof(struct prrt_packet_container) + MAX_DATA_SIZE

/**
 * This structure represents a data packet of the prrt protocol.
 * The header is described with the prrt_header struct.
 * The data field that is allocatet as a char here is in reality an array of chars that starts
 * directly behind the header. This means you have a continuous memory block for one packet.
 * The information about the size of the actually saved data will be stored in the header in the
 * field payload_size.
 * That means a packet is longer than sizeof(struct prrt_data_packet).
 */
struct prrt_data_packet {
	struct prrt_header 	header;	/**< The header of the packet.*/
	struct prrt_data_header	data_header;	/**< The header of the data packet.*/
	char			data;	/**< The first field of the data in the packet.*/
};

/**
 * Structure that represents a redundancy packet for the prrt protocol.
 * Such a redundancy packet holds the normal header, an extra header and the coded redundancy
 * that can be used to regain lost packet information.
 * You can get the prrt_red_packet from a prrt_data_packet by casting.
 */
struct prrt_red_packet {
	struct prrt_header 	header;	/*the header of the packet*/
	struct prrt_red_header	red_header;/*the header for the redundancy*/
	char			data;	/*the data in the packet*/
};

/**
 * This structure was introduced to keep the program extendable.
 * All internal functions don't work on raw packet but on this structure that can hold additional
 * information to administrate the packets.
 */
struct prrt_packet_container {
	u16	payload_size;
	union {
		u32 creation_time;	/*time this packet was allocated in ms*/
		u32 transmit_time;	/*time this packet was transmitted*/
	} time;
	union {
		struct prrt_data_packet	data;
		struct prrt_red_packet	red;
	} pack;
};

/**
 * Initialise the cache structure that is used for packets.
 */
int prrt_packet_cache_init(void);

/**
 * Uninitialise the cache structure for packets.
 */
void prrt_packet_cache_uninit(void);

/**
 * Allocate a packet.
 * \return (struct prrt_packet_container *) on success, 0 on failure.
 */
void *prrt_packet_alloc(struct prrt_time_state *time_state);

/**
 * This initialises a preallocated packet with the given values.
 * \param packetno packet number to set for the packet.
 * \param timestamp timestamp to set for the packet.
 * \param dlen payload_size of the packet data.
 * \param packtype the type of the packet.
 */
int prrt_packet_set_fields(struct prrt_packet_container *pcont, int packetno, int index, unsigned long timestamp, int group_rtt,
        int dlen, int packtype);

/**
 * Deallocate a packet.
 * Does nothing if pcont is NULL.
 */
void prrt_packet_destroy(struct prrt_packet_container **pcont);

/**
 * Returns the pointer to the data are contained in this packet.
 * Because we decode packets but don't want to copy the packets into another struct we will have
 * packets in the same list with different alignments for the data area.
 * Always use this function to retrieve a pointer to the data area.
 */
char *prrt_packet_get_data(struct prrt_packet_container *pcont);

/*
 * Test packet for certain type.
 * Return 1 if it has this type, 0 if not.
 */
extern int prrt_packet_is_redundancy(struct prrt_packet_container *pcont);
extern int prrt_packet_is_preredundancy(struct prrt_packet_container *pcont);
extern int prrt_packet_is_data(struct prrt_packet_container *pcont);
extern int prrt_packet_is_redata(struct prrt_packet_container *pcont);

/**
 * Test if the packet is sent as a replacement, e.g. redata or redundancy.
 */
extern int prrt_packet_is_replacement(struct prrt_packet_container *pcont);

/*
 * Test if packet may already be deleted/read by the receiver/user.
 * Returns 1 for yes and 0 for no.
 */
extern int prrt_packet_may_delete(struct prrt_packet_container *pcont, u32 d_target, struct prrt_time_state *time_state);
extern int prrt_packet_may_read(struct prrt_packet_container *pcont, u32 d_target, struct prrt_time_state *time_state);

/**
 * Convert all variables in the packet to network byte order.
 */
void prrt_packet_hton(struct prrt_packet_container *pcont);

/**
 * Convert all variables in the packet to host byte order.
 */
void prrt_packet_ntoh(struct prrt_packet_container *pcont);

/**
 * This function increments a packetnumber and uses a semaphore too synchronize access to the
 * variable containing the packetnumber.
 * \param lock the semaphore used to lock access to packetno.
 * \return the OLD value of the packetnumber.
 */
extern u16 prrt_increment_packetno_locked(u16 *packetno, u16 i, struct semaphore *lock);

/**
 * This function increments a packetnumber by a certain value.
 * \return the NEW value of the packetnumber.
 */
extern u16 prrt_increment_packetno(u16 packetno, u16 i);

/**
 * This function decrements a packetnumber by a certain value.
 */
extern u16 prrt_decrement_packetno(u16 packetno, u16 i);

/**
 * This function compares two packetnumbers.
 * It is important to always use this function when compairing packetnumbers because a wrap around
 * of the packetnumbers could lead to nasty bugs(e.g. when lists should be ordered according to
 * packetnumbers)
 * \param packetno1 one of the packetnumbers.
 * \param packetno2 one of the packetnumbers.
 * \return -1 if packetno1 < packetno2 | 0  if packetno1 = packetno2 | 1  if packetno1 > packetno2
 */
extern int prrt_comp_packetnos(u16 packetno1, u16 packetno2);

/**
 * Returns the temporaryly seen bigger sequence number.
 */
extern u16 prrt_get_max_packetno(u16 packetno1, u16 packetno2);

/**
 * Compute the distance between two packetnumbers.
 */
extern u16 prrt_dist_packetnos(u16 packetno1, u16 packetno2);

/**
 * Decide if a sequence number lies between two other sequence numbers (start<=packetno<=end).
 */
extern int prrt_packetno_lies_in_limits(u16 packetno, u16 start, u16 end);

/**
 * Print out the header of a packet.
 * For redundancy packets prints also the red_header.
 */
extern void prrt_packet_print_header(struct prrt_packet_container *pcont, int level);

/**
 * Reformats a redundancy packet to a data packet.
 * Does nothing for data packets.
 */
void prrt_packet_redundancy_to_data(struct prrt_packet_container *pcont);

void prrt_packet_log_sending(struct prrt_packet_container *pcont);
void prrt_packet_log_receiving(struct prrt_packet_container *pcont);
void prrt_header_log_general(struct prrt_header *header);

void prrt_packet_print_hexdata(struct prrt_packet_container *pcont);
void prrt_packet_log_omission(struct prrt_packet_container *pcont);

int prrt_packet_data_starts_with_sync_byte(struct prrt_packet_container *pcont);

#endif /*PRRT_PACKET_H*/

