/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#ifndef PRRT_SOCKET_H
#define PRRT_SOCKET_H

#include "kui/kui_option.h"
#include "lists/prrt_out_list.h"
#include "lists/prrt_inc_list.h"
#include "prrt_block.h"
#include "receive_data.h"
#include "send_data.h"
#include "send_replacements.h"
#include "prrt_time.h"
#include "parameters/param_code.h"
#include "parameters/param_local.h"   // For loc_params_in. remove it if loc_params_in isn't used.

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26)
#include <asm/semaphore.h>
#else
#include <linux/semaphore.h>
#endif
#include <linux/socket.h>
#include <linux/in.h>
#include <net/sock.h>

/**
 * This is a type to describe the status of a worker thread.
 */
enum worker_thread_status_t {
  WORKING = 1,
  RELAY_WORKING, // indicates the receiving socket is used in relay scenarios. It prevents from the case where receving socket reading out the data and writing them to the local interface (lo) for the user space.
  ERROR,
  DOWN
};

/**
 * This structure represents a socket for a prrt connection.
 * It is perhaps the singel most important structure in this module around which the complete
 * program is centered.
 * Such sockets are created and administrated by the socket_manager. Their "system calls" are
 * executed through the kui (kernelspace-userspace interface).
 * A prrt socket holds two separate udp sockets, one for feedback and one for data.
 * - Sending part:
 *   A sending socket has two tasks:
 *   1. Sending data.
 *   2. Receiving feedback and sending replacements.
 *   All data that should be send is grouped in blocks (s. prrt_block) and send packet per packet
 *   to the receiver. When feedback arrives the sender searches for the block that contains the
 *   missing packet and sends the next replacement for this block.
 * - Receiving part:
 *   A receiving socket has two tasks:
 *   1. Receiving data and feedback.
 *   2. Informing the sender over lost packets via feedback.
 *   The receiver gathers all incoming packets and checks in regular intervalls if some packets are
 *   missing. If so it sends feedback to the sender that contains i.a. the number of the missing
 *   packets. To suppress the feedback in multicast scenarios the receiver sends the feedback if
 *   no other receivers sent the same feedback before.
 *   For arriving replacements a receiver needs to check for the possibility of decoding/restoring
 *   the block that the replacements belongs to.
 */
/*TODO*/__THREAD_SAFE __PCLASS struct prrt_socket {
        Socketno_t              socketno; /*the number of the socket in the procfs*/

  struct socket   *sock;  /*a container for the system socket*/
  //struct socket   *sockdbg; /*a container for the system socket*/
  struct socket   *sockfb;/*the socket for feedback*/
  struct semaphore  mutex;  /*mutex to lock this shared variable*/
  struct sockaddr_in  *host;  /*address of the host*/
  struct sockaddr_in  *hostfb;/*address of the host for feedback*/
  struct sockaddr_in  *peer;  /*address of the connected peer*/
  struct sockaddr_in  *peerdbg; /*address of the connected peer*/
  struct sockaddr_in  *peerfb;/*address for the feedback*/
  struct in_addr    host_addr;/*this holds the ip address of the host for identification*/

  int kui_port_read;  /*port for kernel user communication*/
  int kui_port_write; /*port for kernel user communication*/
  struct socket *kui_sock_read;/*UDP socket to user space process*/
  struct socket *kui_sock_write;/*UDP socket to user space process*/

  /*incoming packets lists*/
  struct prrt_inc_list  inc_packet_list; /*list for incoming data and redundancy packets*/

  /*lists for outgoing packets*/
  struct prrt_out_list  out_packet_list; /*structure for sending packets*/

  /*variables for sending packets*/
  enum worker_thread_status_t out_status;
  u16       out_pno;  /*packet number for next out packet*/

  struct delayed_work out_work;
  struct semaphore  out_lock; /*locks the variables above*/

  /*coding variables for sending packets*/
  struct prrt_coder *out_cod;
  struct semaphore  out_cod_lock; /*locking for the coder struct*/
	unsigned long time_last_packet; /**< Time when the last packet was sent.*/
	struct prrt_time_state send_time_state; /** Time state for the sending part of the socket.*/

  /*variables for receiving packets*/
  enum worker_thread_status_t poll_status;
  struct delayed_work poll_work;
  u16     poll_pno_fb;  /*packet number for next feedback packet*/
  struct semaphore  poll_lock;  /*locks the variables above*/
	struct prrt_time_state receive_time_state; /** Time state for the receiving part of the socket.*/

  /*variables for sending feedback*/
  struct delayed_work sendfb_work;
  struct semaphore  sendfb_lock;  /*locks the variables above*/

  /*variables for receiving feedback*/
  struct delayed_work recvfb_work;
  struct delayed_work sendrepl_work;
  struct semaphore  recvfb_lock;  /*locks the variables above*/

  /*variables for forwarding packets*/
  struct delayed_work readpkt_work;
  struct delayed_work writepkt_work;

  struct workqueue_struct *recv_queue;  /**< This thread will receive incoming data packets.*/
  struct workqueue_struct *recvfb_queue;  /**< This thread will receive incoming feedback.*/
  struct workqueue_struct *sendrepl_queue;/**< This thread will send replacements.*/
  struct workqueue_struct *send_queue;  /**< This thread will send data packets.*/
  struct workqueue_struct *sendfb_queue;  /**< This thread will send feedback.*/
  struct workqueue_struct *readpkt_queue; /**< This thread will read packets from the prrt stack*/
              /* and forwards it to the user space.*/
  struct workqueue_struct *writepkt_queue;/**< This thread will write packets into the prrt stack*/
              /* using payload data retrieved from user space.*/

  /*private data of the different files(freed on file destruction)*/
  void *control_pd;
  void *write_pd;   /*this is a dummy, no private data for this file type*/
  void *read_pd;    /*this is a dummy, no private data for this file type*/
  void *option_pd;
  void *stats_pd;
  //void *d_target_pd;
  void *plr_pd;
  //void *d_fec_pd;
  //void *codp_pd;
  //void *rtr_pd;
  void *ahec_pd;
  void *bwest_pd;
  void *buffb_pd;

  /*parameters for this socket*/
  //struct params_local loc_params_out; /*parameter set for outgoing connection*/
  struct params_local loc_params_in;  /*parameter set for incoming connection*/
  struct params_code cod_params;
  struct params_network net_params; /*group net params*/

  /*socket stats*/
  unsigned long delivered_packets;  /*# successfully delivered packets*/
  unsigned long delivered_data;
  unsigned long delivered_red;

  enum option_val_t    timeout;           /*should the socket stop receiving after a timeout*/
};

/**
 * Allocate a new prrt socket and initialise the complete structure.
 * \param msock this variable will hold a pointer to the new prrt socket on success.
 * \param kui_port the port for kernel user space communication.
 * \return 0 on success, < 0 on failure.
 */
int prrt_socket(struct prrt_socket **msock, int kui_port);

/**
 * This function binds a socket to the given addresses and the port.
 * \param sock_addr this is the address where the data packets are send to and this socket will
 *    listen on.
 * \param sock_port this is the port where the socket will listen for arriving packets.
 * \param peer_addr this address is the target address for the outgoing feedback. The port for
 *    feedback is sock_port+1.
 * \return 0 on success, <0 on failure.
 */
int prrt_bind(struct prrt_socket *msock, char *sock_addr, u_short sock_port, char *peer_addr);

/**
 * This function binds the socket of the receiver to the given address and port.
 *
 * \param sock_addr the address of the receiver, where the data packets are sent to
 *                  and this socket of the receiver listens on.
 * \param sock_port the port of the receiver, where the socket listens for arriving packets.
 * \return 0 on success, <0 on failure.
 */
int prrt_bind_as_recv(struct prrt_socket *msock, char *sock_addr, u_short sock_port);

/**
 * This function binds the socket of the sender to the given address and port,
 * and adds the socket into multicast group if the feedback is sent via multicast.
 *
 * \param sock_addr the address of the sender, where the feedback packets are sent to
 *                  and the socket of the sender listens on.
 * \param sock_port the port of the sender where the socket listens for arriving feedbacks.
 * \return 0 on success, <0 on failure.
 */
int prrt_bind_as_send(struct prrt_socket *msock, char *sock_addr, u_short sock_port);

/**
 * This function "connects" a socket to the given addresses and the port.
 * \param peer_addr this is the address where the data packets should be send to.
 * \param peer_port this is the port where the data packets should arrive on.
 * \param host_addr this is the address this socket will listen on for arriving feedback. The port
 *    for arrving feedback is peer_port+1.
 * \return 0 on success, <0 on failure.
 */
int prrt_connect(struct prrt_socket *msock, char *peer_addr, u_short peer_port, char *host_addr, int *kui_port);

/**
 * This function "connects" the socket of the sender to the given address and port of the receiver.
 *
 * It stores the address and the port of the receiver in prrt_socket of the sender.
 * It also binds a random free port for kernel user communication and returns the bound
 * port to the caller.
 *
 * \param peer_addr the address of the receiver, where the data packets are sent to.
 * \param peer_port the port of the receiver, where the data packets arrive on.
 * \param kui_port a pointer to the port, used to return the bound port of kernel-user communication to the caller.
 * \return 0 on success, <0 on failure.
 */
int prrt_connect_to_receiver(struct prrt_socket *msock, char *peer_addr, u_short peer_port, int *kui_port);

/**
 * This function "connects" the socket of the receiver to the given address and port of the feedback.
 *
 * It just stores the address and the port of the feedback in prrt_socket of the receiver,
 * and initialises the queues of recv and send feedback (polling), in the case of unicast feedback.
 * In the case of multicast feedback, it binds the feedback address and port to receive feedback,
 * in order to suppress the feedbacks.
 *
 * \param peerfb_addr the address where the feedback packets are sent to.
 * \param peerfb_port the port where the feedback packets arrive on.
 * \return 0 on success, <0 on failure.
 */
int prrt_connect_to_feedback(struct prrt_socket *msock, char *peerfb_addr, u_short peerfb_port);

/*No such functionality because we are almost connection less*/
//int prrt_accept(struct prrt_socket *msock, struct prrt_socket **new_msock);

/*No such functionality because we are almost connection less*/
// int prrt_listen(struct prrt_socket *msock, int backlog);

/**
 * This functions disables the possibility to receive or send packets on this socket and will
 * delete all allocated memory, e.g. data/redundancy packets, blocks, etc..
 */
void prrt_shutdown(struct prrt_socket *msock);

/**
 * Send data over the socket.
 * To send data a socket must be connected.
 * \param pcont the container for the data packet with the data to be sent already included.
 * \param len the number of bytes of data in pcont.
 * \return 0 on success, <0 on failure.
 */
int prrt_send(struct prrt_socket *msock, struct prrt_packet_container *pcont, size_t len);

/**
 * Listen for arriving packets on this socket.
 * To listen for packets a socket must be bound.
 * \param pcont this will hold a pointer to the packet container after returning.
 * \return number of data bytes received on success, <0 on failure.
 */
int prrt_recv(struct prrt_socket *msock, struct prrt_packet_container **pcont);

/**
 * Read packets from the prrt stack and forward them to the user space.
 */
void prrt_read_packet(struct work_struct *work);

/**
 * Writes packets from the user space into the prrt stack.
 */
void prrt_write_packet(struct work_struct *work);

/**
 * This sets an option of the socket.
 * \param opt this is the kind of option to set.
 * \param val this is the value for the option.
 * \return 0 on success.
 */
int prrt_set_option(struct prrt_socket *msock, enum option_t opt, int val);

/**
 * Returns the value of the option.
 * \param opt the option the value should be retrieved of.
 * \return the value of the option that is given.
 */
int prrt_get_option(struct prrt_socket *msock, enum option_t opt);

/**
 * Get the port of the socket on which the packets are received.
 */
int prrtsocket_get_receiveon_port(struct prrt_socket *msock);

/**
 * Get the port of the socket to which the packets are sent.
 */
int prrtsocket_get_sendto_port(struct prrt_socket *msock);

#endif /*PRRT_SOCKET_H*/
