/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "prrt_time.h"

#include <linux/sched.h>

#define DEBUG_NAME "PRRT_TIME"

static void UpdateTimeState(struct prrt_time_state *tstate, int mean_deviation, int clock_skew, int last_measurement, int virtual_time){
	atomic_set(&(tstate->mean_deviation), mean_deviation);
	atomic_set(&(tstate->clock_skew), clock_skew);
	atomic_set(&(tstate->last_measurement), last_measurement);
	atomic_set(&(tstate->virtual_time), virtual_time);
}

//static void SetMeanDeviation(struct prrt_time_state *tstate, int val) {
//atomic_set(&(tstate->mean_deviation), val);
//}

static int GetMeanDeviation(struct prrt_time_state *tstate) {
	return atomic_read(&(tstate->mean_deviation));
}

//static void SetClockSkew(struct prrt_time_state *tstate, int val) {
//	atomic_set(&(tstate->clock_skew), val);
//}
//
static int GetClockSkew(struct prrt_time_state *tstate) {
	return atomic_read(&(tstate->clock_skew));
}

static int GetLastMeasurement(struct prrt_time_state *tstate) {
	return atomic_read(&(tstate->last_measurement));
}

static int GetVirtualTime(struct prrt_time_state *tstate) {
	return atomic_read(&(tstate->virtual_time));
}

void prrttime_init(struct prrt_time_state *tstate) {
	//int cur_time = (int)prrt_get_mstime(tstate);
	UpdateTimeState(tstate, 0, 0, 0, 0);
}

int prrt_compare_timestamps(struct prrt_time_state *tstate, u32 t1, u32 t2)
{
	/*NOTE: The assumption is that two timestamps will never lie more
	  than half of the maximal number range apart*/

	s32 diff = 0;
	diff = (s32)t1 - (s32)t2;

	if(diff < 0) {
		/*t2 > t1*/
		return -1;
	} else if(diff == 0) {
		/*t2 == t1*/
		return 0;
	} else {
		/*t2 < t1*/
		return 1;
	}
}

u32 prrt_diff_timestamps(struct prrt_time_state *tstate, u32 t1, u32 t2)
{
	s32 diff = 0;
	diff = (s32)t1 - (s32)t2;

	return abs(diff);
}

u32 prrt_get_mstime(struct prrt_time_state *tstate)
{
	struct timespec ts;
	getnstimeofday( &ts );

	return ts.tv_sec*1000+ts.tv_nsec/1000/1000;
}

u32 prrt_get_ustime(struct prrt_time_state *tstate)
{
	struct timespec ts;
	getnstimeofday( &ts );

	return ts.tv_sec*1000000+ts.tv_nsec/1000;
}

u32 prrt_get_prrttime(struct prrt_time_state *tstate)
{
  u32 cur_time = prrt_get_mstime(tstate);
  u32 last_measurement = (u32)GetLastMeasurement(tstate);
  u32 virtual_time = GetVirtualTime(tstate);
  //int diff = cur_time - last_measurement;
  int diff = 0;

  int skew = 0;
  u32 time = 0;

  if (virtual_time)
  {
    diff = prrt_diff_timestamps(tstate, cur_time, last_measurement);
    skew = (diff * GetClockSkew(tstate))/400;
    time = (u32)(virtual_time + GetMeanDeviation(tstate) + skew);
  }
  else
    time = cur_time;
  //PDEBUG(MSG_FINAL, "%s:cur_time %u, virtual_time %u, diff %d, skew %d\n", DEBUG_NAME, cur_time, virtual_time, diff, skew);
  return time;
}

void prrt_update_prrttime(struct prrt_time_state *tstate, u32 reference_time, u32 propagation_delay)
{
	//u32 cur_time = prrt_get_prrttime();
	u32 cur_time = 0;
	int phase_err = 0;
	int mean_dev = GetMeanDeviation(tstate);
	int clock_skew = GetClockSkew(tstate);
	int dev_sum = 0;
	int last_measurement = GetLastMeasurement(tstate);
	int period = 0;
	u32 virtual_time = (u32)GetVirtualTime(tstate);

	//prop_delay = (int)((7*prop_delay + propagation_delay) >> 3);
	//FIXME: handle variable wrap around
	cur_time = prrt_get_mstime(tstate);
	phase_err = reference_time - virtual_time + propagation_delay;
	
	if (abs(phase_err) > 10){
		UpdateTimeState(tstate, 0, 0, (int)cur_time, (int)cur_time);
	}
		
	//phase_err = prrt_diff_timestamps(tstate, reference_time, virtual_time) + propagation_delay;
	dev_sum = (int)(phase_err + 3 * GetMeanDeviation(tstate));
	
	mean_dev = (int)(dev_sum >> 2);
	//period = cur_time - last_measurement;
	period = prrt_diff_timestamps(tstate, cur_time, last_measurement);

	if (period > 0){
		clock_skew = (mean_dev * 1000 / period) + 15 * GetClockSkew(tstate);
		clock_skew = (int)(clock_skew >> 4);
	}

	virtual_time = (u32)(virtual_time + mean_dev + period*clock_skew/400);
	
	UpdateTimeState(tstate, mean_dev, clock_skew, (int)cur_time, (int)virtual_time);
	//PDEBUG(MSG_FINAL, "%s: Mean_dev %d, clock_skew %d, cur_time %u, virtual_time %u, period %d\n", DEBUG_NAME, mean_dev, clock_skew, cur_time, virtual_time, period);
	//if(!(counter++%10)){
	//PDEBUG(MSG_KUIPARAM, "%s: Current time %u, reference time %u, phase error %d, recovered prrt time %u, propagation_delay %u\n", DEBUG_NAME, cur_time, reference_time, phase_err, prrt_get_prrttime(), propagation_delay);
	//PDEBUG(MSG_FINAL, "%s: Mean_dev %d\n", DEBUG_NAME, mean_dev);
	//}
}
