/* 
 * (c) 2010 Saarland University, Telecommunications Lab 
 * (prrt@nt.uni-saarland.de)
 *
 * The contents of this file are subject to the Mozilla Public License
 + Version 1.1 (the "License"); you may not use this file except in
 + compliance with the License. You may obtain a copy of the License at
 + http://www.mozilla.org/MPL/
 + 
 + Software distributed under the License is distributed on an "AS IS"
 + basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 + License for the specific language governing rights and limitations
 + under the License.
 + 
 + The Original Code is "Predictably Reliable Real-time Transport (PRRT)".
 + 
 + The Initial Developer of the Original Code is Saarland University.
 + All Rights Reserved.
 + 
 + Contributor(s): Jochen Gruen (Saarland University), Manuel Gorius
 + (Saarland University), Yongtao Shuai (Saarland University).
 + 
 + Disclaimer: This Alpha version of PRRT is not supported and stable yet.
 + Do not use it in a production environment: it may crash your device, 
 + so please make sure that you have a backup of your data before
 + installing PRRT.
 */

#include "receive_data.h"
#include "defines.h"
#include "prrt_socket.h"
#include "kernel_wrappers.h"

#define DEBUG_NAME "RECEIVE_DATA"

static int poll_packet(struct prrt_socket *msock)
{
	int err = 0;
	//u32 time = 0;
	//u32 time_now = 0;
	struct prrt_packet_container *pcont = 0;
	u16 packet_size = 0;

  pcont = prrt_packet_alloc(&(msock->receive_time_state));
	if(!pcont) {
		return -1;
	}

	/*IMPORTANT: Timeout for this socket has to be set or the module will
	 *           deadlock when this part is waiting for messages that
	 *           do not arrive.*/
	PRINT_TRACE("RECV", "P", "poll_pack");
	PDEBUG(MSG_STABILITY, "%s: Before kernel receive in socket with receiver port %d", DEBUG_NAME, prrtsocket_get_receiveon_port(msock));
	err = wrap_kernel_recvmsg(msock->sock, &pcont->pack, MAX_PRRT_PACKET_SIZE);
	PDEBUG(MSG_STABILITY, "%s: After kernel receive in socket with receiver port %d", DEBUG_NAME, prrtsocket_get_receiveon_port(msock));
	if(err < 0) {
		//PERROR("%s: Error on wrap_kernel_recvmsg (error no. %d)\n", DEBUG_NAME, err);
		if(!pcont) {
			return -1;
		}
		prrt_packet_destroy(&pcont);
		return -ETIMEDOUT;	/*TIMEOUT*/
	}

	packet_size = err;

	prrt_packet_ntoh(pcont);
	PRINT_TRACE("RECV", "P", "polled");
	//prrt_packet_print_header(pcont, MSG_PACK);
	prrt_lossgath_check_fb_limit(&msock->inc_packet_list.lossgath, pcont);

	/*decide if this is data or redundancy*/
	/*this works because the flags field is only 8bit long and does not need to be converted
	 *to host byte order*/
	if(prrt_packet_is_redundancy(pcont)) {
		/*this packet is redundancy*/
		PRINT_TRACE("RECV", "P/IL", "insert_rpack");
		PDEBUG(MSG_STABILITY, "%s: Got redundancy packet in socket with receiver port %d", DEBUG_NAME, prrtsocket_get_receiveon_port(msock));

		if (packet_size > PRRT_HEADER_SIZE + PRRT_RED_HEADER_SIZE)
			pcont->payload_size = packet_size - (PRRT_HEADER_SIZE + PRRT_RED_HEADER_SIZE);
		else 
			pcont->payload_size = 0;

		PDEBUG(MSG_FINAL, "%s: Redundancy packet %u, first_packno %u, packet_size %u at %u\n", DEBUG_NAME,
			pcont->pack.red.header.index, pcont->pack.red.red_header.sn_base, pcont->payload_size, prrt_get_prrttime(&msock->receive_time_state));
		/*insert packet in the list of incoming packets*/
		err = prrtilist_try_decode(&(msock->inc_packet_list), pcont);
		pcont = 0;
	} else if (prrt_packet_is_data(pcont) ||
		   prrt_packet_is_redata(pcont)){
		   
		   PDEBUG(MSG_STABILITY, "%s: Got data packet in socket with receiver port %d", DEBUG_NAME, prrtsocket_get_receiveon_port(msock));
		/*this packet contains data*/
		PRINT_TRACE("RECV", "P/IL", "insert_dpack");
		//time_now = prrt_get_prrttime(&(msock->receive_time_state);
		//PDEBUG(MSG_FBSTALLING, "%s: Inserting data packet(no %u) at %u\n", DEBUG_NAME,
		//	pcont->pack.data.header.packetno, time_now);

		if (packet_size > PRRT_HEADER_SIZE + PRRT_DATA_HEADER_SIZE)
			pcont->payload_size = packet_size - (PRRT_HEADER_SIZE + PRRT_DATA_HEADER_SIZE);
		else 
			pcont->payload_size = 0;

		PDEBUG(MSG_FINAL, "%s: Data packet no %u, index %u, packet_size %u, timestamp %u, group_rtt %u, pack_timeout %u, dec_timeout %u, fb_timeout %u, at %u\n", DEBUG_NAME,
			pcont->pack.data.header.seqnum, pcont->pack.data.header.index, pcont->payload_size, pcont->pack.data.data_header.timestamp, pcont->pack.data.data_header.group_rtt, pcont->pack.data.data_header.pack_timeout, pcont->pack.data.data_header.dec_timeout, pcont->pack.data.data_header.fb_timeout, prrt_get_prrttime(&msock->receive_time_state));

		//wrap_kernel_sendmsg(msock->sockdbg, msock->peerdbg, &(pcont->pack), pcont->payload_size);

		PDEBUG(MSG_PRINTDATA, "%s: Message received, %s\n", DEBUG_NAME,
			&pcont->pack.data.data);
		
		prrtnpar_set_ReplaceTime(&msock->net_params, pcont->pack.data.data_header.group_rtt);

		/*insert packet in the list of incoming packets*/
		err = prrtilist_insert_datapacket(&msock->inc_packet_list, pcont);
	} else {
		PDEBUG(MSG_RECV, "%s: Deleting not recognized packet\n", DEBUG_NAME);
		prrt_packet_destroy(&pcont);
		err = 0;
	}

	if(err < 0){
		if(err != -3)
			PDEBUG(MSG_RECV, "%s: Could not insert packet\n", DEBUG_NAME);
		prrt_packet_destroy(&pcont);
		//return err;	/*this can be just a unnecessary packet, not critical*/
	}

	return 0;
}

static void handle_polling(struct prrt_socket *msock)
{
	int err = 0;
//	u32 d_target = (u32)prrtlpar_get_D_target(&msock->loc_params_in);

	err = poll_packet(msock);
//	while(err >= 0 || 
//		(err == -ETIMEDOUT && !prrt_get_option(msock, TIMEOUT)) ) {
//		PRINT_TRACE("RECV", "P", "del_packs");
//		prrtilist_delete_old_packets(&msock->inc_packet_list, d_target);
//		down(&msock->poll_lock);
//		if(msock->poll_status == ERROR || msock->poll_status == DOWN) {
//			up(&msock->poll_lock);
//			return;
//		}
//		up(&msock->poll_lock);
//
//		err = poll_packet(msock);
//	}

	PRINT_TRACE("RECV", "P", "notpolled");
	//PERROR("%s: Pollfailed, error no is %d\n", DEBUG_NAME, err);

	//down(&msock->poll_lock);
	//msock->poll_status = ERROR;
	//up(&msock->poll_lock);
}

void prrt_poll_packet(struct work_struct *work)
{
	struct prrt_socket *msock = 0;
	struct delayed_work *dw = 0;
	enum worker_thread_status_t status= 0;

	dw = container_of(work, struct delayed_work, work);
	msock = container_of(dw, struct prrt_socket, poll_work);

	down(&msock->poll_lock);
	status = msock->poll_status;
	up(&msock->poll_lock);

	//switch(status){
	//case WORKING:
	//	handle_polling(msock);
	//	break;
	//case ERROR:
	//	/*This is an error*/
	//	//printk(MSG_TRACE "%s: Invalid trace..\n", DEBUG_NAME);
	//	//handle_sleeping(msock);
	//	break;
	//case DOWN:
	//	//handle_down(msock);
	//	break;
	//}
	while( status == WORKING || status == RELAY_WORKING ){
	PDEBUG(MSG_COMMONRECV, "%s: Polling..\n", DEBUG_NAME);
		handle_polling(msock);

		down(&msock->poll_lock);
		status = msock->poll_status;
		up(&msock->poll_lock);
	}

	PDEBUG(MSG_COMMONRECV, "%s: Polling done\n", DEBUG_NAME);
}

