/*
 * fec.c -- forward error correction based on Vandermonde matrices
 * 980624
 * (C) 1997-98 Luigi Rizzo (luigi@iet.unipi.it)
 *
 * Portions derived from code by Phil Karn (karn@ka9q.ampr.org),
 * Robert Morelos-Zaragoza (robert@spectra.eng.hawaii.edu) and Hari
 * Thirumoorthy (harit@spectra.eng.hawaii.edu), Aug 1995
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 */

#include "prrt_block_code.h"
#include <linux/slab.h>

#define DEBUG_NAME "PRRT_FEC_CODER"

// static int count_num;
// static uint64_t tot_encoding_time;
// uint64_t myticks[10];       /* vars for timekeeping */
/*
 * Primitive polynomials - see Lin & Costello, Appendix A,
 * and  Lee & Messerschmitt, p. 453.
 */
static char *allPp[] = {    /* GF_BITS	polynomial		*/
    NULL,		    /*  0	no code			*/
    NULL,		    /*  1	no code			*/
    "111",		    /*  2	1+x+x^2			*/
    "1101",		    /*  3	1+x+x^3			*/
    "11001",		    /*  4	1+x+x^4			*/
    "101001",		    /*  5	1+x^2+x^5		*/
    "1100001",		    /*  6	1+x+x^6			*/
    "10010001",		    /*  7	1 + x^3 + x^7		*/
    "101110001",	    /*  8	1+x^2+x^3+x^4+x^8	*/
    "1000100001",	    /*  9	1+x^4+x^9		*/
    "10010000001",	    /* 10	1+x^3+x^10		*/
    "101000000001",	    /* 11	1+x^2+x^11		*/
    "1100101000001",	    /* 12	1+x+x^4+x^6+x^12	*/
    "11011000000001",	    /* 13	1+x+x^3+x^4+x^13	*/
    "110000100010001",	    /* 14	1+x+x^6+x^10+x^14	*/
    "1100000000000001",	    /* 15	1+x+x^15		*/
    "11010000000010001"	    /* 16	1+x+x^3+x^12+x^16	*/
};


/*
 * To speed up computations, we have tables for logarithm, exponent
 * and inverse of a number. If GF_BITS <= 8, we use a table for
 * multiplication as well (it takes 64K, no big deal even on a PDA,
 * especially because it can be pre-initialized an put into a ROM!),
 * otherwhise we use a table of logarithms.
 * In any case the macro gf_mul(x,y) takes care of multiplications.
 */

// static gf gf_exp[2*GF_SIZE];	/* index->poly form conversion table	*/
// static int gf_log[GF_SIZE + 1];	/* Poly->index form conversion table	*/
// static gf inverse[GF_SIZE+1];	/* inverse of field elem.		*/
// 				/* inv[\alpha**i]=\alpha**(GF_SIZE-i-1)	*/

/*
 * modnn(x) computes x % GF_SIZE, where GF_SIZE is 2**GF_BITS - 1,
 * without a slow divide.
 */
static inline gf
modnn(int x)
{
    while (x >= GF_SIZE) {
	x -= GF_SIZE;
	x = (x >> GF_BITS) + (x & GF_SIZE);
    }
    return x;
}

#define SWAP(a,b,t) {t tmp; tmp=a; a=b; b=tmp;}

/*
 * gf_mul(x,y) multiplies two numbers. If GF_BITS<=8, it is much
 * faster to use a multiplication table.
 *
 * USE_GF_MULC, GF_MULC0(c) and GF_ADDMULC(x) can be used when multiplying
 * many numbers by the same constant. In this case the first
 * call sets the constant, and others perform the multiplications.
 * A value related to the multiplication is held in a local variable
 * declared with USE_GF_MULC . See usage in addmul1().
 */
#if (GF_BITS <= 8)
// static gf gf_mul_table[GF_SIZE + 1][GF_SIZE + 1];

#define gf_mul(cod,x,y) (cod->gf_mul_table[x][y])

#define USE_GF_MULC register gf * __gf_mulc_
#define GF_MULC0(cod, c) __gf_mulc_ = cod->gf_mul_table[c]
#define GF_ADDMULC(cod, dst, x) dst ^= __gf_mulc_[x]

static void
init_mul_table(struct prrt_coder *cod)
{
    int i, j;
    for (i=0; i< GF_SIZE+1; i++)
	for (j=0; j< GF_SIZE+1; j++)
	    cod->gf_mul_table[i][j] = cod->gf_exp[modnn(cod->gf_log[i] + cod->gf_log[j]) ] ;

    for (j=0; j< GF_SIZE+1; j++)
	    cod->gf_mul_table[0][j] = cod->gf_mul_table[j][0] = 0;
}
#else	/* GF_BITS > 8 */
static inline gf
gf_mul(struct prrt_coder *cod, x, y)
{
    if ( (x) == 0 || (y)==0 ) return 0;

    return cod->gf_exp[cod->gf_log[x] + cod->gf_log[y] ] ;
}
#define init_mul_table(cod)

#define USE_GF_MULC register gf * __gf_mulc_
#define GF_MULC0(cod, c) __gf_mulc_ = &(cod->gf_exp[ cod->gf_log[c] ])
#define GF_ADDMULC(cod, dst, x) { if (x) dst ^= __gf_mulc_[ cod->gf_log[x] ] ; }
#endif

/*
 * Generate GF(2**m) from the irreducible polynomial p(X) in p[0]..p[m]
 * Lookup tables:
 *     index->polynomial form		gf_exp[] contains j= \alpha^i;
 *     polynomial form -> index form	gf_log[ j = \alpha^i ] = i
 * \alpha=x is the primitive element of GF(2^m)
 *
 * For efficiency, gf_exp[] has size 2*GF_SIZE, so that a simple
 * multiplication of two numbers can be resolved without calling modnn
 */

/*
 * i use malloc so many times, it is easier to put checks all in
 * one place.
 */
static void *
my_malloc(int sz, char *err_string)
{
    void *p = kmalloc( sz, GFP_KERNEL );
    if (p == NULL) {
	PERROR("%s: KMalloc failure allocating %s\n", DEBUG_NAME, err_string);
    }
    return p;
}

#define NEW_GF_MATRIX(rows, cols) \
    (gf *)my_malloc(rows * cols * sizeof(gf), " ## __LINE__ ## " )

/*
 * initialize the data structures used for computations in GF.
 */
static void
generate_gf(struct prrt_coder *cod)
{
    int i;
    gf mask;
    char *Pp =  allPp[GF_BITS] ;

    mask = 1;	/* x ** 0 = 1 */
    cod->gf_exp[GF_BITS] = 0; /* will be updated at the end of the 1st loop */
    /*
     * first, generate the (polynomial representation of) powers of \alpha,
     * which are stored in gf_exp[i] = \alpha ** i .
     * At the same time build gf_log[gf_exp[i]] = i .
     * The first GF_BITS powers are simply bits shifted to the left.
     */
    for (i = 0; i < GF_BITS; i++, mask <<= 1 ) {
	cod->gf_exp[i] = mask;
	cod->gf_log[cod->gf_exp[i]] = i;
	/*
	 * If Pp[i] == 1 then \alpha ** i occurs in poly-repr
	 * gf_exp[GF_BITS] = \alpha ** GF_BITS
	 */
	if ( Pp[i] == '1' )
	    cod->gf_exp[GF_BITS] ^= mask;
    }
    /*
     * now gf_exp[GF_BITS] = \alpha ** GF_BITS is complete, so can als
     * compute its inverse.
     */
    cod->gf_log[cod->gf_exp[GF_BITS]] = GF_BITS;
    /*
     * Poly-repr of \alpha ** (i+1) is given by poly-repr of
     * \alpha ** i shifted left one-bit and accounting for any
     * \alpha ** GF_BITS term that may occur when poly-repr of
     * \alpha ** i is shifted.
     */
    mask = 1 << (GF_BITS - 1 ) ;
    for (i = GF_BITS + 1; i < GF_SIZE; i++) {
	if (cod->gf_exp[i - 1] >= mask)
	    cod->gf_exp[i] = cod->gf_exp[GF_BITS] ^ ((cod->gf_exp[i - 1] ^ mask) << 1);
	else
	    cod->gf_exp[i] = cod->gf_exp[i - 1] << 1;
	cod->gf_log[cod->gf_exp[i]] = i;
    }
    /*
     * log(0) is not defined, so use a special value
     */
    cod->gf_log[0] =	GF_SIZE ;
    /* set the extended gf_exp values for fast multiply */
    for (i = 0 ; i < GF_SIZE ; i++)
	cod->gf_exp[i + GF_SIZE] = cod->gf_exp[i] ;

    /*
     * again special cases. 0 has no inverse. This used to
     * be initialized to GF_SIZE, but it should make no difference
     * since noone is supposed to read from here.
     */
    cod->inverse[0] = 0 ;
    cod->inverse[1] = 1;
    for (i=2; i<=GF_SIZE; i++)
	cod->inverse[i] = cod->gf_exp[GF_SIZE-cod->gf_log[i]];
}

/*
 * Various linear algebra operations that i use often.
 */

/*
 * addmul() computes dst[] = dst[] + c * src[]
 * This is used often, so better optimize it! Currently the loop is
 * unrolled 16 times, a good value for 486 and pentium-class machines.
 * The case c=0 is also optimized, whereas c=1 is not. These
 * calls are unfrequent in my typical apps so I did not bother.
 *
 * Note that gcc on
 */
#define addmul(cod, dst, src, c, sz) \
    if (c != 0) addmul1(cod, dst, src, c, sz)

#define UNROLL 16 /* 1, 4, 8, 16 */
static void
addmul1(struct prrt_coder *cod, gf *dst1, gf *src1, gf c, int sz)
{
	USE_GF_MULC ;
	register gf *dst = dst1, *src = src1 ;
	gf *lim = &dst[sz - UNROLL + 1] ;


	PDEBUGG(MSG_BCDEB, "%s: Addmul 1\n", DEBUG_NAME);
	GF_MULC0(cod, c) ;

	PDEBUGG(MSG_BCDEB, "%s: Addmul 2\n", DEBUG_NAME);

#if (UNROLL > 1) /* unrolling by 8/16 is quite effective on the pentium */
	for (; dst < lim ; dst += UNROLL, src += UNROLL ) {
		GF_ADDMULC( cod, dst[0] , src[0] );
		GF_ADDMULC( cod, dst[1] , src[1] );
		GF_ADDMULC( cod, dst[2] , src[2] );
		GF_ADDMULC( cod, dst[3] , src[3] );
#if (UNROLL > 4)
		GF_ADDMULC( cod, dst[4] , src[4] );
		GF_ADDMULC( cod, dst[5] , src[5] );
		GF_ADDMULC( cod, dst[6] , src[6] );
		GF_ADDMULC( cod, dst[7] , src[7] );
#endif
#if (UNROLL > 8)
		GF_ADDMULC( cod, dst[8] , src[8] );
		GF_ADDMULC( cod, dst[9] , src[9] );
		GF_ADDMULC( cod, dst[10] , src[10] );
		GF_ADDMULC( cod, dst[11] , src[11] );
		GF_ADDMULC( cod, dst[12] , src[12] );
		GF_ADDMULC( cod, dst[13] , src[13] );
		GF_ADDMULC( cod, dst[14] , src[14] );
		GF_ADDMULC( cod, dst[15] , src[15] );
#endif
	}
#endif
	PDEBUGG(MSG_BCDEB, "%s: Addmul 3\n", DEBUG_NAME);

	lim += UNROLL - 1 ;
	for (; dst < lim; dst++, src++ )		/* final components */
		GF_ADDMULC( cod, *dst , *src );
}

/*
 * computes C = AB where A is n*k, B is k*m, C is n*m
 */
static void
matmul(struct prrt_coder *cod, gf *a, gf *b, gf *c, int n, int k, int m)
{
    int row, col, i ;

    for (row = 0; row < n ; row++) {
	for (col = 0; col < m ; col++) {
	    gf *pa = &a[ row * k ];
	    gf *pb = &b[ col ];
	    gf acc = 0 ;
	    for (i = 0; i < k ; i++, pa++, pb += m )
		acc ^= gf_mul( cod, *pa, *pb ) ;
	    c[ row * m + col ] = acc ;
	}
    }
}

#ifdef FECDEBUG
/*
 * returns 1 if the square matrix is identiy
 * (only for test)
 */
static int
is_identity(gf *m, int k)
{
    int row, col ;
    for (row=0; row<k; row++)
	for (col=0; col<k; col++)
	    if ( (row==col && *m != 1) ||
		 (row!=col && *m != 0) )
		 return 0 ;
	    else
		m++ ;
    return 1 ;
}
#endif /* debug */

/*
 * invert_mat() takes a matrix and produces its inverse
 * k is the size of the matrix.
 * (Gauss-Jordan, adapted from Numerical Recipes in C)
 * Return non-zero if singular.
 */
DEB( int pivloops=0; int pivswaps=0 ; /* diagnostic */)
static int
invert_mat(struct prrt_coder *cod, gf *src, int k)
{
	gf c, *p ;
	int irow, icol, row, col, i, ix ;

	int error = 1 ;
	int *indxc = (int*)my_malloc(k*sizeof(int), "indxc");
	int *indxr = (int*)my_malloc(k*sizeof(int), "indxr");
	int *ipiv = (int*)my_malloc(k*sizeof(int), "ipiv");
	gf *id_row = NEW_GF_MATRIX(1, k);
	gf *temp_row = NEW_GF_MATRIX(1, k);

	if(!indxc || !indxr || !ipiv || !id_row || !temp_row) {
		PERROR("%s: Could not allocate ressources needed for invert_mat\n", DEBUG_NAME);
		goto fail;
	}

	bzero(id_row, k*sizeof(gf));
	DEB( pivloops=0; pivswaps=0 ; /* diagnostic */ )
	/*
	* ipiv marks elements already used as pivots.
	*/
	for (i = 0; i < k ; i++)
		ipiv[i] = 0 ;

	for (col = 0; col < k ; col++) {
		gf *pivot_row ;
		/*
		* Zeroing column 'col', look for a non-zero element.
		* First try on the diagonal, if it fails, look elsewhere.
		*/
		irow = icol = -1 ;
		if (ipiv[col] != 1 && src[col*k + col] != 0) {
			irow = col ;
			icol = col ;
			goto found_piv ;
		}
		for (row = 0 ; row < k ; row++) {
			if (ipiv[row] != 1) {
				for (ix = 0 ; ix < k ; ix++) {
					DEB( pivloops++ ; )
					if (ipiv[ix] == 0) {
						if (src[row*k + ix] != 0) {
						irow = row ;
						icol = ix ;
						goto found_piv ;
						}
					} else if (ipiv[ix] > 1) {
						PERROR("%s: singular matrix\n", DEBUG_NAME);
						goto fail ;
					}
				}
			}
		}
		if (icol == -1) {
			PERROR("%s: XXX pivot not found!\n", DEBUG_NAME);
			goto fail ;
		}
found_piv:
		++(ipiv[icol]) ;
		/*
		* swap rows irow and icol, so afterwards the diagonal
		* element will be correct. Rarely done, not worth
		* optimizing.
		*/
		if (irow != icol) {
			for (ix = 0 ; ix < k ; ix++ ) {
				SWAP( src[irow*k + ix], src[icol*k + ix], gf) ;
			}
		}
		indxr[col] = irow ;
		indxc[col] = icol ;
		pivot_row = &src[icol*k] ;
		c = pivot_row[icol] ;
		if (c == 0) {
			PERROR("%s: singular matrix 2\n", DEBUG_NAME);
			goto fail ;
		}
		if (c != 1 ) { /* otherwhise this is a NOP */
		/*
		* this is done often , but optimizing is not so
		* fruitful, at least in the obvious ways (unrolling)
		*/
		DEB( pivswaps++ ; )
		c = cod->inverse[ c ] ;
		pivot_row[icol] = 1 ;
		for (ix = 0 ; ix < k ; ix++ )
			pivot_row[ix] = gf_mul(cod, c, pivot_row[ix] );
		}
		/*
		* from all rows, remove multiples of the selected row
		* to zero the relevant entry (in fact, the entry is not zero
		* because we know it must be zero).
		* (Here, if we know that the pivot_row is the identity,
		* we can optimize the addmul).
		*/
		id_row[icol] = 1;
		if (bcmp(pivot_row, id_row, k*sizeof(gf)) != 0) {
			for (p = src, ix = 0 ; ix < k ; ix++, p += k ) {
				if (ix != icol) {
					c = p[icol] ;
					p[icol] = 0 ;
					addmul(cod, p, pivot_row, c, k );
				}
			}
		}
		id_row[icol] = 0;
	} /* done all columns */
	for (col = k-1 ; col >= 0 ; col-- ) {
		if (indxr[col] <0 || indxr[col] >= k) {
			PERROR("%s: AARGH, indxr[col] %d\n", DEBUG_NAME, indxr[col]);
		} else if (indxc[col] <0 || indxc[col] >= k) {
			PERROR("%s: AARGH, indxc[col] %d\n", DEBUG_NAME, indxc[col]);
		} else if (indxr[col] != indxc[col] ) {
			for (row = 0 ; row < k ; row++ ) {
				SWAP( src[row*k + indxr[col]], src[row*k + indxc[col]], gf) ;
			}
		}
	}
	error = 0 ;
fail:
	kfree(indxc);
	kfree(indxr);
	kfree(ipiv);
	kfree(id_row);
	kfree(temp_row);
	return error ;
}

/*
 * fast code for inverting a vandermonde matrix.
 * XXX NOTE: It assumes that the matrix
 * is not singular and _IS_ a vandermonde matrix. Only uses
 * the second column of the matrix, containing the p_i's.
 *
 * Algorithm borrowed from "Numerical recipes in C" -- sec.2.8, but
 * largely revised for my purposes.
 * p = coefficients of the matrix (p_i)
 * q = values of the polynomial (known)
 */

int
invert_vdm(struct prrt_coder *cod, gf *src, int k)
{
	int i, j, row, col ;
	gf *b, *c, *p;
	gf t, xx ;
	int err = 0;

	if (k == 1) 	/* degenerate case, matrix must be p^0 = 1 */
		return 0 ;
	/*
	* c holds the coefficient of P(x) = Prod (x - p_i), i=0..k-1
	* b holds the coefficient for the matrix inversion
	*/
	c = NEW_GF_MATRIX(1, k);
	b = NEW_GF_MATRIX(1, k);
	p = NEW_GF_MATRIX(1, k);

	if(!c || !b || !p) {
		err = -1;
		goto fail;
	}

	for ( j=1, i = 0 ; i < k ; i++, j+=k ) {
		c[i] = 0 ;
		p[i] = src[j] ;    /* p[i] */
	}
	/*
	* construct coeffs. recursively. We know c[k] = 1 (implicit)
	* and start P_0 = x - p_0, then at each stage multiply by
	* x - p_i generating P_i = x P_{i-1} - p_i P_{i-1}
	* After k steps we are done.
	*/
	c[k-1] = p[0] ;	/* really -p(0), but x = -x in GF(2^m) */
	for (i = 1 ; i < k ; i++ ) {
		gf p_i = p[i] ; /* see above comment */
		for (j = k-1  - ( i - 1 ) ; j < k-1 ; j++ )
			c[j] ^= gf_mul( cod, p_i, c[j+1] ) ;
		c[k-1] ^= p_i ;
	}

	for (row = 0 ; row < k ; row++ ) {
		/*
		* synthetic division etc.
		*/
		xx = p[row] ;
		t = 1 ;
		b[k-1] = 1 ; /* this is in fact c[k] */
		for (i = k-2 ; i >= 0 ; i-- ) {
			b[i] = c[i+1] ^ gf_mul(cod, xx, b[i+1]) ;
			t = gf_mul(cod, xx, t) ^ b[i] ;
		}
		for (col = 0 ; col < k ; col++ )
			src[col*k + row] = gf_mul(cod, cod->inverse[t], b[col] );
	}

fail:
	kfree(c) ;
	kfree(b) ;
	kfree(p) ;
	return err ;
}

//static int fec_initialized = 0 ;
static void init_fec(struct prrt_coder *cod)
{
    TICK(myticks[0]);
    generate_gf(cod);
    TOCK(myticks[0]);
    DDB(PDEBUG(MSG_COMMON, "%s: generate_gf took %ldus\n", DEBUG_NAME, myticks[0]);)
    TICK(myticks[0]);
    init_mul_table(cod);
    TOCK(myticks[0]);
    DDB(PDEBUG(MSG_COMMON,, "%s: init_mul_table took %ldus\n", DEBUG_NAME, myticks[0]);)
}

/*
 * This section contains the proper FEC encoding/decoding routines.
 * The encoding matrix is computed starting with a Vandermonde matrix,
 * and then transforming it into a systematic matrix.
 */

int prrt_fec_new(struct prrt_coder **cod, int k, int n);

int prrt_fec_get_coder(struct prrt_coder **cod, int n, int k)
{
	int err = 0;

	if(*cod == 0 ||
	   (*cod)->params.n != n ||
	   (*cod)->params.k != k) {
		prrt_fec_free(*cod);
		err = prrt_fec_new(cod, k, n);
	}

	return err;
}

void prrt_fec_free(struct prrt_coder *cod)
{
	int k;
	int n;
	gf *enc_matrix;

	if( cod == 0 )
		return;

	enc_matrix = cod->params.enc_matrix;
	k = cod->params.k;
	n = cod->params.n;

	if ( cod->params.magic != ( ( (FEC_MAGIC ^ k) ^ n) ^ (int)(enc_matrix)) ){
		PERROR("%s: bad parameters to fec_free\n", DEBUG_NAME);
		return ;
	}
	kfree(enc_matrix);
	kfree(cod);
}

int prrt_fec_new(struct prrt_coder **cod, int k, int n)
{
	int row, col ;
	gf *p, *tmp_m ;
	int err = 0;

	PDEBUG(MSG_BCDEB, "%s: In prrt_fec_new\n", DEBUG_NAME);
	if (k > GF_SIZE + 1 || n > GF_SIZE + 1 || k > n ) {
		PERROR("%s: Invalid parameters k %d n %d GF_SIZE %d\n", DEBUG_NAME,
			k, n, GF_SIZE );
		return -1;
	}

	PDEBUG(MSG_BCDEB, "%s: Allocating memory for coder\n", DEBUG_NAME);
	*cod = kmalloc(sizeof(struct prrt_coder), GFP_KERNEL);
	if(!(*cod)) {
		PERROR("%s: Not enough memory for coder!\n", DEBUG_NAME);
		return -2;
	}

	PDEBUG(MSG_BCDEB, "%s: Setting fec parameter\n", DEBUG_NAME);
	/*set fec parameter*/
	(*cod)->params.k = k;
	(*cod)->params.n = n;
	(*cod)->params.enc_matrix = NEW_GF_MATRIX(n, k);
	if(!(*cod)->params.enc_matrix) {
		(*cod)->params.enc_matrix = 0;
		PERROR("%s: Could not allocate enc_matrix for coder\n", DEBUG_NAME);
		err = -2;
		goto error;
	}

	(*cod)->params.magic = ( ( FEC_MAGIC ^ k) ^ n) ^ (int)((*cod)->params.enc_matrix) ;

	PDEBUG(MSG_BCDEB, "%s: Initialising tables\n", DEBUG_NAME);
	/*init tables*/
	init_fec(*cod);

	/*init auxiliary*/
	(*cod)->count_num=0;
	(*cod)->tot_encoding_time=0;

	PDEBUG(MSG_BCDEB, "%s: Allocating new encoding matrix\n", DEBUG_NAME);
	tmp_m = NEW_GF_MATRIX(n, k);
	if(!tmp_m) {
		PERROR("%s: Could not allocate tmp_m matrix for new coder\n", DEBUG_NAME);
		err = -2;
		goto error;
	}
	/*
	* fill the matrix with powers of field elements, starting from 0.
	* The first row is special, cannot be computed with exp. table.
	*/
	PDEBUG(MSG_BCDEB, "%s: Setting first element of first column of tmp_m\n", DEBUG_NAME);
	tmp_m[0] = 1 ;
	PDEBUG(MSG_BCDEB, "%s: Setting rest of first column of tmp_m\n", DEBUG_NAME);
	for (col = 1; col < k ; col++)
		tmp_m[col] = 0 ;
	PDEBUG(MSG_BCDEB, "%s: Setting other columns of tmp_m\n", DEBUG_NAME);
	for (p = tmp_m + k, row = 0; row < n-1 ; row++, p += k) {
		for ( col = 0 ; col < k ; col ++ )
		p[col] = (*cod)->gf_exp[modnn(row*col)];
	}

	/*
	* quick code to build systematic matrix: invert the top
	* k*k vandermonde matrix, multiply right the bottom n-k rows
	* by the inverse, and construct the identity matrix at the top.
	*/
	TICK(myticks[3]);
	PDEBUG(MSG_BCDEB, "%s: Inverting vdm matrix\n", DEBUG_NAME);
	err = invert_vdm(*cod, tmp_m, k); /* much faster than invert_mat */
	if(err < 0) {
		PERROR("%s: Could not invert vdm matrix\n", DEBUG_NAME);
		err = -2;
		goto error;
	}

	PDEBUG(MSG_BCDEB, "%s: Multiplicate matrixes\n", DEBUG_NAME);
	matmul(*cod, tmp_m + k*k, tmp_m, (*cod)->params.enc_matrix + k*k, n - k, k, k);
	/*
	* the upper matrix is I so do not bother with a slow multiply
	*/
	PDEBUG(MSG_BCDEB, "%s: Zero out enc_matrix\n", DEBUG_NAME);
	bzero((*cod)->params.enc_matrix, k*k*sizeof(gf) );
	PDEBUG(MSG_BCDEB, "%s: Set elements of enc_matrix\n", DEBUG_NAME);
	for (p = (*cod)->params.enc_matrix, col = 0 ; col < k ; col++, p += k+1 )
		*p = 1 ;
	PDEBUG(MSG_BCDEB, "%s: Free tmp_m\n", DEBUG_NAME);
	kfree(tmp_m);
	TOCK(myticks[3]);

	PDEBUG(MSG_BCDEB, "%s: Coder ready\n", DEBUG_NAME);
	DDB(PDEBUG(MSG_COMMON, "%s: %ld us to build encoding matrix\n", DEBUG_NAME, myticks[3]);)
	/*DEB(pr_matrix((*cod)->params.enc_matrix, n, k, "encoding_matrix");)*/

	return 0;

error:
	kfree(*cod);
	kfree((*cod)->params.enc_matrix);
	return err;
}

/*
 * fec_encode accepts as input pointers to n data packets of size sz,
 * and produces as output a packet pointed to by fec, computed
 * with index "index".
 */
void
prrt_fec_encode(struct prrt_coder *cod, gf *src[], gf *fec, int index, int sz)
{
    int i, k = cod->params.k ;
    gf *p ;
// #if (TEST > 0)
//     int par_num=cod->params.n-cod->params.k;
//     uint64_t avg_encoding_time=0;
//     struct timeval t1, t2;
//     gettimeofday(&t1, NULL);
// #endif

    if (GF_BITS > 8)
	sz /= 2 ;

    if (index < k)
         bcopy(src[index], fec, sz*sizeof(gf) ) ;
    else if (index < cod->params.n) {
	p = &(cod->params.enc_matrix[index*k] );
	bzero(fec, sz*sizeof(gf));
	for (i = 0; i < k ; i++)
	    addmul(cod, fec, src[i], p[i], sz ) ;
    } else
	PERROR("%s: Invalid index %d (max %d)\n", DEBUG_NAME, index, cod->params.n - 1 );
// #if (TEST > 0)
//    gettimeofday(&t2, NULL);
//    cod->count_num++;
//    cod->tot_encoding_time=cod->tot_encoding_time+DIFF_T(t2,t1);
//    if ((cod->count_num%par_num)==0) {
//      avg_encoding_time=cod->tot_encoding_time/par_num;
//      PDEBUG(MSG_COMMON, "%s: Average Encoding Time:  one packet %ld us\n", DEBUG_NAME,
// 		avg_encoding_time);
//      cod->count_num=0;
//      cod->tot_encoding_time=0;
//    }
// #endif
}

/*
 * shuffle move src packets in their position
 * NOTE:not used atm
 */
// static int
// shuffle(gf *pkt[], int index[], int k)
// {
//     int i;
//
//     for ( i = 0 ; i < k ; ) {
// 	if (index[i] >= k || index[i] == i)
// 	    i++ ;
// 	else {
// 	    /*
// 	     * put pkt in the right position (first check for conflicts).
// 	     */
// 	    int c = index[i] ;
//
// 	    if (index[c] == c) {
// 		DEB(PERROR("%s: shuffle, error at %d\n", DEBUG_NAME, i);)
// 		return 1 ;
// 	    }
// 	    SWAP(index[i], index[c], int) ;
// 	    SWAP(pkt[i], pkt[c], gf *) ;
// 	}
//     }
//     DEB( /* just test that it works... */
// 	for ( i = 0 ; i < k ; i++ ) {
// 		if (index[i] < k && index[i] != i) {
// 		PDEBUG(MSG_BCDEB, "%s: shuffle: after\n", DEBUG_NAME);
// 		for (i=0; i<k ; i++)
// 			PDEBUG(MSG_BCDEB, "%s: %3d\n", DEBUG_NAME, index[i]);
// 		return 1 ;
// 		}
// 	}
//     )
//     return 0 ;
// }

/*
 * build_decode_matrix constructs the encoding matrix given the
 * indexes. The matrix must be already allocated as
 * a vector of k*k elements, in row-major order
 */
static gf *
build_decode_matrix(struct prrt_coder *cod, int index[])
{
	int i , k = cod->params.k ;
	gf *p, *matrix = NEW_GF_MATRIX(k, k);
	if(!matrix)
		goto out;

	TICK(myticks[9]);
	for (i = 0, p = matrix ; i < k ; i++, p += k ) {
#if 1 /* this is simply an optimization, not very useful indeed */
		if (index[i] < k) {
			bzero(p, k*sizeof(gf) );
			p[i] = 1 ;
		} else
#endif
		if (index[i] < cod->params.n )
			bcopy( &(cod->params.enc_matrix[index[i]*k]), p, k*sizeof(gf) );
		else {
			PERROR("%s: decode: invalid index %d (max %d)\n", DEBUG_NAME,
				index[i], cod->params.n - 1 );
			kfree(matrix) ;
			return NULL ;
		}
	}
	TICK(myticks[9]);
	if (invert_mat(cod, matrix, k)) {
		kfree(matrix);
		matrix = NULL ;
	}
	TOCK(myticks[9]);
out:
	return matrix ;
}

int prrt_fec_decode(struct prrt_coder *cod, gf *pkt[], int index[], int sz)
{
	gf *m_dec = 0;
	gf **new_pkt = 0;
	int row, col , k = cod->params.k ;
	int err = 0;
	int i = 0;
// #if (TEST > 0)
//     struct timeval t1, t2;
//     gettimeofday(&t1, NULL);
// #endif

	/*check input fields*/
	for(i=0; i<k; i++) {
		if(pkt[i] == 0 || index[i] == -1) {
			PERROR("%s: Input data not valid, missing packet at position %d\n",
				DEBUG_NAME, i);
			return -1;
		}
	}

	if (GF_BITS > 8)
		sz /= 2 ;

//     if (shuffle(pkt, index, k))	/* error if true */
// 	return 1 ;
	PDEBUG(MSG_BCDEB, "%s: Building decode matrix\n", DEBUG_NAME);
	m_dec = build_decode_matrix(cod, index);

	if (!m_dec) {
		PERROR("%s: Could not build decode matrix\n", DEBUG_NAME);
		err = -1 ; /* error */
		goto error;
	}
	/*
	* do the actual decoding
	*/
	PDEBUG(MSG_BCDEB, "%s: Decode packets\n", DEBUG_NAME);
	new_pkt = (gf**)my_malloc (k * sizeof (gf * ), "new pkt pointers" );
	if(!new_pkt) {
		PERROR("%s: Could not allocate memory for temporary packet\n", DEBUG_NAME);
		err = -1;
		goto error;
	}
	bzero(new_pkt, k * sizeof (gf *));

	for (row = 0 ; row < k ; row++ ) {
		if (index[row] >= k) {
			new_pkt[row] = (gf*)my_malloc (sz * sizeof (gf), "new pkt buffer" );
			if(!new_pkt[row]) {
				PERROR("%s: No memory for packet row\n", DEBUG_NAME);
				goto free_rows;
			}
			bzero(new_pkt[row], sz * sizeof(gf) ) ;
			for (col = 0 ; col < k ; col++ )
				addmul(cod, new_pkt[row], pkt[col], m_dec[row*k + col], sz) ;
		}
	}
	/*
	* move pkts to their final destination
	*/
	PDEBUG(MSG_BCDEB, "%s: Move packets to final positions\n", DEBUG_NAME);
	for (row = 0 ; row < k ; row++ ) {
		if (index[row] >= k) {
			bcopy(new_pkt[row], pkt[row], sz*sizeof(gf));
			kfree(new_pkt[row]);
		}
	}
// #if (TEST > 0)
//     gettimeofday(&t2, NULL);
//     PDEBUG(MSG_COMMON, "%s: Decoding Time:  one block %ld us\n", DEBUG_NAME, DIFF_T(t2,t1));
// #endif
error:
	kfree(new_pkt);
	kfree(m_dec);


	return err;

free_rows:
	for (row = 0 ; row < k ; row++ ) {
		if (index[row] >= k) {
			kfree(new_pkt[row]);
		}
	}

	kfree(new_pkt);
	kfree(m_dec);

	return -1;
}

/*********** end of FEC code -- beginning of test code ************/

// #if (TEST || FECDEBUG)
// void
// test_gf(struct prrt_coder *cod)
// {
//     int i ;
//     /*
//      * test gf tables. Sufficiently tested...
//      */
//     for (i=0; i<= GF_SIZE; i++) {
//         if (gf_exp[gf_log[i]] != i)
// 	    PDEBUG(MSG_COMMON, "%s: bad exp/log i %d log %d exp(log) %d\n", DEBUG_NAME,
// 		i, gf_log[i], gf_exp[gf_log[i]]);
//
//         if (i != 0 && gf_mul(cod, i, inverse[i]) != 1)
// 	    PDEBUG(MSG_COMMON, "%s: bad mul/inv i %d inv %d i*inv(i) %d\n", DEBUG_NAME,
// 		i, inverse[i], gf_mul(cod, i, inverse[i]) );
// 	if (gf_mul(cod, 0,i) != 0)
// 	    PDEBUG(MSG_COMMON, "%s: bad mul table 0,%d\n", DEBUG_NAME, i);
// 	if (gf_mul(cod, i,0) != 0)
// 	    PDEBUG(MSG_COMMON, "%s: bad mul table %d,0\n", DEBUG_NAME, i);
//     }
// }
// #endif /* TEST */
